/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "lorenz95/LocalizationMatrixL95.h"

#include <algorithm>
#include <cmath>

#include "lorenz95/IncrementL95.h"
#include "lorenz95/Resolution.h"

#include "eckit/config/Configuration.h"

// -----------------------------------------------------------------------------
namespace lorenz95 {
// -----------------------------------------------------------------------------

LocalizationMatrixL95::LocalizationMatrixL95(const Resolution & geom,
                                             const eckit::Configuration & config)
  : resol_(geom.npoints()),
    rscale_(1.0/config.getDouble("length_scale"))
{
// Gaussian structure function
  unsigned int size = resol_/2+1;
  std::vector<double> locfct(resol_);
  for (unsigned int jj = 0; jj < resol_; ++jj) {
    double zz = rscale_ * std::min(jj, resol_-jj);
    locfct[jj] = std::exp(-0.5*zz*zz);
  }
// Go to Fourier space
  std::vector<std::complex<double> > four(resol_);
  fft_.fwd(four, locfct);
// Save Fourier coefficients
  coefs_.resize(size);
  for (unsigned int jj = 0; jj < size; ++jj) {
    coefs_[jj] = std::real(four[jj]);
  }
}

// -----------------------------------------------------------------------------

LocalizationMatrixL95::~LocalizationMatrixL95() {}

// -----------------------------------------------------------------------------

void LocalizationMatrixL95::multiply(IncrementL95 & dx) const {
  unsigned int size = resol_/2+1;
  std::vector<std::complex<double> > four(resol_);
  fft_.fwd(four, dx.asVector());
  for (unsigned int jj = 0; jj < size; ++jj) {
    four[jj] *= coefs_[jj];
  }
  fft_.inv(dx.asVector(), four);
}

// -----------------------------------------------------------------------------

void LocalizationMatrixL95::print(std::ostream & os) const {
  os << "LocalizationMatrixL95: resolution = " << resol_
     << ", length scale = " << 1.0/rscale_ << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace lorenz95
