/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "lorenz95/ObsBiasCorrection.h"

#include <iostream>
#include <string>

#include "lorenz95/ObsBias.h"
#include "lorenz95/ObsBiasCovariance.h"

#include "eckit/config/Configuration.h"
#include "util/Logger.h"

// -----------------------------------------------------------------------------
namespace lorenz95 {
// -----------------------------------------------------------------------------
static oops::ObsAuxIncrementMaker<ObsBiasCorrection> makesObsBiasCorrection_("ObsBiasL95");
// -----------------------------------------------------------------------------
ObsBiasCorrection::ObsBiasCorrection(const eckit::Configuration & conf)
  : bias_(0.0), active_(false), dummyKey_(0)
{
  active_ = conf.has("standard_deviation");
  if (active_) {oops::Log::trace() << "ObsBiasCorrection::ObsBiasCorrection created." << std::endl;}
}
// -----------------------------------------------------------------------------
ObsBiasCorrection::ObsBiasCorrection(const ObsBiasCorrection & other,
                                     const bool copy)
  : bias_(0.0), active_(other.active_), dummyKey_(other.dummyKey_)
{
  if (active_ && copy) bias_ = other.bias_;
}
// -----------------------------------------------------------------------------
ObsBiasCorrection::ObsBiasCorrection(const ObsBiasCorrection & other,
                                     const eckit::Configuration &)
  : bias_(0.0), active_(other.active_)
{
  if (active_) bias_ = other.bias_;
}
// -----------------------------------------------------------------------------
ObsBiasCorrection* ObsBiasCorrection::clone(const bool copy) const {
  return new ObsBiasCorrection(*this, copy);
}
// -----------------------------------------------------------------------------
ObsBiasCorrection* ObsBiasCorrection::clone(const eckit::Configuration & conf) const {
  return new ObsBiasCorrection(*this, conf);
}
// -----------------------------------------------------------------------------
void ObsBiasCorrection::diff(const oops::ObsAuxControlBase & b1, const oops::ObsAuxControlBase & b2) {
  const ObsBias * pb1 = dynamic_cast<const ObsBias*>(&b1);
  const ObsBias * pb2 = dynamic_cast<const ObsBias*>(&b2);
  ASSERT(pb1 != nullptr && pb2 != nullptr);
  if (active_) bias_ = (*pb1).value() - (*pb2).value();
}
// -----------------------------------------------------------------------------
void ObsBiasCorrection::zero() {
  bias_ = 0.0;
}
// -----------------------------------------------------------------------------
ObsBiasCorrection & ObsBiasCorrection::operator=(const oops::ObsAuxIncrementBase & rhs) {
  if (active_) {
    const ObsBiasCorrection * prhs = dynamic_cast<const ObsBiasCorrection*>(&rhs);
    ASSERT(prhs != nullptr);
    bias_ = (*prhs).value();
  } else {
    bias_ = 0.0;
  }
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasCorrection & ObsBiasCorrection::operator=(const ObsBiasCorrection & rhs) {
  if (active_) {
    bias_ = rhs.bias_;
  } else {
    bias_ = 0.0;
  }
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasCorrection & ObsBiasCorrection::operator+=(const oops::ObsAuxIncrementBase & rhs) {
  if (active_) {
    const ObsBiasCorrection * prhs = dynamic_cast<const ObsBiasCorrection*>(&rhs);
    ASSERT(prhs != nullptr);
    bias_ += (*prhs).value();
  }
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasCorrection & ObsBiasCorrection::operator-=(const oops::ObsAuxIncrementBase & rhs) {
  if (active_) {
    const ObsBiasCorrection * prhs = dynamic_cast<const ObsBiasCorrection*>(&rhs);
    ASSERT(prhs != nullptr);
    bias_ -= (*prhs).value();
  }
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasCorrection & ObsBiasCorrection::operator*=(const double & fact) {
  if (active_) bias_ *= fact;
  return *this;
}
// -----------------------------------------------------------------------------
void ObsBiasCorrection::axpy(const double & fact, const oops::ObsAuxIncrementBase & rhs) {
  if (active_) {
    const ObsBiasCorrection * prhs = dynamic_cast<const ObsBiasCorrection*>(&rhs);
    ASSERT(prhs != nullptr);
    bias_ += fact * (*prhs).value();
  }
}
// -----------------------------------------------------------------------------
double ObsBiasCorrection::dot_product_with(const oops::ObsAuxIncrementBase & rhs) const {
  double zz = 0.0;
  if (active_) {
    const ObsBiasCorrection * prhs = dynamic_cast<const ObsBiasCorrection*>(&rhs);
    ASSERT(prhs != nullptr);
    zz = bias_ * (*prhs).value();
  }
  return zz;
}
// -----------------------------------------------------------------------------
void ObsBiasCorrection::print(std::ostream & os) const {
  if (active_) {os << std::endl << "ObsBiasCorrection = " << bias_;}
}
// -----------------------------------------------------------------------------
}  // namespace lorenz95
