/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/GenericCtlVecQG.h"

#include <cmath>
#include <string>

#include "model/QgFortran.h"

#include "eckit/config/Configuration.h"

#include "util/Logger.h"
#include "util/DateTime.h"

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------
GenericCtlVecQG::GenericCtlVecQG(const int & nx, const int & ny, const int & nl)
  : nx_(nx), ny_(ny), nl_(nl), keyCtlVec_(0)
{
  qg_generic_ctl_vec_create_f90(keyCtlVec_, nx, ny, nl);
}
// -----------------------------------------------------------------------------
GenericCtlVecQG::GenericCtlVecQG(const GenericCtlVecQG & other, const bool copy)
  : nx_(other.nx_), ny_(other.ny_), nl_(other.nl_), keyCtlVec_(0)
{
  qg_generic_ctl_vec_create_f90(keyCtlVec_, other.nx_, other.ny_, other.nl_);
  if (copy) {
    qg_generic_ctl_vec_copy_f90(keyCtlVec_, other.keyCtlVec_);
  } else {
    qg_generic_ctl_vec_zero_f90(keyCtlVec_);
  }
}
// -----------------------------------------------------------------------------
GenericCtlVecQG::GenericCtlVecQG(const GenericCtlVecQG & other)
  : nx_(other.nx_), ny_(other.ny_), nl_(other.nl_), keyCtlVec_(0)
{
  qg_generic_ctl_vec_create_f90(keyCtlVec_, other.nx_, other.ny_, other.nl_);
  qg_generic_ctl_vec_copy_f90(keyCtlVec_, other.keyCtlVec_);
}
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
GenericCtlVecQG::~GenericCtlVecQG() {
  qg_generic_ctl_vec_delete_f90(keyCtlVec_);
}
// -----------------------------------------------------------------------------
GenericCtlVecQG & GenericCtlVecQG::operator=(const GenericCtlVecQG & rhs) {
  ASSERT(nx_ == rhs.nx_ && ny_ == rhs.ny_ && nl_ == rhs.nl_);
  qg_generic_ctl_vec_copy_f90(keyCtlVec_, rhs.keyCtlVec_);
  return *this;
}
// -----------------------------------------------------------------------------
GenericCtlVecQG & GenericCtlVecQG::operator+=(const GenericCtlVecQG & rhs) {
  ASSERT(nx_ == rhs.nx_ && ny_ == rhs.ny_ && nl_ == rhs.nl_);
  qg_generic_ctl_vec_self_add_f90(keyCtlVec_, rhs.keyCtlVec_);
  return *this;
}
// -----------------------------------------------------------------------------
GenericCtlVecQG & GenericCtlVecQG::operator-=(const GenericCtlVecQG & rhs) {
  ASSERT(nx_ == rhs.nx_ && ny_ == rhs.ny_ && nl_ == rhs.nl_);
  qg_generic_ctl_vec_self_sub_f90(keyCtlVec_, rhs.keyCtlVec_);
  return *this;
}
// -----------------------------------------------------------------------------
GenericCtlVecQG & GenericCtlVecQG::operator*=(const double & zz) {
  qg_generic_ctl_vec_self_mul_f90(keyCtlVec_, zz);
  return *this;
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::zero() {
  qg_generic_ctl_vec_zero_f90(keyCtlVec_);
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::axpy(const double & zz, const GenericCtlVecQG & rhs) {
  ASSERT(nx_ == rhs.nx_ && ny_ == rhs.ny_ && nl_ == rhs.nl_);
  qg_generic_ctl_vec_axpy_f90(keyCtlVec_, zz, rhs.keyCtlVec_);
}
// -----------------------------------------------------------------------------
double GenericCtlVecQG::dot_product_with(const GenericCtlVecQG & dx) const {
  double zz;
  ASSERT(nx_ == dx.nx_ && ny_ == dx.ny_ && nl_ == dx.nl_);
  qg_generic_ctl_vec_dot_prod_f90(keyCtlVec_, dx.keyCtlVec_, zz);
  return zz;
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::schur_product_with(const GenericCtlVecQG & dx) {
  ASSERT(nx_ == dx.nx_ && ny_ == dx.ny_ && nl_ == dx.nl_);
  qg_generic_ctl_vec_self_schur_f90(keyCtlVec_, dx.keyCtlVec_);
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::random() {
  qg_generic_ctl_vec_random_f90(keyCtlVec_);
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::changeResolution(const GenericCtlVecQG & other) {
  nx_ = other.nx_;
  ny_ = other.ny_;
  nl_ = other.nl_;
  qg_generic_ctl_vec_change_resol_f90(keyCtlVec_, other.keyCtlVec_);
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::getSection(std::vector<double> & section,
                                 const int & iistart,
                                 const int & iiend) const {
   ASSERT(iiend >= iistart);
   ASSERT(section.size() > static_cast<size_t>(iiend - iistart));
   qg_generic_ctl_vec_get_section_f90(keyCtlVec_, &section[0], iistart, iiend);
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::setSection(const std::vector<double> & section,
                                 const int & iistart,
                                 const int & iiend) {
   ASSERT(iiend >= iistart);
   ASSERT(section.size() > static_cast<size_t>(iiend - iistart));
   qg_generic_ctl_vec_set_section_f90(keyCtlVec_, &section[0], iistart, iiend);
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::read(const eckit::Configuration & config) {
  const eckit::Configuration * conf = &config;
  qg_generic_ctl_vec_read_file_f90(keyCtlVec_, &conf);
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::write(const eckit::Configuration & config) const {
  const eckit::Configuration * conf = &config;
  qg_generic_ctl_vec_write_file_f90(keyCtlVec_, &conf);
}
// -----------------------------------------------------------------------------
double GenericCtlVecQG::norm() const {
  double zz = 0.0;
  qg_generic_ctl_vec_rms_f90(keyCtlVec_, zz);
  return zz;
}
// -----------------------------------------------------------------------------
void GenericCtlVecQG::print(std::ostream & os) const {
  os << std::endl << "  Resolution = " << nx_ << ", " << ny_ << ", " << nl_;
  std::vector<double> zstat(3);
  qg_generic_ctl_vec_gpnorm_f90(keyCtlVec_, zstat[0]);
  os << std::endl << "  Min=" << zstat[0]
     << ", Max=" << zstat[1] << ", RMS=" << zstat[2];
}
// -----------------------------------------------------------------------------
}  // namespace qg
