/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef QG_MODEL_LOGBOOKQG_H_
#define QG_MODEL_LOGBOOKQG_H_

#include <string>

#include <boost/noncopyable.hpp>
#include <boost/scoped_ptr.hpp>

#include "eckit/config/LocalConfiguration.h"
#include "util/LogbookHelper.h"

namespace qg {

/// Connects to the OOPS execution Logbook
/*!
 * This class allows to connect to oops::LogbookHelper - an execution tracking mechanism;
 * when the execution Logbook is updated, a callback to LogbookQG update method is invoked,
 * where the Logbook information can be used to update local information; here stored in 
 * eckit::LocalConfiguration, but possibly in Fortran for other models (for fast access).
 */
// -----------------------------------------------------------------------------

class LogbookQG : public util::Printable,
                  private boost::noncopyable {
 public:
  static const std::string classname() {return "qg::LogbookQG";}

/// Initialization, finalization
  static void start();
  static void stop();

/// Destructor
  ~LogbookQG() {}

 private:
  static LogbookQG & getInstance();
  LogbookQG(); 

/// Basic operators
  void update(const eckit::Configuration & conf);
  void connectToLogbook();

/// Print
  void print(std::ostream & os) const;

/// Data
  boost::scoped_ptr<eckit::LocalConfiguration> conf_;
};

// -----------------------------------------------------------------------------

LogbookQG & LogbookQG::getInstance() {
  static LogbookQG theLogbook;
  return theLogbook;
}

// -----------------------------------------------------------------------------

LogbookQG::LogbookQG()
 : conf_(new eckit::LocalConfiguration()) 
{}

// -----------------------------------------------------------------------------

void LogbookQG::start() { 
  oops::Log::info() << "LogbookQG::strat LogbookQG starting" << std::endl; 
  getInstance().connectToLogbook();
}

// -----------------------------------------------------------------------------

void LogbookQG::stop() { 
  oops::Log::info() << "LogbookQG::stop LogbookQG stopping" << std::endl; 
}

// -----------------------------------------------------------------------------

void LogbookQG::update(const eckit::Configuration & conf) {
   *conf_ = conf;
   oops::Log::info() << "LogbookQG::update done: " << *this << std::endl;
}

// -----------------------------------------------------------------------------

void LogbookQG::connectToLogbook() {
  util::LogbookHelper::connectCallback([this]
                                       (const eckit::Configuration & conf)
                                       { return update(conf); });
}

// -----------------------------------------------------------------------------

void LogbookQG::print(std::ostream & os) const {
  os << *conf_ << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace qg

#endif  // QG_MODEL_LOGBOOKQG_H_
