/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ModelBias.h"

#include <iostream>
#include <string>
#include <boost/noncopyable.hpp>

#include "model/GeometryQG.h"
#include "model/ModelBiasIncrement.h"
#include "model/VariablesQG.h"

#include "eckit/config/LocalConfiguration.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"

using oops::Log;

namespace qg {

// -----------------------------------------------------------------------------
/// Constructor, destructor
// -----------------------------------------------------------------------------

ModelBias::ModelBias(const GeometryQG & resol, const eckit::Configuration & conf)
  : active_(conf.has("variables")), fields_(), stash_()
{
  if (active_) {  
    VariablesQG vars(conf);
    util::DateTime vt(conf.getString("date"));
    fields_.reset(new FieldsQG(resol, vars, vt));
    if (conf.has("filename")) {
      fields_->read(conf);
    } else {
      fields_->zero();
    }
     
    Log::trace() << "ModelBias::ModelBias created." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBias::ModelBias(const GeometryQG & resol, const ModelBias & other) 
  : active_(false), fields_(), stash_()
{
  if (other.active_) {
    active_ = true;
    fields_.reset(new FieldsQG(*other.fields_, resol));
    Log::trace() << "ModelBias::ModelBias created by interpolation." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBias::ModelBias(const ModelBias & other, const bool copy) 
  : active_(false), fields_(), stash_()
{
  if (other.active_) {
    active_ = true;
    fields_.reset(new FieldsQG(*other.fields_, copy));
    Log::trace() << "ModelBias::ModelBias copy constructed." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBias::~ModelBias() {
  if (active_) Log::trace() << "ModelBias::ModelBias destructed." << std::endl;
}

// -----------------------------------------------------------------------------
/// Basic operators
// -----------------------------------------------------------------------------

ModelBias & ModelBias::operator+=(const ModelBiasIncrement & dx) {
  if (active_) {
    ASSERT(dx.isActive());
    fields_->add(dx.fields());
  }
  return *this;
}

// -----------------------------------------------------------------------------
/// I/O and diagnostics
// -----------------------------------------------------------------------------

void ModelBias::read(const eckit::Configuration & files) {
  if (active_) fields_->read(files);  
}

// -----------------------------------------------------------------------------

void ModelBias::write(const eckit::Configuration & files) const {
  if (active_) fields_->write(files);
}

// -----------------------------------------------------------------------------

double ModelBias::norm() const {
  double zz = 0.0;
  if (active_) zz = fields_->norm();
  return zz;
}

// -----------------------------------------------------------------------------

void ModelBias::print(std::ostream & os) const {
  if (active_) os << *fields_;
}

// -----------------------------------------------------------------------------

}  // namespace qg
