/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ModelBiasCovariance.h"

#include <ostream>
#include <string>
#include <boost/noncopyable.hpp>

#include "model/GeometryQG.h"
#include "model/ModelBias.h"
#include "model/ModelBiasCtlVec.h"
#include "model/ModelBiasIncrement.h"

#include "eckit/config/LocalConfiguration.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"

using oops::Log;

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------

ModelBiasCovariance::ModelBiasCovariance(const eckit::Configuration & conf, 
                                         const GeometryQG & resol) 
  : active_(false), conf_(conf)
{
  active_ = conf.has("standard_deviation") && conf.has("vertical_correlation") &&
            conf.has("horizontal_length_scale") && conf.has("maximum_condition_number");
  if (active_) {            
    const eckit::Configuration * configc = &conf;
    qg_b_setup_f90(keyBiasCov_, &configc, resol.toFortran());
    Log::trace() << "ModelBiasCovariance created" << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBiasCovariance::~ModelBiasCovariance() {
  if (active_) qg_b_delete_f90(keyBiasCov_);
  Log::trace() << "ModelBiasCovariance destructed" << std::endl;
}

// -----------------------------------------------------------------------------

void ModelBiasCovariance::linearize(const ModelBias &, const GeometryQG & resol) {
  if (active_) geom_.reset(new GeometryQG(resol));
}

// -----------------------------------------------------------------------------

void ModelBiasCovariance::multiply(const ModelBiasIncrement & dxin, 
                                   ModelBiasIncrement & dxout) const {
  if (active_) {                                     
    ASSERT(dxin.isActive());
    ASSERT(dxout.isActive());
    qg_b_mult_f90(keyBiasCov_, dxin.fields().toFortran(),
                               dxout.fields().toFortran());
  }
}

// -----------------------------------------------------------------------------

void ModelBiasCovariance::inverseMultiply(const ModelBiasIncrement & dxin, 
                                          ModelBiasIncrement & dxout) const {
  if (active_) {                                     
    ASSERT(dxin.isActive());
    ASSERT(dxout.isActive());
    qg_b_invmult_f90(keyBiasCov_, dxin.fields().toFortran(),
                                  dxout.fields().toFortran());
  }
}

// -----------------------------------------------------------------------------

void ModelBiasCovariance::multiplySqrt(const ModelBiasCtlVec & dv, 
                                       ModelBiasIncrement & dx) const {
  if (active_) {                                     
    ASSERT(dv.isActive());
    ASSERT(dx.isActive());
    qg_b_mult_sqrt_f90(keyBiasCov_, dv.ctlvec().toFortran(),
                                    dx.fields().toFortran());
  }
}

// -----------------------------------------------------------------------------

void ModelBiasCovariance::multiplySqrtTrans(const ModelBiasIncrement & dx, 
                                            ModelBiasCtlVec & dv) const {
  if (active_) {                                     
    ASSERT(dv.isActive());
    ASSERT(dx.isActive());
    qg_b_mult_sqrt_trans_f90(keyBiasCov_, dx.fields().toFortran(),
                                          dv.ctlvec().toFortran());
  }
}

// -----------------------------------------------------------------------------

void ModelBiasCovariance::randomize(ModelBiasIncrement & dx) const {
  if (active_) {
    ASSERT(dx.isActive());
    qg_b_randomize_f90(keyBiasCov_, dx.fields().toFortran());
  }
}

// -----------------------------------------------------------------------------

const std::vector<int> ModelBiasCovariance::getDim() const {
    std::vector<int> dim(3);
    if (active_) qg_b_get_dim_f90(keyBiasCov_, dim[0]);
    return dim;
}

// -----------------------------------------------------------------------------

void ModelBiasCovariance::print(std::ostream & os) const {
  if (active_) os << "ModelBiasCovariance not implemented";
}

// -----------------------------------------------------------------------------

}  // namespace qg
