/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ModelBiasCtlVec.h"

#include <iostream>
#include <vector>

#include "model/ModelBias.h"
#include "model/ModelBiasCovariance.h"

#include "eckit/config/LocalConfiguration.h"
#include "util/dot_product.h"
#include "util/Printable.h"
#include "util/Logger.h"

using oops::Log;

namespace qg {

// -----------------------------------------------------------------------------
/// Constructor, destructor
// -----------------------------------------------------------------------------

ModelBiasCtlVec::ModelBiasCtlVec(const ModelBiasCovariance & cov)
  : active_(cov.isActive()), ctlvec_()
{
  if (active_) {
    std::vector<int> dim(cov.getDim());
    ctlvec_.reset(new GenericCtlVecQG(dim[0], dim[1], dim[2]));
    ctlvec_->zero();
    Log::trace() << "ModelBiasCtlVec constructed." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBiasCtlVec::ModelBiasCtlVec(const ModelBiasCovariance & cov, 
                                 const ModelBiasCtlVec & other)
  : active_(cov.isActive() && other.isActive()), ctlvec_()
{  
  if (active_) {
    std::vector<int> dim(cov.getDim());
    ctlvec_.reset(new GenericCtlVecQG(dim[0], dim[1], dim[2]));
    *ctlvec_ = *other.ctlvec_;
    Log::trace() << "ModelBiasCtlVec constructed from other." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBiasCtlVec::ModelBiasCtlVec(const ModelBiasCtlVec & other, const bool copy)
  : active_(other.isActive()), ctlvec_()
{
  if (active_) {
    ctlvec_.reset(new GenericCtlVecQG(*other.ctlvec_, copy));
    Log::trace() << "ModelBiasCtlVec copy-created." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBiasCtlVec::~ModelBiasCtlVec() {
  if (active_) {
    ctlvec_.reset();
    Log::trace() << "ModelBiasCtlVec destructed" << std::endl;
  }
}

// -----------------------------------------------------------------------------
/// Basic operators
// -----------------------------------------------------------------------------

void ModelBiasCtlVec::zero() {
  if (active_) ctlvec_->zero();
}

// -----------------------------------------------------------------------------

ModelBiasCtlVec & ModelBiasCtlVec::operator=(const ModelBiasCtlVec & rhs) {
  if (active_) {
    ASSERT(rhs.isActive());
    *ctlvec_ = *rhs.ctlvec_;
  }
  return *this;
}

// -----------------------------------------------------------------------------

ModelBiasCtlVec & ModelBiasCtlVec::operator+=(const ModelBiasCtlVec & dx) {
  if (active_) {
    ASSERT(dx.isActive());
    *ctlvec_ += *dx.ctlvec_;
  }
  return *this;
}

// -----------------------------------------------------------------------------

ModelBiasCtlVec & ModelBiasCtlVec::operator-=(const ModelBiasCtlVec & dx) {
  if (active_) {
    ASSERT(dx.isActive());
    *ctlvec_ -= *dx.ctlvec_;
  }
  return *this;
}

// -----------------------------------------------------------------------------

ModelBiasCtlVec & ModelBiasCtlVec::operator*=(const double & zz) {
  if (active_) *ctlvec_ *= zz;
  return *this;
}

// -----------------------------------------------------------------------------

void ModelBiasCtlVec::axpy(const double & zz, const ModelBiasCtlVec & dx) {
  if (active_) {
    ASSERT(dx.isActive());
    ctlvec_->axpy(zz, *dx.ctlvec_);
  }
}

// -----------------------------------------------------------------------------

double ModelBiasCtlVec::dot_product_with(const ModelBiasCtlVec & other) const {
  double zz = 0.0;
  if (active_) zz = dot_product(*ctlvec_, *other.ctlvec_);
  return zz;
}

// -----------------------------------------------------------------------------
/// I/O and diagnostics
// -----------------------------------------------------------------------------

void ModelBiasCtlVec::read(const eckit::Configuration & files) {
  if (active_) ctlvec_->read(files);
}

// -----------------------------------------------------------------------------

void ModelBiasCtlVec::write(const eckit::Configuration & files) const {
  if (active_) ctlvec_->write(files);
}

// -----------------------------------------------------------------------------

double ModelBiasCtlVec::norm() const {
  double zz = 0.0;
  if (active_) zz = ctlvec_->norm();
  return zz;
}

// -----------------------------------------------------------------------------

void ModelBiasCtlVec::print(std::ostream & os) const {
  if (active_) os << *ctlvec_;
}

// -----------------------------------------------------------------------------

}  // namespace qg
