/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ModelBiasIncrement.h"

#include <iostream>

#include "model/GeometryQG.h"
#include "model/ModelBias.h"
#include "model/ModelBiasCovariance.h"
#include "model/VariablesQG.h"

#include "eckit/config/LocalConfiguration.h"
#include "util/dot_product.h"
#include "util/Printable.h"
#include "util/Logger.h"

using oops::Log;

namespace qg {

// -----------------------------------------------------------------------------
/// Constructor, destructor
// -----------------------------------------------------------------------------

ModelBiasIncrement::ModelBiasIncrement(const GeometryQG & resol, const eckit::Configuration & conf) 
  : active_(conf.has("variables")), fields_(), stash_()
{
  if (active_) {
    VariablesQG vars(conf);
    util::DateTime vt(conf.getString("date"));
    fields_.reset(new FieldsQG(resol, vars, vt));
    fields_->zero();
    Log::trace() << "ModelBiasIncrement constructed." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBiasIncrement::ModelBiasIncrement(const ModelBiasIncrement & other, const bool copy)
  : active_(false), fields_(), stash_()
{
  if (other.isActive()) {
    active_ = true;
    fields_.reset(new FieldsQG(*other.fields_, copy));
    Log::trace() << "ModelBiasIncrement::ModelBiasIncrement copy constructed." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBiasIncrement::ModelBiasIncrement(const ModelBiasIncrement & other, const eckit::Configuration &)
  : active_(false), fields_(), stash_()
{
  if (other.isActive()) {
    active_ = true;
    fields_.reset(new FieldsQG(*other.fields_));
    Log::trace() << "ModelBiasIncrement::ModelBiasIncrement copy constructed." << std::endl;
  }
}

// -----------------------------------------------------------------------------

ModelBiasIncrement::~ModelBiasIncrement() {
  if (active_) Log::trace() << "ModelBiasIncrement::ModelBiasIncrement destructed." << std::endl;
}

// -----------------------------------------------------------------------------
/// Basic operators
// -----------------------------------------------------------------------------

void ModelBiasIncrement::diff(const ModelBias & x1, const ModelBias & x2) {
  if (active_) {
    ASSERT(x1.isActive());
    ASSERT(x2.isActive());
    fields_->diff(x1.fields(), x2.fields());
  }
}

// -----------------------------------------------------------------------------

void ModelBiasIncrement::zero() {
  if (active_) fields_->zero();
}

// -----------------------------------------------------------------------------

ModelBiasIncrement & ModelBiasIncrement::operator=(const ModelBiasIncrement & rhs) {
  if (active_) {
    ASSERT(rhs.isActive());
    *fields_ = *rhs.fields_;
  }
  return *this;
}

// -----------------------------------------------------------------------------

ModelBiasIncrement & ModelBiasIncrement::operator+=(const ModelBiasIncrement & rhs) {
  if (active_) {
    ASSERT(rhs.isActive());
    *fields_ += *rhs.fields_;
  }
  return *this;
}

// -----------------------------------------------------------------------------

ModelBiasIncrement & ModelBiasIncrement::operator-=(const ModelBiasIncrement & rhs) {
  if (active_) {
    ASSERT(rhs.isActive());
    *fields_ -= *rhs.fields_;
  }
  return *this;
}

// -----------------------------------------------------------------------------

ModelBiasIncrement & ModelBiasIncrement::operator*=(const double & zz) {
  if (active_) *fields_ *= zz;
  return *this;
}

// -----------------------------------------------------------------------------

void ModelBiasIncrement::axpy(const double zz, const ModelBiasIncrement & dx) {
  if (active_) {
    ASSERT(dx.isActive());
    fields_->axpy(zz, *dx.fields_);
  }
}

// -----------------------------------------------------------------------------

double ModelBiasIncrement::dot_product_with(const ModelBiasIncrement & other) const {
  double zz = 0.0;
  if (active_) {
    ASSERT(other.isActive());
    zz = dot_product(*fields_, *other.fields_);
  }
  return zz;
}

// -----------------------------------------------------------------------------
/// I/O and diagnostics
// -----------------------------------------------------------------------------

void ModelBiasIncrement::read(const eckit::Configuration & files) {
  if (active_) fields_->read(files);
}

// -----------------------------------------------------------------------------

void ModelBiasIncrement::write(const eckit::Configuration & files) const {
  if (active_) fields_->write(files);
}

// -----------------------------------------------------------------------------

double ModelBiasIncrement::norm() const {
  double zz = 0.0;
  if (active_) zz = fields_->norm();
  return zz;
}

// -----------------------------------------------------------------------------

void ModelBiasIncrement::print(std::ostream & os) const {
  if (active_) os << *fields_;
}

// -----------------------------------------------------------------------------

}  // namespace qg
