/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ObsBias.h"

#include <cmath>
#include <iostream>
#include <sstream>
#include <string>

#include "model/ObsBiasIncrement.h"

#include "eckit/filesystem/LocalPathName.h"
#include "eckit/filesystem/PathName.h"
#include "eckit/exception/Exceptions.h"
#include "eckit/config/Configuration.h"
#include "oops/interface/ObsAuxControlBase.h"
#include "util/formats.h"
#include "util/Logger.h"

using oops::Log;

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------
static oops::ObsAuxControlMaker<ObsBias> makesObsBias_("ObsBiasQG");
// -----------------------------------------------------------------------------

ObsBias::ObsBias(const eckit::Configuration & conf)
  : active_(false), bias_(ntypes, 0.0), dummyKey_(0)
{
  Log::info() << "ObsBias: conf = " << conf << std::endl;
  active_ = conf.has("stream") || conf.has("uwind") || conf.has("vwind") || conf.has("wspeed");
  if (active_) {
    if (conf.has("stream")) bias_[0] = conf.getDouble("stream");
    if (conf.has("uwind"))  bias_[1] = conf.getDouble("uwind");
    if (conf.has("vwind"))  bias_[2] = conf.getDouble("vwind");
    if (conf.has("wspeed")) bias_[3] = conf.getDouble("wspeed");
    std::string strn = "";
    for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
      if (jj > 0) strn += ", ";
      std::ostringstream strs;
      strs << bias_[jj];
      strn += strs.str();
    }
    Log::info() << "ObsBias::ObsBias created, bias = " << strn << std::endl;
  }
}
// -----------------------------------------------------------------------------
ObsBias::ObsBias(const ObsBias & other, const bool copy)
  : active_(other.active_), bias_(ntypes, 0.0), dummyKey_(other.dummyKey_)
{
  if (active_ && copy) {
    for (unsigned int jj = 0; jj < ntypes; ++jj) bias_[jj] = other.bias_[jj];
  }
}
// -----------------------------------------------------------------------------
ObsBias* ObsBias::clone(const bool copy) const {
  return new ObsBias(*this, copy);
}
// -----------------------------------------------------------------------------
oops::ObsAuxControlBase & ObsBias::operator+=(const oops::ObsAuxIncrementBase & dx) {
  if (active_) {
    const ObsBiasIncrement * pdx = dynamic_cast<const ObsBiasIncrement*>(&dx);
    ASSERT(pdx != nullptr);
    for (unsigned int jj = 0; jj < ntypes; ++jj) bias_[jj] += (*pdx)[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
void ObsBias::reset(const std::string & stateid,
                    const eckit::Configuration & conf) {
  // No need to reset for ObsBias
}
// -----------------------------------------------------------------------------
void ObsBias::read(const eckit::Configuration & conf) {
  std::string filename = conf.getString("filename"); 
  Log::info() << " ObsBias::read opening" << filename << std::endl;
  eckit::PathName path = eckit::LocalPathName::cwd() + "/" + filename;
  std::ifstream ifs(path.localPath());
  if (!ifs)
    throw eckit::CantOpenFile(path.localPath(), Here());
  for (size_t ii = 0; ii < ntypes; ++ii) {
    if (!ifs.bad()) {
      std::string bias;
      ifs >> bias;
      if (ifs.good()) {
        bias_[ii] = std::stod(bias);
      } else if (ifs.eof()) {
        throw eckit::ReadError("Failed to read entire ObsBias", Here()); 
      } else if (ifs.fail()) {
        throw eckit::ReadError("Failed to read ObsBias", Here()); 
      }
    }
  }
  Log::info() << " DEBUG ObsBias::read " << util::full_precision(bias_) << std::endl;
}
// -----------------------------------------------------------------------------
void ObsBias::write(const eckit::Configuration & conf) const {
  std::string filename = conf.getString("filename"); 
  Log::info() << " ObsBias::write opening" << filename << std::endl;
  eckit::PathName path = eckit::LocalPathName::cwd() + "/" + filename; 
  std::ofstream ofs(path.localPath());
  if (!ofs)
    throw eckit::CantOpenFile(path.localPath(), Here());
  for (size_t ii = 0; ii < ntypes; ++ii) {
    ofs << util::full_precision(bias_[ii]) << std::endl;
    if (ofs.fail()) 
      throw eckit::WriteError("Failed to write ObsBias", Here());
  }
  ofs.close();
  Log::info() << " DEBUG ObsBias::write " << util::full_precision(bias_) << std::endl;
}
// -----------------------------------------------------------------------------
double ObsBias::norm() const {
  double zz = 0.0;
  if (active_) {
    for (unsigned int jj = 0; jj < ntypes; ++jj) zz += bias_[jj]*bias_[jj];
    zz = std::sqrt(zz/3.0);
  }
  return zz;
}
// -----------------------------------------------------------------------------
void ObsBias::print(std::ostream & os) const {
  if (active_) {
    std::string strn = "";
    for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
      if (jj > 0) strn += ", ";
      std::ostringstream strs;
      strs << bias_[jj];
      strn += strs.str();
    }
    os << std::endl << "ObsBias = " << strn;
  }
}
// -----------------------------------------------------------------------------
}  // namespace qg

