/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ObsBiasCtlVec.h"

#include <cmath>
#include <iostream>
#include <sstream>
#include <string>

#include "model/ObsBias.h"

#include "eckit/filesystem/LocalPathName.h"
#include "eckit/filesystem/PathName.h"
#include "eckit/exception/Exceptions.h"
#include "eckit/config/LocalConfiguration.h"
#include "oops/interface/ObsAuxCovarianceBase.h"
#include "util/formats.h"
#include "util/Logger.h"

using oops::Log;

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------
static oops::ObsAuxCtlVecMaker<ObsBiasCtlVec> makeObsBiasCtlVec_("ObsBiasQG");
// -----------------------------------------------------------------------------

ObsBiasCtlVec::ObsBiasCtlVec(const oops::ObsAuxCovarianceBase & covar)
  : active_(ObsBias::ntypes, false), sqrtbias_(ObsBias::ntypes, 0.0),
    dummyKey_(0)
{
  eckit::LocalConfiguration covarConfig = covar.config();
  active_[0] = covarConfig.has("stream");
  active_[1] = covarConfig.has("uwind");
  active_[2] = covarConfig.has("vwind");
  active_[3] = covarConfig.has("wspeed");
  bool on = false;
  std::string strn = "";
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (jj > 0) strn += ", ";
    if (active_[jj]) {
      strn += "on";
      on = true;
    } else {
      strn += "off";
    }
  }
  if (on) {Log::trace() << "ObsBiasCtlVec created : " << strn << std::endl;}
}
// -----------------------------------------------------------------------------
ObsBiasCtlVec::ObsBiasCtlVec(const oops::ObsAuxCovarianceBase & cov,
                             const ObsBiasCtlVec & other)
  : active_(other.active()), sqrtbias_(ObsBias::ntypes, 0.0),
    dummyKey_(other.dummyKey_)
{
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) sqrtbias_[jj] = other[jj];
  this->makePassive();
}
// -----------------------------------------------------------------------------
ObsBiasCtlVec::ObsBiasCtlVec(const ObsBiasCtlVec & other,
                             const bool copy)
  : active_(other.active()), sqrtbias_(ObsBias::ntypes, 0.0),
    dummyKey_(other.dummyKey_)
{
  if (copy) {
    for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) sqrtbias_[jj] = other[jj];
  }
  this->makePassive();
}
// -----------------------------------------------------------------------------
ObsBiasCtlVec* ObsBiasCtlVec::clone(const oops::ObsAuxCovarianceBase & cov) const {
  return new ObsBiasCtlVec(cov, *this);
}
// -----------------------------------------------------------------------------
ObsBiasCtlVec* ObsBiasCtlVec::clone(const bool copy) const {
  return new ObsBiasCtlVec(*this, copy);
}
// -----------------------------------------------------------------------------
void ObsBiasCtlVec::makePassive() {
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (!active_[jj]) sqrtbias_[jj] = 0.0;
  }
}
// -----------------------------------------------------------------------------
void ObsBiasCtlVec::read(const eckit::Configuration & conf) {
  std::string filename = conf.getString("filename");
  Log::info() << " ObsBiasCtlVec::read " << filename << std::endl;
  eckit::PathName path = eckit::LocalPathName::cwd() + "/" + filename;
  std::ifstream ifs(path.localPath());
  if (!ifs)
    throw eckit::CantOpenFile(path.localPath(), Here());
  for (size_t ii = 0; ii < ObsBias::ntypes; ++ii) {
    if (!ifs.bad()) {
      std::string sqrtbias;
      ifs >> sqrtbias;
      if (ifs.good()) {
        sqrtbias_[ii] = std::stod(sqrtbias);
      } else if (ifs.eof()) {
        throw eckit::ReadError("Failed to read entire ObsBias", Here());
      } else if (ifs.fail()) {
        throw eckit::ReadError("Failed to read ObsBias", Here());
      }
    }
  }
  Log::info() << " DEBUG ObsBiasCtlVec::read " << util::full_precision(sqrtbias_) << std::endl;
}
// -----------------------------------------------------------------------------
void ObsBiasCtlVec::write(const eckit::Configuration & conf) const {
  std::string filename = conf.getString("filename");
  Log::info() << " ObsBiasCtlVec::write " << filename << std::endl;
  eckit::PathName path = eckit::LocalPathName::cwd() + "/" + filename;
  std::ofstream ofs(path.localPath());
  if (!ofs)
    throw eckit::CantOpenFile(path.localPath(), Here());
  for (size_t ii = 0; ii < ObsBias::ntypes; ++ii) {
    ofs << util::full_precision(sqrtbias_[ii]) << std::endl;
    if (ofs.fail())
      throw eckit::WriteError("Failed to write ObsBiasCtlVec", Here());
  }
  ofs.close();
  Log::info() << " DEBUG ObsBiasCtlVec::write " << util::full_precision(sqrtbias_) << std::endl;
}
// -----------------------------------------------------------------------------
void ObsBiasCtlVec::zero() {
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) sqrtbias_[jj] = 0.0;
}
// -----------------------------------------------------------------------------
ObsBiasCtlVec & ObsBiasCtlVec::operator=(const oops::ObsAuxCtlVecBase & rhs) {
  const ObsBiasCtlVec * prhs = dynamic_cast<const ObsBiasCtlVec*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) sqrtbias_[jj] = (*prhs)[jj];
  this->makePassive();
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasCtlVec & ObsBiasCtlVec::operator+=(const oops::ObsAuxCtlVecBase & rhs) {
  const ObsBiasCtlVec * prhs = dynamic_cast<const ObsBiasCtlVec*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) sqrtbias_[jj] += (*prhs)[jj];
  this->makePassive();
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasCtlVec & ObsBiasCtlVec::operator-=(const oops::ObsAuxCtlVecBase & rhs) {
  const ObsBiasCtlVec * prhs = dynamic_cast<const ObsBiasCtlVec*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) sqrtbias_[jj] -= (*prhs)[jj];
  this->makePassive();
  return *this;
}
// -----------------------------------------------------------------------------
ObsBiasCtlVec & ObsBiasCtlVec::operator*=(const double & fact) {
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) sqrtbias_[jj] *= fact;
  this->makePassive();
  return *this;
}
// -----------------------------------------------------------------------------
void ObsBiasCtlVec::axpy(const double & fact, const oops::ObsAuxCtlVecBase & rhs) {
  const ObsBiasCtlVec * prhs = dynamic_cast<const ObsBiasCtlVec*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) sqrtbias_[jj] += fact * (*prhs)[jj];
  this->makePassive();
}
// -----------------------------------------------------------------------------
double ObsBiasCtlVec::dot_product_with(const oops::ObsAuxCtlVecBase & rhs) const {
  double zz = 0.0;
  const ObsBiasCtlVec * prhs = dynamic_cast<const ObsBiasCtlVec*>(&rhs);
  ASSERT(prhs != nullptr);
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (active_[jj]) zz += sqrtbias_[jj] * (*prhs)[jj];
  }
  return zz;
}
// -----------------------------------------------------------------------------
double ObsBiasCtlVec::norm() const {
  double zz = 0.0;
  int ii = 0;
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (active_[jj]) {
      zz += sqrtbias_[jj] * sqrtbias_[jj];
      ++ii;
    }
  }
  if (ii>0) zz = std::sqrt(zz/ii);
  return zz;
}
// -----------------------------------------------------------------------------
void ObsBiasCtlVec::print(std::ostream & os) const {
  bool on = false;
  std::string strn = "";
  for (unsigned int jj = 0; jj < ObsBias::ntypes; ++jj) {
    if (jj > 0) strn += ", ";
    if (active_[jj]) {
      on = true;
      std::ostringstream strs;
      strs << sqrtbias_[jj];
      strn += strs.str();
    } else {
      strn += "0.0";
    }
  }
  if (on) os << std::endl << "ObsBiasCtlVec = " << strn;
}
// -----------------------------------------------------------------------------
}  // namespace qg
