/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ObsErrWSpeedQG.h"
#include "model/QgFortran.h"

#include "eckit/config/Configuration.h"
#include "util/Logger.h"

namespace qg {

static oops::ObsErrorMaker<QgTraits, ObsErrWSpeedQG> makerObsErrStreamIFS_("WSpeedObsError");

// =============================================================================

ObsErrWSpeedQG::ObsErrWSpeedQG(const ObsSpaceQG & obsgeom, const eckit::Configuration & config)
  : lvarqc_(false), cleft_(0.0), cright_(0.0),
    obsgeom_(obsgeom), stddev_(), inverseVariance_(), wghtsqrt_()
{
/// Setup R
  stddev_.reset(new ObsVecQG(obsgeom));
  const std::string col = config.getString("obserror");
  stddev_->read(col);

  inverseVariance_.reset(new ObsVecQG(*stddev_));
  *inverseVariance_ *= *inverseVariance_;
  inverseVariance_->invert();

/// Setup W^1/2
  if (config.has("varqc")) {
    lvarqc_ = true;
    cleft_ = config.getDouble("varqc.cleft");
    cright_ = config.getDouble("varqc.cright");
    wghtsqrt_.reset(new ObsVecQG(obsgeom));
  }

  oops::Log::trace() << "ObsErrWSpeedQG:ObsErrWSpeedQG constructed" << std::endl;
}

// -----------------------------------------------------------------------------

void ObsErrWSpeedQG::linearize(const ObsVecQG & dy)
{
  oops::Log::trace() << "ObsErrWSpeedQG:linearize starting" << std::endl;

  oops::Log::trace() << "ObsErrWSpeedQG:linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

void ObsErrWSpeedQG::setupWeights(const ObsVecQG & dy)  
{
  oops::Log::trace() << "ObsErrWSpeedQG:setupWeights starting" << std::endl;
  if (lvarqc_) {
    qg_obserr_setup_wghts_sqrt_f90(wghtsqrt_->toFortran(), stddev_->toFortran(), 
                                   dy.toFortran(), cleft_, cright_);   
    oops::Log::info() << "ObsErrWSpeedQG:setupWeights done" << std::endl;
    oops::Log::info() << "  Weights " << *wghtsqrt_ << std::endl;
  }
  oops::Log::trace() << "ObsErrWSpeedQG:setupWeights done" << std::endl;
}

// -----------------------------------------------------------------------------

ObsErrWSpeedQG::~ObsErrWSpeedQG() {
  oops::Log::trace() << "ObsErrWSpeedQG:~ObsErrWSpeedQG destructed" << std::endl;
}

// -----------------------------------------------------------------------------

ObsVecQG * ObsErrWSpeedQG::multiply(const ObsVecQG & dy) const {
  ObsVecQG * res = new ObsVecQG(dy);
  *res /= *inverseVariance_;
  return res;
}

// -----------------------------------------------------------------------------

ObsVecQG * ObsErrWSpeedQG::inverseMultiply(const ObsVecQG & dy) const {
  ObsVecQG * res = new ObsVecQG(dy);
  *res *= *inverseVariance_;
  return res;
}

// -----------------------------------------------------------------------------

ObsVecQG * ObsErrWSpeedQG::multiplyWghtSqrt(const ObsVecQG & dy) const {
  ObsVecQG * res = new ObsVecQG(dy);
  if (lvarqc_) *res *= *wghtsqrt_;
  return res;
}

// -----------------------------------------------------------------------------

void ObsErrWSpeedQG::randomize(ObsVecQG & dy) const {
  dy.random();
  dy *= *stddev_;
}

// -----------------------------------------------------------------------------

void ObsErrWSpeedQG::print(std::ostream & os) const {
  os << "ObsErrWSpeedQG::print not implemeted yet";
}

// -----------------------------------------------------------------------------


}  // namespace qg
