/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#include "model/ObsWindTLAD.h"

#include "eckit/config/Configuration.h"
#include "model/GomQG.h"
#include "model/ObsBias.h"
#include "model/ObsBiasIncrement.h"
#include "model/ObsSpaceQG.h"
#include "model/ObsVecQG.h"
#include "model/QgFortran.h"
#include "model/VariablesQG.h"

#include "oops/interface/ObsAuxControlBase.h"
#include "util/Logger.h"

// -----------------------------------------------------------------------------
namespace qg {
// -----------------------------------------------------------------------------
static oops::LinearObsOpMaker<QgTraits, ObsWindTLAD> makerWindTL_("Wind");
// -----------------------------------------------------------------------------

ObsWindTLAD::ObsWindTLAD(const ObsSpaceQG &, const eckit::Configuration & config)
  : keyOperWind_(0), varin_()
{
  const eckit::Configuration * configc = &config;
  qg_wind_setup_f90(keyOperWind_, &configc);
  int keyVarin;
  qg_obsoper_inputs_f90(keyOperWind_, keyVarin);
  varin_.reset(new VariablesQG(keyVarin));
  oops::Log::trace() << "ObsWindTLAD created" << std::endl;
}

// -----------------------------------------------------------------------------

ObsWindTLAD::~ObsWindTLAD() {
  oops::Log::trace() << "ObsWindTLAD destructed" << std::endl;
}

// -----------------------------------------------------------------------------

void ObsWindTLAD::setTrajectory(const GomQG &, 
                                const boost::ptr_map<std::string, oops::ObsAuxControlBase> &) 
{}

// -----------------------------------------------------------------------------

void ObsWindTLAD::obsEquivTL(const GomQG & gom, ObsVecQG & ovec,
                             const boost::ptr_map<std::string, oops::ObsAuxIncrementBase> & bias) const {
  typedef typename boost::ptr_map<std::string, oops::ObsAuxIncrementBase>::const_iterator icst_;
  icst_ it = bias.find("ObsBiasQG");
  if (it == bias.end()) {
    std::vector<double> dummyBias(2, 0.0);
    qg_wind_equiv_tl_f90(gom.toFortran(), ovec.toFortran(), dummyBias[0]);
  } else {
    const ObsBiasIncrement * pbias = dynamic_cast<const ObsBiasIncrement*>(it->second);
    ASSERT(pbias != nullptr);
    qg_wind_equiv_tl_f90(gom.toFortran(), ovec.toFortran(), pbias->wind());
  }
}

// -----------------------------------------------------------------------------

void ObsWindTLAD::obsEquivAD(GomQG & gom, const ObsVecQG & ovec,
                             boost::ptr_map<std::string, oops::ObsAuxIncrementBase> & bias) const {
  typedef typename boost::ptr_map<std::string, oops::ObsAuxIncrementBase>::iterator iter_;
  iter_ it = bias.find("ObsBiasQG");
  if (it == bias.end()) {
    std::vector<double> dummyBias(2, 0.0);
    qg_wind_equiv_ad_f90(gom.toFortran(), ovec.toFortran(), dummyBias[0]);
  } else {
    ObsBiasIncrement * pbias = dynamic_cast<ObsBiasIncrement*>(it->second);
    ASSERT(pbias != nullptr);
    qg_wind_equiv_ad_f90(gom.toFortran(), ovec.toFortran(), pbias->wind());
  }
}

// -----------------------------------------------------------------------------

void ObsWindTLAD::obsBiasEquivTL(const GomQG & gom, ObsVecQG & obias,
                                 const boost::ptr_map<std::string, oops::ObsAuxIncrementBase> & bias) const {
  typedef typename boost::ptr_map<std::string, oops::ObsAuxIncrementBase>::const_iterator icst_;
  icst_ it = bias.find("ObsBiasQG");
  if (it == bias.end()) {
    std::vector<double> dummyBias(2, 0.0);
    qg_wind_bias_equiv_tl_f90(gom.toFortran(), obias.toFortran(), dummyBias[0]);
  } else {
    const ObsBiasIncrement * pbias = dynamic_cast<const ObsBiasIncrement*>(it->second);
    ASSERT(pbias != nullptr);
    qg_wind_bias_equiv_tl_f90(gom.toFortran(), obias.toFortran(), pbias->wind());
  }
}

// -----------------------------------------------------------------------------

void ObsWindTLAD::obsBiasEquivAD(const GomQG & gom, const ObsVecQG & obias,
                                 boost::ptr_map<std::string, oops::ObsAuxIncrementBase> & bias) const {
  typedef typename boost::ptr_map<std::string, oops::ObsAuxIncrementBase>::iterator iter_;
  iter_ it = bias.find("ObsBiasQG");
  if (it == bias.end()) {
    std::vector<double> dummyBias(2, 0.0);
    qg_wind_bias_equiv_ad_f90(gom.toFortran(), obias.toFortran(), dummyBias[0]);
  } else {
    ObsBiasIncrement * pbias = dynamic_cast<ObsBiasIncrement*>(it->second);
    ASSERT(pbias != nullptr);
    qg_wind_bias_equiv_ad_f90(gom.toFortran(), obias.toFortran(), pbias->wind());
  }
}

// -----------------------------------------------------------------------------

void ObsWindTLAD::print(std::ostream & os) const {
  os << "ObsStreamTLAD::print not implemented" << std::endl;
}

// -----------------------------------------------------------------------------


}  // namespace qg
