! (C) Copyright 2009-2016 ECMWF.
! 
! This software is licensed under the terms of the Apache Licence Version 2.0
! which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
! In applying this licence, ECMWF does not waive the privileges and immunities 
! granted to it by virtue of its status as an intergovernmental organisation nor
! does it submit to any jurisdiction.

!> Handle control vector for the QG model

module qg_generic_ctl_vec_mod

use config_mod
use qg_geom_mod
use qg_locs_mod
use qg_vars_mod
use qg_goms_mod
use kinds

implicit none
private

public :: qg_generic_ctl_vec, &
        & create, delete, zerosctlvec, random, copy, &
        & self_add, self_schur, self_sub, self_mul, axpy, &
        & dot_prod, &
        & read_file, write_file, gpnorm, fldrms, &
        & change_resol, &
        & get_section, set_section
public :: qg_generic_ctl_vec_registry

! ------------------------------------------------------------------------------

!> Fortran derived type to hold QG fields
type :: qg_generic_ctl_vec
  integer :: nx                     !< Zonal grid dimension
  integer :: ny                     !< Meridional grid dimension
  integer :: nl                     !< Number of levels
  real(kind=kind_real), pointer :: ctlvec(:,:,:)    !< Control Vector
end type qg_generic_ctl_vec

#define LISTED_TYPE qg_generic_ctl_vec

!> Linked list interface - defines registry_t type
#include "linkedList_i.f"

!> Global registry
type(registry_t) :: qg_generic_ctl_vec_registry

! ------------------------------------------------------------------------------
contains
! ------------------------------------------------------------------------------
!> Linked list implementation
#include "linkedList_c.f"

! ------------------------------------------------------------------------------

subroutine create(self, nx, ny, nl)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self
integer, intent(in) :: nx
integer, intent(in) :: ny
integer, intent(in) :: nl

self%nx = nx
self%ny = ny
self%nl = nl

allocate(self%ctlvec(self%nx,self%ny,self%nl))
self%ctlvec(:,:,:)=0.0_kind_real

call check(self)

end subroutine create

! ------------------------------------------------------------------------------

subroutine delete(self)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self

call check(self)

if (associated(self%ctlvec)) deallocate(self%ctlvec)

end subroutine delete

! ------------------------------------------------------------------------------

subroutine zerosctlvec(self)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self

call check(self)

self%ctlvec(:,:,:) = 0.0_kind_real

end subroutine zerosctlvec

! ------------------------------------------------------------------------------

subroutine random(self)
use random_vectors_mod
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self

call check(self)

call random_vector(self%ctlvec(:,:,:))

end subroutine random

! ------------------------------------------------------------------------------

subroutine copy(self,rhs)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self
type(qg_generic_ctl_vec), intent(in)    :: rhs

call check_resolution(self, rhs)

self%ctlvec(:,:,:) = rhs%ctlvec(:,:,:)

return
end subroutine copy

! ------------------------------------------------------------------------------

subroutine self_add(self,rhs)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self
type(qg_generic_ctl_vec), intent(in)    :: rhs

call check_resolution(self, rhs)

self%ctlvec(:,:,:) = self%ctlvec(:,:,:) + rhs%ctlvec(:,:,:)

return
end subroutine self_add

! ------------------------------------------------------------------------------

subroutine self_schur(self,rhs)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self
type(qg_generic_ctl_vec), intent(in)    :: rhs

call check_resolution(self, rhs)

self%ctlvec(:,:,:) = self%ctlvec(:,:,:) * rhs%ctlvec(:,:,:)

return
end subroutine self_schur

! ------------------------------------------------------------------------------

subroutine self_sub(self,rhs)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self
type(qg_generic_ctl_vec), intent(in)    :: rhs

call check_resolution(self, rhs)

self%ctlvec(:,:,:) = self%ctlvec(:,:,:) - rhs%ctlvec(:,:,:)

return
end subroutine self_sub

! ------------------------------------------------------------------------------

subroutine self_mul(self,zz)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self
real(kind=kind_real), intent(in) :: zz

call check(self)

self%ctlvec(:,:,:) = zz * self%ctlvec(:,:,:)

return
end subroutine self_mul

! ------------------------------------------------------------------------------

subroutine axpy(self,zz,rhs)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: self
real(kind=kind_real), intent(in) :: zz
type(qg_generic_ctl_vec), intent(in)    :: rhs

call check_resolution(self, rhs)

self%ctlvec(:,:,:) = self%ctlvec(:,:,:) + zz * rhs%ctlvec(:,:,:)

return
end subroutine axpy

! ------------------------------------------------------------------------------

subroutine dot_prod(fld1,fld2,zprod)
implicit none
type(qg_generic_ctl_vec), intent(in) :: fld1, fld2
real(kind=kind_real), intent(inout) :: zprod

integer :: jx,jy,jz
real(kind=kind_real) :: zz

call check_resolution(fld1, fld2)

zz=0.0_kind_real
do jy=1,fld1%ny
do jx=1,fld1%nx
do jz=1,fld1%nl
  zz = zz + fld1%ctlvec(jx,jy,jz) * fld2%ctlvec(jx,jy,jz)
enddo
enddo
enddo

zprod = zz

return
end subroutine dot_prod

! ------------------------------------------------------------------------------

subroutine change_resol(fld,rhs)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: fld
type(qg_generic_ctl_vec), intent(in)    :: rhs
real(kind=kind_real), allocatable :: ztmp(:,:)
real(kind=kind_real) :: dy1, dy2, ya, yb, dx1, dx2, xa, xb
integer :: jx, jy, jf, iy, ia, ib

call check(fld)
call check(rhs)

if (fld%nx==rhs%nx .and. fld%ny==rhs%ny) then
  call copy(fld, rhs)
else
  call abor1_ftn("qg_generic_ctl_vec:field_resol: untested code")

allocate(ztmp(rhs%nx,rhs%nl))
dy1=1.0_kind_real/real(fld%ny,kind_real)
dy2=1.0_kind_real/real(rhs%ny,kind_real)
dx1=1.0_kind_real/real(fld%nx,kind_real)
dx2=1.0_kind_real/real(rhs%nx,kind_real)
do jy=1,fld%ny
! North-south interpolation
  if (jy==1 .or. jy==fld%ny) then
    if (jy==1) then
      iy=1
    else
      iy=rhs%ny
    endif
    do jx=1,rhs%nx
      do jf=1,rhs%nl
        ztmp(jx,jf)=rhs%ctlvec(jx,iy,jf)
      enddo
    enddo
  else
    call lin_weights(jy,dy1,dy2,ia,ib,ya,yb)
    if (ib>rhs%ny) call abor1_ftn("qg_generic_ctl_vec:field_resol: ib too large")
    do jx=1,rhs%nx
      do jf=1,rhs%nl
        ztmp(jx,jf)=ya*rhs%ctlvec(jx,ia,jf)+yb*rhs%ctlvec(jx,ib,jf)
      enddo
    enddo
  endif
! East-west interpolation (periodic)
  do jx=1,fld%nx
    call lin_weights(jx,dx1,dx2,ia,ib,xa,xb)
    if (ib>rhs%nx) ib=ib-rhs%nx
    do jf=1,fld%nl
      fld%ctlvec(jx,jy,jf)=xa*ztmp(jx,jf)+xb*ztmp(jx+1,jf)
    enddo
  enddo
enddo
deallocate(ztmp)

endif

return
end subroutine change_resol

! ------------------------------------------------------------------------------

subroutine get_section(fld,section,iistart,iiend)
implicit none
type(qg_generic_ctl_vec), intent(in) :: fld
real(kind=kind_real),     intent(out):: section(iiend-iistart+1)
integer,                  intent(in) :: iistart,iiend

integer :: jx, jy, jz
integer :: gcounter, lcounter

call check(fld)

gcounter=1
lcounter=1

loop_jz: do jz=1,fld%nl
loop_jy: do jy=1,fld%ny
loop_jx: do jx=1,fld%nx
  if (gcounter >= iistart ) then
    section(lcounter) = fld%ctlvec(jx,jy,jz)
    lcounter = lcounter + 1
  endif
  if (gcounter == iiend) exit loop_jz
  gcounter = gcounter + 1
enddo loop_jx
enddo loop_jy
enddo loop_jz

return
end subroutine get_section

! ------------------------------------------------------------------------------

subroutine set_section(fld,section,iistart,iiend)
implicit none
type(qg_generic_ctl_vec), intent(inout) :: fld
real(kind=kind_real),     intent(in)    :: section(iiend-iistart+1)
integer,                  intent(in)    :: iistart,iiend

integer :: jx, jy, jz
integer :: gcounter, lcounter

call check(fld)

gcounter=1
lcounter=1

loop_jz: do jz=1,fld%nl
loop_jy: do jy=1,fld%ny
loop_jx: do jx=1,fld%nx
  if (gcounter >= iistart ) then
    fld%ctlvec(jx,jy,jz) = section(lcounter)
    lcounter = lcounter + 1
  endif
  if (gcounter == iiend) exit loop_jz
  gcounter = gcounter + 1
enddo loop_jx
enddo loop_jy
enddo loop_jz

return
end subroutine set_section
! ------------------------------------------------------------------------------

subroutine read_file(fld, c_conf)
! Needs more interface clean-up here...
use iso_c_binding
use fckit_log_module, only : fckit_log

implicit none
type(qg_generic_ctl_vec), intent(inout) :: fld      !< Fields
type(c_ptr), intent(in)                 :: c_conf   !< Configuration

integer, parameter :: iunit=10
integer, parameter :: max_string_length=800 ! Yuk!
character(len=max_string_length+50) :: record
character(len=max_string_length) :: filename
character(len=20) :: fmtn
character(len=4)  :: cnx
character(len=11) :: fmt1='(X,ES24.16)'
character(len=1024) :: buf
integer :: ix, iy, il, jx, jy, jl 

call zerosctlvec(fld)

filename = config_get_string(c_conf,len(filename),"filename")
WRITE(buf,*) 'qg_generic_ctl_vec:read_file: opening '//filename
call fckit_log%info(buf)
open(unit=iunit, file=trim(filename), form='formatted', action='read')

read(iunit,*) ix, iy, il
if (ix /= fld%nx .or. iy /= fld%ny .or. il /= fld%nl) then
  write (record,*) "qg_generic_ctl_vec:read_file: ", &
                 & "input fields have wrong dimensions: ",ix,iy,il
  call fckit_log%error(record)
  write (record,*) "qg_generic_ctl_vec:read_file: expected: ",fld%nx,fld%ny,fld%nl
  call fckit_log%error(record)
  call abor1_ftn("qg_generic_ctl_vec:read_file: input fields have wrong dimensions")
endif

if (fld%nx>9999)  call abor1_ftn("Format too small")
write(cnx,'(I4)')fld%nx
fmtn='('//trim(cnx)//fmt1//')'

do jl=1,il
  do jy=1,fld%ny
    read(iunit,fmtn) (fld%ctlvec(jx,jy,jl), jx=1,fld%nx)
  enddo
enddo

close(iunit)

call check(fld)

return
end subroutine read_file

! ------------------------------------------------------------------------------

subroutine write_file(fld, c_conf)
use iso_c_binding
use fckit_log_module, only : fckit_log

implicit none
type(qg_generic_ctl_vec), intent(in) :: fld    !< Fields
type(c_ptr), intent(in)    :: c_conf !< Configuration

integer, parameter :: iunit=11
integer, parameter :: max_string_length=800 ! Yuk!
character(len=max_string_length+50) :: record
character(len=max_string_length) :: filename
character(len=20) :: fmtn
character(len=4)  :: cnx
character(len=11) :: fmt1='(X,ES24.16)'
character(len=1024):: buf
integer :: jx, jy, jl

call check(fld)

filename = genfilename(c_conf,max_string_length)
WRITE(buf,*) 'qg_generic_ctl_vec:write_file: writing '//filename
call fckit_log%info(buf)
open(unit=iunit, file=trim(filename), form='formatted', action='write')

write(iunit,*) fld%nx, fld%ny, fld%nl
if (fld%nx>9999)  call abor1_ftn("Format too small")
write(cnx,'(I4)')fld%nx
fmtn='('//trim(cnx)//fmt1//')'

do jl=1,fld%nl
  do jy=1,fld%ny
    write(iunit,fmtn) (fld%ctlvec(jx,jy,jl), jx=1,fld%nx)
  enddo
enddo

close(iunit)

return
end subroutine write_file

! ------------------------------------------------------------------------------

function genfilename (c_conf,length)
use iso_c_binding
type(c_ptr), intent(in):: c_conf  !< Configuration
integer, intent(in)    :: length
character(len=length)  :: genfilename

character(len=length)  :: fle
integer lenfn

! here we should query the length and then allocate "string".
! But Fortran 90 does not allow variable-length allocatable strings.
! config_get_string checks the string length and aborts if too short.
fle = config_get_string(c_conf,len(fle),"filename")

lenfn = LEN_TRIM(fle)
genfilename = TRIM(fle)

if (lenfn>length) &
  & call abor1_ftn("qg_generic_ctl_vec:genfilename: filename too long")

end function genfilename

! ------------------------------------------------------------------------------

subroutine gpnorm(fld, pstat)
implicit none
type(qg_generic_ctl_vec), intent(in) :: fld
real(kind=kind_real), intent(inout) :: pstat(3)

call check(fld)

pstat(1)=minval(fld%ctlvec(:,:,:))
pstat(2)=maxval(fld%ctlvec(:,:,:))
pstat(3)=sqrt(sum(fld%ctlvec(:,:,:)**2) &
          & /real(fld%nl*fld%nx*fld%ny,kind_real))

return
end subroutine gpnorm

! ------------------------------------------------------------------------------

subroutine fldrms(fld, prms)
implicit none
type(qg_generic_ctl_vec), intent(in) :: fld
real(kind=kind_real), intent(out) :: prms
integer :: jf,jy,jx,ii
real(kind=kind_real) :: zz

call check(fld)

zz = 0.0

do jf=1,fld%nl
  do jy=1,fld%ny
    do jx=1,fld%nx
      zz = zz + fld%ctlvec(jx,jy,jf)*fld%ctlvec(jx,jy,jf)
    enddo
  enddo
enddo

ii = fld%nl*fld%ny*fld%nx

prms = sqrt(zz/real(ii,kind_real))

end subroutine fldrms

! ------------------------------------------------------------------------------

subroutine lin_weights(kk,delta1,delta2,k1,k2,w1,w2)
implicit none
integer, intent(in)  :: kk
real(kind=kind_real), intent(in)     :: delta1,delta2
integer, intent(out) :: k1,k2
real(kind=kind_real), intent(out)    :: w1,w2

integer :: ii
real(kind=kind_real) :: zz

zz=real(kk-1,kind_real)*delta1
zz=zz/delta2
ii=int(zz)
w1=zz-real(ii,kind_real)
w2=1.0_kind_real-w1
k1=ii+1
k2=ii+2

return
end subroutine lin_weights

! ------------------------------------------------------------------------------

subroutine check_resolution(x1, x2)

implicit none
type(qg_generic_ctl_vec), intent(in) :: x1, x2

if (x1%nx /= x2%nx .or.  x1%ny /= x2%ny .or.  x1%nl /= x2%nl) then
  call abor1_ftn ("qg_generic_ctl_vec: resolution error")
endif
call check(x1)
call check(x2)

end subroutine check_resolution

! ------------------------------------------------------------------------------

subroutine check(self)
implicit none
type(qg_generic_ctl_vec), intent(in) :: self
logical :: bad

bad = .not.associated(self%ctlvec)

bad = bad .or. (size(self%ctlvec, 1) /= self%nx)
bad = bad .or. (size(self%ctlvec, 2) /= self%ny)
bad = bad .or. (size(self%ctlvec, 3) /= self%nl)

!allocate(self%fldnames(self%nf))

if (bad) then
  write(0,*)'nx, ny, nl = ',self%nx,self%ny,self%nl
  if (associated(self%ctlvec)) write(0,*)'shape(ctlvec) = ',shape(self%ctlvec)
  call abor1_ftn ("qg_generic_ctl_vec: field not consistent")
endif

end subroutine check

! ------------------------------------------------------------------------------

end module qg_generic_ctl_vec_mod
