/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_CONTROLOBSVECTOR_H_
#define OOPS_INTERFACE_CONTROLOBSVECTOR_H_

#include <math.h>
#include <ostream>
#include <string>

#include <boost/scoped_ptr.hpp>

#include "oops/interface/ObservationSpace.h"
#include "oops/interface/ObsVector.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
}

namespace util {
  class DateTime;
}

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class ControlObsVector : public util::Printable,
                         private util::ObjectCounter<ControlObsVector<MODEL> > {
  typedef ObsVector<MODEL>        ObsVector_;
  typedef ObservationSpace<MODEL> ObsSpace_;  

 public:
  static const std::string classname() {return "oops::ControlObsVector";}

  explicit ControlObsVector(const ObsSpace_ &);
  explicit ControlObsVector(const ControlObsVector &, const bool copy = true);
  explicit ControlObsVector(ObsVector_ *, ObsVector_ *);
  ~ControlObsVector();

/// Interfacing
  ObsVector_ & obsvector() {return *obsvec_;}
  const ObsVector_ & obsvector() const {return *obsvec_;}
  ObsVector_ & biasvector() {ASSERT(biasvec_); return *biasvec_;}
  const ObsVector_ & biasvector() const {ASSERT(biasvec_); return *biasvec_;}
  inline bool hasBias() const {return biasvec_ ? true : false;}

// Linear algebra
  ControlObsVector & operator = (const ControlObsVector &);
  ControlObsVector & operator*= (const double &);
  ControlObsVector & operator+= (const ControlObsVector &);
  ControlObsVector & operator-= (const ControlObsVector &);
  ControlObsVector & operator*= (const ControlObsVector &);
  ControlObsVector & operator/= (const ControlObsVector &);

  void zero();
  void axpy(const double &, const ControlObsVector &);
  void invert();
  void random();
  double dot_product_with(const ControlObsVector &) const;
  double rms() const;
  unsigned int size() const {return obsvec_->size();}

// I/O
  void read(const eckit::Configuration &);
  void save(const eckit::Configuration &) const;

 private:
  void print(std::ostream &) const;
  boost::scoped_ptr<ObsVector_> obsvec_;
  boost::scoped_ptr<ObsVector_> biasvec_;
};

// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL>::ControlObsVector(const ObsSpace_ & os)
  : obsvec_(0), biasvec_(0) {
  Log::trace() << "ControlObsVector<MODEL>::ControlObsVector starting" << std::endl;
  util::Timer timer(classname(), "ControlObsVector");

  obsvec_.reset(new ObsVector_(os));
  if (os.config().has("ObsBias") ? os.config().getBool("ObsBias") : false) 
    biasvec_.reset(new ObsVector_(os));

  Log::trace() << "ControlObsVector<MODEL>::ControlObsVector done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL>::ControlObsVector(const ControlObsVector & other, const bool copy)
  : obsvec_(0), biasvec_(0) {
  Log::trace() << "ControlObsVector<MODEL>::ControlObsVector starting" << std::endl;
  util::Timer timer(classname(), "ControlObsVector");

  obsvec_.reset(new ObsVector_(*other.obsvec_, copy));
  if (other.hasBias()) biasvec_.reset(new ObsVector_(*other.biasvec_, copy));

  Log::trace() << "ControlObsVector<MODEL>::ControlObsVector done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL>::ControlObsVector(ObsVector_ * obsvec, ObsVector_ * biasvec)
  : obsvec_(obsvec), biasvec_(biasvec) {
  Log::trace() << "ControlObsVector<MODEL>::ControlObsVector done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL>::~ControlObsVector() {
  Log::trace() << "ControlObsVector<MODEL>::~ControlObsVector starting" << std::endl;
  util::Timer timer(classname(), "~ControlObsVector");

  obsvec_.reset();
  biasvec_.reset();

  Log::trace() << "ControlObsVector<MODEL>::~ControlObsVector done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL> & ControlObsVector<MODEL>::operator=(const ControlObsVector & rhs) {
  Log::trace() << "ControlObsVector<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");

  *obsvec_ = *rhs.obsvec_;
  if (this->hasBias()) *biasvec_ = *rhs.biasvec_;

  Log::trace() << "ControlObsVector<MODEL>::operator= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL> & ControlObsVector<MODEL>::operator*=(const double & zz) {
  Log::trace() << "ControlObsVector<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");

  *obsvec_ *= zz;
  if (this->hasBias()) *biasvec_ *= zz;

  Log::trace() << "ControlObsVector<MODEL>::operator*= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL> & ControlObsVector<MODEL>::operator+=(const ControlObsVector & rhs) {
  Log::trace() << "ControlObsVector<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");

  *obsvec_ += *rhs.obsvec_;
  if (this->hasBias()) *biasvec_ += *rhs.biasvec_;

  Log::trace() << "ControlObsVector<MODEL>::operator+= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL> & ControlObsVector<MODEL>::operator-=(const ControlObsVector & rhs) {
  Log::trace() << "ControlObsVector<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");

  *obsvec_ -= *rhs.obsvec_;
  if (this->hasBias()) *biasvec_ -= *rhs.biasvec_;

  Log::trace() << "ControlObsVector<MODEL>::operator-= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL> & ControlObsVector<MODEL>::operator*=(const ControlObsVector & rhs) {
  Log::trace() << "ControlObsVector<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");

  *obsvec_ *= *rhs.obsvec_;
  if (this->hasBias()) *biasvec_ *= *rhs.biasvec_;

  Log::trace() << "ControlObsVector<MODEL>::operator*= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
ControlObsVector<MODEL> & ControlObsVector<MODEL>::operator/=(const ControlObsVector & rhs) {
  Log::trace() << "ControlObsVector<MODEL>::operator/= starting" << std::endl;
  util::Timer timer(classname(), "operator/=");

  *obsvec_ /= *rhs.obsvec_;
  if (this->hasBias()) *biasvec_ /= *rhs.biasvec_;

  Log::trace() << "ControlObsVector<MODEL>::operator/= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlObsVector<MODEL>::zero() {
  Log::trace() << "ControlObsVector<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");

  obsvec_->zero();
  if (this->hasBias()) biasvec_->zero();

  Log::trace() << "ControlObsVector<MODEL>::zero done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlObsVector<MODEL>::axpy(const double & zz, const ControlObsVector & rhs) {
  Log::trace() << "ControlObsVector<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");

  obsvec_->axpy(zz, *rhs.obsvec_);
  if (this->hasBias()) biasvec_->axpy(zz, *rhs.biasvec_);

  Log::trace() << "ControlObsVector<MODEL>::axpy done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlObsVector<MODEL>::invert() {
  Log::trace() << "ControlObsVector<MODEL>::invert starting" << std::endl;
  util::Timer timer(classname(), "invert");

  obsvec_->invert();
  if (this->hasBias()) biasvec_->invert();

  Log::trace() << "ControlObsVector<MODEL>::invert done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlObsVector<MODEL>::random() {
  Log::trace() << "ControlObsVector<MODEL>::random starting" << std::endl;
  util::Timer timer(classname(), "random");

  obsvec_->random();
  if (this->hasBias()) biasvec_->random();

  Log::trace() << "ControlObsVector<MODEL>::random done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ControlObsVector<MODEL>::dot_product_with(const ControlObsVector & other) const {
  Log::trace() << "ControlObsVector<MODEL>::dot_product starting" << std::endl;
  util::Timer timer(classname(), "dot_product");

  double zz = 0.0;
  zz += obsvec_->dot_product_with(*other.obsvec_);
  if (this->hasBias()) zz += biasvec_->dot_product_with(*other.biasvec_);

  Log::trace() << "ControlObsVector<MODEL>::dot_product done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
double ControlObsVector<MODEL>::rms() const {
  Log::trace() << "ControlObsVector<MODEL>::rms starting" << std::endl;
  util::Timer timer(classname(), "rms");

  const double zz = 0.0;
  zz += obsvec_->rms();

  Log::trace() << "ControlObsVector<MODEL>::rms done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlObsVector<MODEL>::print(std::ostream & os) const {
  Log::trace() << "ControlObsVector<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");

  os << *obsvec_;
  if (this->hasBias()) os << std::endl << *biasvec_;

  Log::trace() << "ControlObsVector<MODEL>::print done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlObsVector<MODEL>::read(const eckit::Configuration & conf) {
  Log::trace() << "ControlObsVector<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");

  std::string obsname = conf.getString("obsvalue");
  obsvec_->read(obsname);

  if (this->hasBias()) {
    if (conf.has("biasvalue")) {
      std::string biasname = conf.getString("biasvalue");
      biasvec_->read(biasname);
    } else {
      biasvec_->zero();
    }
  }

  Log::trace() << "ControlObsVector<MODEL>::read done" << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlObsVector<MODEL>::save(const eckit::Configuration & conf) const {
  Log::trace() << "ControlObsVector<MODEL>::save starting";
  util::Timer timer(classname(), "save");

  std::string obsname = conf.getString("obsvalue");
  obsvec_->save(obsname);

  if (this->hasBias()) {
    if (conf.has("biasvalue")) {
      std::string biasname = conf.getString("biasvalue");
      biasvec_->save(biasname);
    }
  }

  Log::trace() << "ControlObsVector<MODEL>::save done" << std::endl;
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_CONTROLOBSVECTOR_H_
