/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_CONTROLVECTOR_H_
#define OOPS_ASSIMILATION_CONTROLVECTOR_H_

#include <ostream>
#include <sstream>
#include <string>

#include "oops/assimilation/IncrCtlVec4D.h"
#include "oops/interface/ModelAuxCtlVec.h"
#include "oops/interface/ObsAuxCtlVec.h"

#include "eckit/config/Configuration.h"

#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/dot_product.h"

namespace oops {

  /// Control Vector
  /*!
   * The control vector acts as a container for the inputs of the variational
   * data assimilation cost functions in the control variable space.
   * That includes the control vectors at the start the assimilation window or of each
   * sub-window but also additional variables such as model bias, VarBC
   * coefficients, or other control variables for algorithms that use them.
   */

template<typename MODEL> class CostJbTotal;

// -----------------------------------------------------------------------------
template<typename MODEL>
class ControlVector : public util::Printable,
                      private util::ObjectCounter<ControlVector<MODEL> > {
  typedef CostJbTotal<MODEL>         JbTotal_;
  typedef IncrCtlVec4D<MODEL>        IncrCtlVec4D_;
  typedef ModelAuxCtlVec<MODEL>      ModelAuxCtlVec_;
  typedef ObsAuxCtlVec<MODEL>        ObsAuxCtlVec_;

 public:
  static const std::string classname() {return "oops::ControlVector";}

/// Constructor, destructor
  explicit ControlVector(const JbTotal_ &);
  ControlVector(const JbTotal_ &, const ControlVector &);
  ControlVector(const ControlVector &, const bool copy = true);
  ControlVector(const ControlVector &, const eckit::Configuration &);
  ~ControlVector();

/// Linear algebra operators
  void zero();
  ControlVector & operator=(const ControlVector &);
  ControlVector & operator+=(const ControlVector &);
  ControlVector & operator-=(const ControlVector &);
  ControlVector & operator*=(const double);
  void axpy(const double, const ControlVector &);
  double dot_product_with(const ControlVector &) const;

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;

/// Get state control variable
  IncrCtlVec4D_ & state() {return incrCtlVec4D_;}
  const IncrCtlVec4D_ & state() const {return incrCtlVec4D_;}

/// Get augmented model control variable
  ModelAuxCtlVec_ & modVar() {return modAuxCtlVec_;}
  const ModelAuxCtlVec_ & modVar() const {return modAuxCtlVec_;}

/// Get augmented observation control variable
  ObsAuxCtlVec_ & obsVar() {return obsAuxCtlVec_;}
  const ObsAuxCtlVec_ & obsVar() const {return obsAuxCtlVec_;}

 private:
  void print(std::ostream &) const;

  IncrCtlVec4D_   incrCtlVec4D_;
  ModelAuxCtlVec_ modAuxCtlVec_;
  ObsAuxCtlVec_   obsAuxCtlVec_;
};

// =============================================================================

template<typename MODEL>
ControlVector<MODEL>::ControlVector(const JbTotal_ & jb)
  : incrCtlVec4D_(jb.jbState()), 
    modAuxCtlVec_(jb.jbModBias().errorCovariance()),
    obsAuxCtlVec_(jb.jbObsBias())
{
  Log::trace() << "ControlVector:ControlVector created." << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ControlVector<MODEL>::ControlVector(const JbTotal_ & jb, const ControlVector & other)
  : incrCtlVec4D_(jb.jbState()), 
    modAuxCtlVec_(jb.jbModBias().errorCovariance()),
    obsAuxCtlVec_(jb.jbObsBias())
{
  incrCtlVec4D_ = other.incrCtlVec4D_;
  modAuxCtlVec_ = other.modAuxCtlVec_;
  obsAuxCtlVec_ = other.obsAuxCtlVec_;
  Log::trace() << "ControlVector:ControlVector created." << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ControlVector<MODEL>::ControlVector(const ControlVector & other, const bool copy)
  : incrCtlVec4D_(other.incrCtlVec4D_, copy), 
    modAuxCtlVec_(other.modAuxCtlVec_, copy),
    obsAuxCtlVec_(other.obsAuxCtlVec_, copy)
{
  Log::trace() << "ControlVector:ControlVector copied." << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ControlVector<MODEL>::ControlVector(const ControlVector & other,
                                    const eckit::Configuration & tlConf)
  : incrCtlVec4D_(other.incrCtlVec4D_, tlConf), 
    modAuxCtlVec_(other.modAuxCtlVec_, tlConf),
    obsAuxCtlVec_(other.obsAuxCtlVec_)
{
  Log::trace() << "ControlVector:ControlVector copied." << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ControlVector<MODEL>::~ControlVector() {}
// -----------------------------------------------------------------------------
template<typename MODEL> ControlVector<MODEL> &
ControlVector<MODEL>::operator=(const ControlVector & rhs) {
  incrCtlVec4D_ = rhs.incrCtlVec4D_;
  modAuxCtlVec_ = rhs.modAuxCtlVec_;
  obsAuxCtlVec_ = rhs.obsAuxCtlVec_;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL> ControlVector<MODEL> &
ControlVector<MODEL>::operator+=(const ControlVector & rhs) {
  incrCtlVec4D_ += rhs.incrCtlVec4D_;
  modAuxCtlVec_ += rhs.modAuxCtlVec_;
  obsAuxCtlVec_ += rhs.obsAuxCtlVec_;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL> ControlVector<MODEL> &
ControlVector<MODEL>::operator-=(const ControlVector & rhs) {
  incrCtlVec4D_ -= rhs.incrCtlVec4D_;
  modAuxCtlVec_ -= rhs.modAuxCtlVec_;
  obsAuxCtlVec_ -= rhs.obsAuxCtlVec_;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ControlVector<MODEL> & ControlVector<MODEL>::operator*=(const double zz) {
  incrCtlVec4D_ *= zz;
  modAuxCtlVec_ *= zz;
  obsAuxCtlVec_ *= zz;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ControlVector<MODEL>::zero() {
  incrCtlVec4D_.zero();
  modAuxCtlVec_.zero();
  obsAuxCtlVec_.zero();
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ControlVector<MODEL>::axpy(const double zz, const ControlVector & rhs) {
  incrCtlVec4D_.axpy(zz, rhs.incrCtlVec4D_);
  modAuxCtlVec_.axpy(zz, rhs.modAuxCtlVec_);
  obsAuxCtlVec_.axpy(zz, rhs.obsAuxCtlVec_);
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ControlVector<MODEL>::read(const eckit::Configuration & config) {
  incrCtlVec4D_.read(config);
  modAuxCtlVec_.read(config);
  obsAuxCtlVec_.read(config);
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ControlVector<MODEL>::write(const eckit::Configuration & config) const {
  incrCtlVec4D_.write(config);
  modAuxCtlVec_.write(config);
  obsAuxCtlVec_.write(config);
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void ControlVector<MODEL>::print(std::ostream & outs) const {
  outs << incrCtlVec4D_;
  outs << modAuxCtlVec_;
  outs << obsAuxCtlVec_;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
double ControlVector<MODEL>::dot_product_with(const ControlVector & x2) const {
  double zz = 0.0;
  zz += dot_product(incrCtlVec4D_, x2.incrCtlVec4D_);
  zz += dot_product(modAuxCtlVec_, x2.modAuxCtlVec_);
  zz += dot_product(obsAuxCtlVec_, x2.obsAuxCtlVec_);
  return zz;
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_ASSIMILATION_CONTROLVECTOR_H_
