/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTJB4D_H_
#define OOPS_ASSIMILATION_COSTJB4D_H_

#include <memory>
#include <vector>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/scoped_ptr.hpp>

#include "eckit/config/LocalConfiguration.h"
#include "oops/assimilation/CostJbState.h"
#include "oops/assimilation/Increment4D.h"
#include "oops/assimilation/IncrCtlVec4D.h"
#include "oops/assimilation/State4D.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/base/ModelSpaceCovarianceBase.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/Increment.h"
#include "oops/interface/Variables.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/dot_product.h"
#include "util/Logger.h"

namespace oops {
  template<typename MODEL> class ControlIncrement;
  template<typename MODEL> class IncrCtlVec;
  template<typename MODEL> class JqTerm;
  template<typename MODEL> class JqTermTL;
  template<typename MODEL> class JqTermAD;

// -----------------------------------------------------------------------------

/// 4D Jb Cost Function
/*!
 * CostJb4D encapsulates the generalized four dimensional Jb term of
 * the 4D-Ens-Var cost function. NOTE: for the square root formulation
 * it is assumed that the localization matrix is constant 
 * L_bar = [I I .. I]^T L [I I .. I]
 */

template<typename MODEL> class CostJb4D : public CostJbState<MODEL> {
  typedef ControlIncrement<MODEL>    CtrlInc_;
  typedef EnsemblesCollection<MODEL> EnsemblesCollection_;
  typedef Geometry<MODEL>            Geometry_;
  typedef Increment<MODEL>           Increment_;
  typedef Increment4D<MODEL>         Increment4D_;
  typedef IncrCtlVec<MODEL>          IncrCtlVec_;
  typedef IncrCtlVec4D<MODEL>        IncrCtlVec4D_;
  typedef State4D<MODEL>             State4D_;
  typedef Variables<MODEL>           Variables_;

 public:
/// Construct \f$ J_b\f$.
  CostJb4D(const eckit::Configuration &, const Geometry_ &, const Variables_ &,
           const util::Duration &, const State4D_ &);

/// Destructor
  virtual ~CostJb4D() {}

/// Empty Jq observer.
  JqTerm<MODEL> * initializeJq() const override {return 0;}

/// Get increment from state (usually first guess).
  void computeIncrement(const State4D_ &, const State4D_ &, Increment4D_ &) const override;

/// Linearize before the linear computations.
  void linearize(const State4D_ &, const Geometry_ &) override;

/// Add Jb gradient.
  void addGradient(const Increment4D_ &, Increment4D_ &, Increment4D_ &) const override;

/// Empty TL Jq observer.
  JqTermTL<MODEL> * initializeJqTL() const override {return 0;}

/// Empty AD Jq observer.
  JqTermAD<MODEL> * initializeJqAD(const Increment4D_ &) const override {return 0;}

/// Multiply by \f$ B\f$ and \f$ B^{-1}\f$.
  void Bmult(const Increment4D_ &, Increment4D_ &) const override;
  void Bminv(const Increment4D_ &, Increment4D_ &) const override;

/// Multiply by \f$ B^{1/2}\f$ and \f$ B^{1/2}^T\f$.
  void BmultSqrt(const IncrCtlVec4D_ &, Increment4D_ &) const override;
  void BmultSqrtTrans(const Increment4D_ &, IncrCtlVec4D_ &) const override;

/// Access to B
  const ModelSpaceCovarianceBase<MODEL> & covar() const {return B_[0];}

/// Randomize
  void randomize(Increment4D_ &) const override;

/// Create new increment (set to 0).
  unsigned int nstates() const override {return B_.size();}
  Increment_ * newStateIncrement(const unsigned int) const override;

/// Create new control vector 
  unsigned int nchistates() const override {return 1;}
  IncrCtlVec_ * newIncrCtlVec(const unsigned int) const override;

 private:
  boost::ptr_vector< ModelSpaceCovarianceBase<MODEL> > B_;
  const Variables_ controlvars_;
  boost::scoped_ptr<const Geometry_> resol_;
  std::vector<util::DateTime> times_;
};

// =============================================================================

//  Generalized Jb Term of Cost Function
// -----------------------------------------------------------------------------

template<typename MODEL>
CostJb4D<MODEL>::CostJb4D(const eckit::Configuration & config, const Geometry_ & resolouter,
                          const Variables_ & ctlvars, const util::Duration &, const State4D_ & xb)
  : B_(), controlvars_(ctlvars), resol_(), times_()
{
// Create one row of blocks of the whole BMatrix, one object for each
// subwindow. Each object stands for all blocks in the same column of the B Matrix.
// It can be from any concrete class of ModelSpaceCovarianceBase,
// according to the configuration file.
  const eckit::LocalConfiguration covar(config, "Covariance");
  std::vector<eckit::LocalConfiguration> confs;
  covar.get("covariance_time", confs);

  for (size_t jsub = 0; jsub < confs.size(); ++jsub) {
    B_.push_back(CovarianceFactory<MODEL>::create(confs[jsub], resolouter, ctlvars, xb[jsub]));
  }

  Log::trace() << "CostJb4D contructed." << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJb4D<MODEL>::linearize(const State4D_ & fg, const Geometry_ & resolinner) {
  ASSERT(fg.checkStatesNumber(B_.size()));
  resol_.reset(new Geometry_(resolinner));
  times_.clear();
  for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
    times_.push_back(fg[jsub].validTime());
    B_[jsub].linearize(fg[jsub], *resol_);
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJb4D<MODEL>::computeIncrement(const State4D_ & xb, const State4D_ & fg,
                                       Increment4D_ & dx) const {
  dx.diff(fg, xb);
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJb4D<MODEL>::addGradient(const Increment4D_ & dxFG, Increment4D_ & grad,
                                  Increment4D_ & gradJb) const {
  grad -= gradJb;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJb4D<MODEL>::Bmult(const Increment4D_ & dxin, Increment4D_ & dxout) const {
  for (unsigned k1 = 0; k1 < B_.size(); ++k1) {
    // Apply the line k1 of the whole B matrix to the StateIncrement dxin
    // Result is the part of dxout at time k1
    dxout[k1].zero();
    Increment_ dout(dxout[k1]);
    for (unsigned k2 = 0; k2 < B_.size(); ++k2) {
      // Apply to increment at time k2 the block B_k1k2 of the whole B matrix.
      // We need an object B which is also at time k2
      // Result "dout" is an increment at time k1
      B_[k2].multiply(dxin[k2], dout);
      dxout[k1] += dout;
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJb4D<MODEL>::Bminv(const Increment4D_ & dxin, Increment4D_ & dxout) const {
  Log::warning() << "*** B inverse might not always exist ***" << std::endl;
  for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
    B_[jsub].inverseMultiply(dxin[jsub], dxout[jsub]);
  }
  Log::warning() << "*** B inverse might not always exist ***" << std::endl;
}


// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJb4D<MODEL>::BmultSqrt(const IncrCtlVec4D_ & dv, Increment4D_ & dxout) const {
  // Be^{1/2} = D^hat L^hat^{1/2}
  //   D^hat = (D_bar_1, .., D_bar_Ne)
  //   L_hat^{1/2} = diag( 1_bar L^{1/2}, .., 1_bar L^{1/2} )
  //     D_bar_i = diag(X_0^i, X_1^i, .., X_Ns^i)
  //     1_bar = [1 1 .. 1]^T
  //
  //     Ne - number of ensemble members
  //     Ns - number of time steps 
  for (unsigned k1 = 0; k1 < B_.size(); ++k1) {
    B_[k1].multiplySqrt(dv[0], dxout[k1]);
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJb4D<MODEL>::BmultSqrtTrans(const Increment4D_ & dxin, IncrCtlVec4D_ & dvout) const {
  IncrCtlVec_ dout(dvout[0]);
  dvout[0].zero();
  for (unsigned k1 = 0; k1 < B_.size(); ++k1) {
    B_[k1].multiplySqrtTrans(dxin[k1], dout);
    dvout[0] += dout;
  }
}
   
// -----------------------------------------------------------------------------


template<typename MODEL>
void CostJb4D<MODEL>::randomize(Increment4D_ & dx) const {
  for (unsigned jsub = 0; jsub < B_.size(); ++jsub) {
    B_[jsub].randomize(dx[jsub]);
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
Increment<MODEL> *
CostJb4D<MODEL>::newStateIncrement(const unsigned int isub) const {
  Increment_ * incr = new Increment_(*resol_, controlvars_, times_[isub]);
  return incr;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL> * 
CostJb4D<MODEL>::newIncrCtlVec(const unsigned int isub) const {
  return new IncrCtlVec_(*this, times_[isub]);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTJB4D_H_
