/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_COSTJO_H_
#define OOPS_ASSIMILATION_COSTJO_H_

#include <string>
#include <boost/noncopyable.hpp>
#include <boost/pointer_cast.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "eckit/config/LocalConfiguration.h"

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/base/Departures.h"
#include "oops/base/LinearObsOperators.h"
#include "oops/base/Observations.h"
#include "oops/base/Observer.h"
#include "oops/base/ObserverTL.h"
#include "oops/base/ObserverAD.h"
#include "oops/base/ObsErrors.h"
#include "oops/base/ObsOperators.h"
#include "oops/base/ObsSpaces.h"
#include "oops/base/PostBase.h"
#include "oops/base/PostBaseTL.h"
#include "oops/base/PostBaseAD.h"
#include "oops/interface/Geometry.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/Increment.h"
#include "oops/interface/ObsAuxControl.h"
#include "oops/interface/ObsAuxIncrement.h"
#include "oops/interface/State.h"

#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Logger.h"

namespace oops {

// -----------------------------------------------------------------------------

/// Jo Cost Function
/*!
 * The CostJo class encapsulates the Jo term of the cost function.
 * The Observer to be called during the model integration is managed
 * inside the CostJo class.
 */

template<typename MODEL> class CostJo : public CostTermBase<MODEL>,
                                        private boost::noncopyable {
  typedef ControlVariable<MODEL>     CtrlVar_;
  typedef ControlIncrement<MODEL>    CtrlInc_;
  typedef Departures<MODEL>          Departures_;
  typedef Observations<MODEL>        Observations_;
  typedef Geometry<MODEL>            Geometry_;
  typedef GeoVaLs<MODEL>             GeoVaLs_;
  typedef State<MODEL>               State_;
  typedef Increment<MODEL>           Increment_;
  typedef ObsAuxControl<MODEL  >     ObsAuxControl_;
  typedef ObsAuxIncrement<MODEL>     ObsAuxIncr_;
  typedef ObsErrors<MODEL>           ObsErrors_;
  typedef ObsOperators<MODEL>        ObsOperators_;
  typedef ObsSpaces<MODEL>           ObsSpaces_;
  typedef LinearObsOperators<MODEL>  LinearObsOperators_;

 public:
  static const std::string classname() {return "oops::CostJo";}

  /// Define which ObservationSpace to use
  enum OS : bool {
    NotScreened = false,
    Screened = true
  };

  /// Construct \f$ J_o\f$ from \f$ R\f$ and \f$ y_{obs}\f$.
  CostJo(const eckit::Configuration &,
         const util::DateTime &, const util::DateTime &,
         const util::Duration &, const bool subwindows = false);

  /// Destructor
  virtual ~CostJo() {}

  /// Initialize \f$ J_o\f$ before starting the integration of the model.
  boost::shared_ptr<PostBase<State_> > initialize(const CtrlVar_ &,
                                                  const eckit::Configuration &) const override;
  boost::shared_ptr<PostBase<State_> > initializeTraj(const CtrlVar_ &,
                                                      const Geometry_ &,
                                                      const eckit::Configuration &) override;
  /// Finalize \f$ J_o\f$ after the integration of the model.
  double finalize(const eckit::Configuration &) const override;
  double finalizeTraj(const eckit::Configuration &) override;

  /// Screen observations
  void screenObservations(boost::scoped_ptr<Departures_> &, 
                          boost::scoped_ptr<Observations_> &,
                          const eckit::Configuration &) const;

  /// Initialize \f$ J_o\f$ before starting the TL run.
  boost::shared_ptr<PostBaseTL<Increment_> > setupTL(const CtrlInc_ &) const override;

  /// Initialize \f$ J_o\f$ before starting the AD run.
  boost::shared_ptr<PostBaseAD<Increment_> > setupAD(
           boost::shared_ptr<const GeneralizedDepartures>, CtrlInc_ &) const override;

  /// Multiply by \f$ R\f$ and \f$ R^{-1}\f$.
  Departures_ * multiplyCovar(const GeneralizedDepartures &) const override;
  Departures_ * multiplyCoInv(const GeneralizedDepartures &) const override;

  /// Provide new departure.
  Departures_ * newDualVector() const override;

  /// Return gradient at first guess ie \f$ R^{-1} {\cal H}(x^t ) - y\f$.
  Departures_ * newGradientFG() const override {return new Departures_(*gradFG_);}

  /// Reset obs operator trajectory.
  void resetLinearization() override;

  /// Print Jo
  double printJo(const Departures_ &, 
                 const Departures_ &, 
                 const bool & lscreened) const;

  /// Checkpoint Jo
  void checkpoint(eckit::LocalConfiguration &) const;

  /// Restart Jo
  void restart(const eckit::Configuration &);

  /// Set xb
  void setBackground(const CtrlVar_ & xb) const {xb_ = &xb;}

 private:
  /// Config
  const eckit::LocalConfiguration conf_;

  /// Window
  const util::DateTime winbgn_; // Begining of the window
  const util::DateTime winend_; // End of the window
  const util::Duration tslot_;  // Time slot

  /// Controls
  const bool subwindows_;
  mutable bool lpreScreenedObservations_;
  mutable bool lscreenObservations_;
  mutable bool lsetupOnce_;

  /// Background
  const CtrlVar_ mutable * xb_;

  /// First guess
  const CtrlVar_ mutable * fg_;

  /// Observation space
  mutable boost::scoped_ptr<ObsSpaces_> obsSpaces_;

  /// Operators
  mutable boost::scoped_ptr<ObsOperators_> hop_;
  boost::shared_ptr<LinearObsOperators_> hoptlad_;
  mutable boost::scoped_ptr<ObsErrors_> R_;

  /// Observations
  mutable boost::scoped_ptr<Observations_> yobs_;

  /// Gradient at first guess : \f$ R^{-1} (H(x_{fg})-y_{obs}) \f$.
  mutable boost::scoped_ptr<Departures_> gradFG_;

  /// Observer passed by \f$ J_o\f$ to the model during integration.
  mutable boost::shared_ptr<Observer<MODEL, State_> > pobs_;
  mutable boost::shared_ptr<Observer<MODEL, State_> > pobstraj_;
  mutable boost::shared_ptr<ObserverTL<MODEL, Increment_> > pobstl_;
  mutable boost::shared_ptr<ObserverAD<MODEL, Increment_> > pobsad_;

};

// =============================================================================

template<typename MODEL>
CostJo<MODEL>::CostJo(const eckit::Configuration & joConf,
                      const util::DateTime & winbgn, const util::DateTime & winend,
                      const util::Duration & ts, const bool subwindows)
  : conf_(joConf), winbgn_(winbgn), winend_(winend), tslot_(ts), subwindows_(subwindows),
    lpreScreenedObservations_(false), lscreenObservations_(false), lsetupOnce_(true), 
    xb_(NULL), fg_(NULL), obsSpaces_(), hop_(), hoptlad_(), R_(), yobs_(), gradFG_(),
    pobs_(), pobstraj_(), pobstl_(), pobsad_()
{
  Log::debug() << "CostJo:setup tslot_ = " << tslot_ << std::endl;
  if (conf_.has("PreScreenedObservations")) {
    lpreScreenedObservations_ = conf_.getBool("PreScreenedObservations");
    if (lpreScreenedObservations_)
      obsSpaces_.reset(new ObsSpaces_(conf_, winbgn_, winend_, OS::Screened));
  } else {
    obsSpaces_.reset(new ObsSpaces_(conf_, winbgn_, winend_, OS::NotScreened));
  }
  Log::trace() << "CostJo:CostJo done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
boost::shared_ptr<PostBase<State<MODEL> > >
CostJo<MODEL>::initialize(const CtrlVar_ & fg,
                          const eckit::Configuration & conf) const {
  ASSERT(lscreenObservations_ == false);
  // At each outer loop setup here a switch to control if screening is on;
  // If observations were already pre-screened don't allow screening
  if (conf.has("screening") && not lpreScreenedObservations_) 
    lscreenObservations_ = conf.getBool("screening.active");

  // Get a reference to the first guess
  fg_ = &fg;

  if (lsetupOnce_ || lscreenObservations_) {
    // This should be setup just once if screening is off
    if (not lsetupOnce_)
      obsSpaces_.reset(new ObsSpaces_(conf_, winbgn_, winend_, OS::NotScreened));
    hop_.reset(new ObsOperators_(*obsSpaces_)); 
    R_.reset(new ObsErrors_(*obsSpaces_)); 
    yobs_.reset(new Observations_(*obsSpaces_));
    yobs_->read(conf_);
    //
    lsetupOnce_ = false;
  }

  pobs_.reset(new Observer<MODEL, State_>(*obsSpaces_, *hop_, fg.obsVar(),
                                          tslot_, subwindows_));
  return pobs_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double CostJo<MODEL>::finalize(const eckit::Configuration & conf) const {
  // Write the observation space component of the first guess
  if (conf.has("stateObsAux")) {
    fg_->obsVar().write(conf);
  }

  // Get observation equivalents
  boost::scoped_ptr<Observations_> yeqv(pobs_->release());
  Log::info() << "Jo Observation Equivalent: " << *yeqv << std::endl;

  // Calculate departures
  boost::scoped_ptr<Departures_> ydep(new Departures_(*yobs_ - *yeqv));
  Log::info() << "Jo Departures: " << *ydep << std::endl;

  // Screen observations
  if (lscreenObservations_) 
    this->screenObservations(ydep, yeqv, conf); 
  
  // Linearize R
  R_->linearize(*yeqv);

  // Setup VarQC weights
  if (conf.getBool("varQC.activate",false))
    R_->setupWeights(*ydep);

  // Calculate FG gradient
  gradFG_.reset(R_->inverseMultiply(*ydep));

  // Print JO table
  double zjo = this->printJo(*ydep, *gradFG_, OS::Screened);

  // Diagnostics
  eckit::LocalConfiguration diagnostics;
  if (conf.has("diagnostics")) { 
    diagnostics = eckit::LocalConfiguration(conf, "diagnostics");
    if (diagnostics.has("departures")) {
      const eckit::LocalConfiguration depconf(diagnostics, "departures");
      ydep->save(depconf);
    }
  }

  pobs_.reset();

  // Reset the first guess pointer
  fg_ = NULL;

  return zjo;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJo<MODEL>::screenObservations(boost::scoped_ptr<Departures_> & ydep,
                                       boost::scoped_ptr<Observations_> & yeqv,
                                       const eckit::Configuration & conf) const {
    // ---------------------------------------------------------------------------
    // MC: should be implemented as a post-processor in the Observer at some point
    // ---------------------------------------------------------------------------
    // Print unscreened Jo table
    R_->linearize(*yeqv);
    if (conf.getBool("varQC.activate",false))
      R_->setupWeights(*ydep);
    boost::scoped_ptr<Departures_> gradFG(R_->inverseMultiply(*ydep));
    double zjo = this->printJo(*ydep, *gradFG, OS::NotScreened);
    Log::debug() << "CostJo (before screening)" << zjo << std::endl;

    // Save departures in the current ObservationSpace
    const eckit::LocalConfiguration depconf(conf, "screening.departures");
    ydep->save(depconf);

    // Get a vector of pointers to GeoVaLs from the Observer
    const std::vector<boost::shared_ptr<GeoVaLs_> > & gvals = pobs_->geovals();

    // Run screening - this saves a screened ObservationSpace
    obsSpaces_->screenObservations(*ydep, gvals);

    // Close the current ObservationSpace and read in the screened ObservationSpace
    obsSpaces_.reset();
    obsSpaces_.reset(new ObsSpaces_(conf_, winbgn_, winend_, OS::Screened));

    // Reset R
    R_.reset(new ObsErrors_(*obsSpaces_)); 

    // Reset non-linear observation operator
    hop_.reset(new ObsOperators_(*obsSpaces_)); 

    // Reset y0
    yobs_.reset(new Observations_(*obsSpaces_));
    yobs_->read(conf_);

    // Reset ydep
    ydep.reset(new Departures_(*obsSpaces_));
    ydep->read(depconf);

    // Reset yeqv
    yeqv.reset(new Observations_(*yobs_));
    *yeqv -= *ydep;

    // Reset the observation space component of the first guess
    ObsAuxControl_ & obsvarfg = const_cast<ObsAuxControl_ &>(fg_->obsVar());
    obsvarfg.reset("fg", conf);

    // Reset the observation space component of the background
    ObsAuxControl_ & obsvarbg = const_cast<ObsAuxControl_ &>(xb_->obsVar());
    obsvarbg.reset("xb", conf);
}
// -----------------------------------------------------------------------------

template<typename MODEL>
boost::shared_ptr<PostBase<State<MODEL> > >
CostJo<MODEL>::initializeTraj(const CtrlVar_ & fg, const Geometry_ &,
                              const eckit::Configuration & conf) {
  util::Duration tltstep = tslot_;
  if (conf.has("linearmodel"))
    tltstep = util::Duration(conf.getString("linearmodel.tstep"));

  hoptlad_.reset();  
  pobstraj_.reset();

  hoptlad_.reset(new LinearObsOperators_(*obsSpaces_));
  pobstraj_.reset(new Observer<MODEL, State_>(*obsSpaces_, *hop_, fg.obsVar(),
                                              tltstep, subwindows_, hoptlad_));
  return pobstraj_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double CostJo<MODEL>::finalizeTraj(const eckit::Configuration & conf) {
  boost::scoped_ptr<Observations_> yeqv(pobstraj_->release());
  Log::info() << "Jo Observation Equivalent:" << *yeqv << std::endl;

  Departures_ ydep(*yobs_ - *yeqv);
  Log::info() << "Jo Departures:" << ydep << std::endl;
  
  boost::scoped_ptr<Departures_> grad(R_->inverseMultiply(ydep)); 

  // Check config if the inner loop trajectory Jo needs to be printed 
  bool printInnerLoopTrajJo = false;
  if (conf.has("printInnerLoopTrajJo")) 
    printInnerLoopTrajJo = conf.getBool("printInnerLoopTrajJo");

  double zjo = 0.0;
  if (printInnerLoopTrajJo) {
     // Print JO table
     Log::info() << "Print inner loop trajectory JO-table" << std::endl;
     zjo = this->printJo(ydep, *grad, OS::Screened);
  } else {
    // Calculate JO
    for (std::size_t jj = 0; jj < ydep.size(); ++jj) {
      const double zz = 0.5 * dot_product(ydep[jj].obsvector(), (*grad)[jj].obsvector());
      zjo += zz;
    }
  }

  if (conf.has("innerloop_departures")) {
    const eckit::LocalConfiguration depconf(conf, "innerloop_departures");
    ydep.save(depconf);
  }

  pobstraj_.reset();
  return zjo;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
boost::shared_ptr<PostBaseTL<Increment<MODEL> > > CostJo<MODEL>::setupTL(
                               const CtrlInc_ & dx) const {
  ASSERT(hoptlad_);
  if (pobstl_ == nullptr)
    pobstl_.reset(new ObserverTL<MODEL, Increment_>(*obsSpaces_, *hoptlad_, dx.obsVar(),
                                                    tslot_, subwindows_));
  else {
    pobstl_->setupTL(dx.obsVar());
  }
  return pobstl_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
boost::shared_ptr<PostBaseAD<Increment<MODEL> > > CostJo<MODEL>::setupAD(
                               boost::shared_ptr<const GeneralizedDepartures> pv,
                               CtrlInc_ & dx) const {
  ASSERT(hoptlad_);
  boost::shared_ptr<const Departures_> dy = boost::dynamic_pointer_cast<const Departures_>(pv);
  if (pobsad_ == nullptr)
    pobsad_.reset(new ObserverAD<MODEL, Increment_>(*obsSpaces_, *hoptlad_, dy, dx.obsVar(),
                                                    tslot_, subwindows_));
  else {
    pobsad_->setupAD(dy, dx.obsVar());
  }
  return pobsad_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
Departures<MODEL> * CostJo<MODEL>::multiplyCovar(const GeneralizedDepartures & v1) const {
  const Departures_ & y1 = dynamic_cast<const Departures_ &>(v1);
  return R_->multiply(y1);
}

// -----------------------------------------------------------------------------

template<typename MODEL>
Departures<MODEL> * CostJo<MODEL>::multiplyCoInv(const GeneralizedDepartures & v1) const {
  const Departures_ & y1 = dynamic_cast<const Departures_ &>(v1);
  return R_->inverseMultiply(y1);
}

// -----------------------------------------------------------------------------

template<typename MODEL>
Departures<MODEL> * CostJo<MODEL>::newDualVector() const {
  Departures_ * ydep = new Departures_(*obsSpaces_);
  ydep->zero();
  return ydep;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJo<MODEL>::resetLinearization() {
  hoptlad_.reset();
  pobs_.reset();
  pobstraj_.reset();
  pobstl_.reset();
  pobsad_.reset();
  lscreenObservations_ = false;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double CostJo<MODEL>::printJo(const Departures_ & dy, 
                              const Departures_ & grad, 
                              const bool & lscreened ) const {
  obsSpaces_->printJo(dy, grad);

  double zjo = 0.0;
  for (std::size_t jj = 0; jj < dy.size(); ++jj) {
    const double zz = 0.5 * dot_product(dy[jj].obsvector(), grad[jj].obsvector());
    const unsigned nobs = dy[jj].size();
    if (nobs > 0) {
      if (lscreened) {
        Log::test() << "CostJo   : Nonlinear Jo = " << zz
                    << ", nobs = " << nobs << ", Jo/n = " << zz/nobs
                    << ", err = " << (*R_)[jj].getRMSE() << std::endl;
      } else {
        Log::info() << "CostJo (not screened)   : Nonlinear Jo = " << zz
                    << ", nobs = " << nobs << ", Jo/n = " << zz/nobs
                    << ", err = " << (*R_)[jj].getRMSE() << std::endl;
      }
      if (dy[jj].hasBias()) {
        ASSERT(grad[jj].hasBias());
        double zjc = 0.5 * dot_product(dy[jj].biasvector(), grad[jj].biasvector());
        if (lscreened) {
          Log::test() << "CostJo   : Nonlinear JoCVarBC = " 
                      << zjc << std::endl;
        } else {
          Log::info() << "CostJo (not screened)  : Nonlinear JoCVarBC = " 
                      << zjc << std::endl;
        }
      }
    } else {
      if (lscreened) {
        Log::test() << "CostJo   : Nonlinear Jo = " 
                    << zz << " --- No Observations" << std::endl;
      } else {
        Log::info() << "CostJo (not screened)  : Nonlinear Jo = " 
                    << zz << " --- No Observations" << std::endl;
      }
      Log::warning() << "CostJo: No Observations!!!" << std::endl;
    }
    zjo += zz;
  }

  return zjo;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJo<MODEL>::checkpoint(eckit::LocalConfiguration & conf) const {
   if (conf.has("jo")) {
      eckit::LocalConfiguration jochckptconfg(conf, "jo"); 
      obsSpaces_->write(jochckptconfg);
   }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void CostJo<MODEL>::restart(const eckit::Configuration & conf) {
   if (conf.has("jo")) {
      eckit::LocalConfiguration joconfg(conf, "jo");
      obsSpaces_->read(joconfg);
   }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_COSTJO_H_
