/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_DRMINIMIZER_H_
#define OOPS_ASSIMILATION_DRMINIMIZER_H_

#include <string>
#include <vector>

#include <boost/scoped_ptr.hpp>

#include "eckit/config/Configuration.h"
#include "oops/assimilation/BMatrix.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/HtRinvHdxFG.h"
#include "oops/assimilation/HtRinvHMatrix.h"
#include "oops/assimilation/Minimizer.h"

#include "util/dot_product.h"
#include "util/formats.h"
#include "util/Logger.h"

namespace oops {

/// DR (Derber and Rosati) Minimizers
/*!
 * DRMinimizer is the base class for all minimizers that use \f$ B\f$ to
 * precondition the variational minimisation problem and use the auxiliary
 * variable \f$ x = B \hat{x}\f$ or \f$ x = B \hat{x} + x_b -x_fg\f$ and
 * to update it in parallel to \f$ x\f$  based on
 * Derber and Rosati, 1989, J. Phys. Oceanog. 1333-1347.
 * \f$ J_b\f$ is then computed as \f$ x^T\hat{x}\f$ eliminating the need for
 * \f$ B^{-1}\f$ or \f$ B^{1/2}\f$.
 */

// -----------------------------------------------------------------------------

template<typename MODEL> class DRMinimizer : public Minimizer<MODEL> {
  typedef BMatrix<MODEL>          Bmat_;
  typedef CostFunction<MODEL>     CostFct_;
  typedef ControlIncrement<MODEL> CtrlInc_;
  typedef HtRinvHdxFG<MODEL>      HtRinvHdxFG_;
  typedef HtRinvHMatrix<MODEL>    HtRinvH_;
  typedef Minimizer<MODEL>        Minimizer_;

 public:
  virtual const std::string classname() const override =0;

  explicit DRMinimizer(const CostFct_ & J)
   : Minimizer_(J), J_(J), gradJb_(0), dxh0_(0), dxh_(), costJ0Jb_(0) {}
  ~DRMinimizer() {}

  void checkpoint(eckit::LocalConfiguration &) const;
  void restart(const eckit::Configuration &);

 private:
  CtrlInc_ * doMinimize(const eckit::Configuration &) override;
  virtual double solve(CtrlInc_ &, CtrlInc_ &, CtrlInc_ &,
                       const Bmat_ &, const HtRinvH_ &,
                       const double, const double, 
                       const int, const double) =0;

  const CostFct_ & J_;
  boost::scoped_ptr<CtrlInc_> gradJb_;
  boost::scoped_ptr<CtrlInc_> dxh0_;
  std::vector<CtrlInc_> dxh_;
  double costJ0Jb_;
};

// =============================================================================

template<typename MODEL>
ControlIncrement<MODEL> * DRMinimizer<MODEL>::doMinimize(const eckit::Configuration & config) {
  int ninner = config.getInt("ninner");
  double gnreduc = config.getDouble("gradient_norm_reduction");

  bool runOnlineAdjTest = false;
  if (config.has("onlineDiagnostics")) {
    const eckit::LocalConfiguration onlineDiag(config, "onlineDiagnostics");
    runOnlineAdjTest = onlineDiag.getBool("onlineAdjTest");
  }

// Interpolate gradJB/Set initial guess to the new resolution
  if (J_.classicIncrementalFormulation()) {
    if (gradJb_ == 0) {
      gradJb_.reset(new CtrlInc_(J_.jb()));
    } else {
      gradJb_.reset(new CtrlInc_(J_.jb().resolution(), *gradJb_));
    }
  } else {
    if (dxh0_ == 0) {
      dxh0_.reset(new CtrlInc_(J_.jb()));
    } else {
      dxh0_.reset(new CtrlInc_(J_.jb().resolution(), *dxh0_));
    }
  }

  Log::info() << std::endl;
  Log::info() << classname() << ": max iter = " << ninner 
              <<  ", requested norm reduction = " << gnreduc << std::endl;

// Define the matrices
  const Bmat_    B(J_);
  const HtRinvH_ HtRinvH(J_, runOnlineAdjTest);

// Compute H^T R^{-1} d
// dx_i = x_i - x_{i-1}; dx_1 = x_1 - x_b
  CtrlInc_ rhs(J_.jb());
  J_.computeGradientFG(rhs);

// Calculate first guess gradient
  if (J_.classicIncrementalFormulation()) {
    // -SUM dx' + H^T R^{-1} d
    J_.jb().addGradientFG(rhs, *gradJb_);
  } else {
    // H^T R^{-1} [d + H (x_b - x_fg)]
    const HtRinvHdxFG_ HtRinvHdxFG(J_);
    CtrlInc_ ht_rinv_h_dxfg(J_.jb());
    HtRinvHdxFG.compute(ht_rinv_h_dxfg);
    rhs += ht_rinv_h_dxfg;
  }

  Log::info() << classname() << " rhs" << rhs << std::endl;

// Define minimisation starting point
  // dx
  CtrlInc_ * dx = new CtrlInc_(J_.jb());
  // dxh = B^{-1} dx
  CtrlInc_ dxh(J_.jb());
  if (!J_.classicIncrementalFormulation()) {
    dxh = *dxh0_;
    B.multiply(dxh, *dx);
  }

// Set J[0] = 0.5 (x_i - x_b)^T B^{-1} (x_i - x_b) + 0.5 d^T R^{-1} d
  const double costJ0Jb = costJ0Jb_;
  const double costJ0JoJc = J_.getCostJoJc();

// Solve the linear system
  double reduc = this->solve(*dx, dxh, rhs, B, HtRinvH, costJ0Jb, costJ0JoJc, ninner, gnreduc);

  Log::test() << classname() << ": reduction in residual norm = " << reduc << std::endl;
  Log::info() << classname() << " output increment:" << *dx << std::endl;

// Update gradient Jb/Save inital guess for next minimization
  if (J_.classicIncrementalFormulation()) {
    *gradJb_ += dxh;
    dxh_.push_back(dxh);
  } else {
    *dxh0_ = dxh;
  }

// Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
  if (J_.classicIncrementalFormulation()) {
    costJ0Jb_ += 0.5 * dot_product(*dx, dxh);
    for (unsigned int jouter = 1; jouter < dxh_.size(); ++jouter) {
      CtrlInc_ dxhtmp(dx->geometry(), dxh_[jouter-1]);
      costJ0Jb_ += dot_product(*dx, dxhtmp);
    }
  } else {
    costJ0Jb_ = 0.5 * dot_product(*dx, dxh);
  }

  return dx;
}

// -----------------------------------------------------------------------------

template<typename MODEL> 
void DRMinimizer<MODEL>::checkpoint(eckit::LocalConfiguration & conf) const {
  Log::info() << classname() << "::checkpoint starting" << std::endl;

  // GradJb
  ASSERT(conf.has("gradJb"));
  eckit::LocalConfiguration gradjbconf(conf, "gradJb");
  gradJb_->write(gradjbconf);
  Log::info() << classname() << "::checkpoint gradJb: " 
              << util::full_precision(dot_product(*gradJb_, *gradJb_)) 
              << std::endl;
  // dxh
  ASSERT(conf.has("dxh"));
  std::vector<eckit::LocalConfiguration> dxhconfs;
  conf.get("dxh", dxhconfs);
  ASSERT(dxhconfs.size() == dxh_.size());
  for (unsigned int jj = 0; jj < dxhconfs.size(); ++jj) {
    dxh_[jj].write(dxhconfs[jj]);
    Log::info() << classname() << "::checkpoint dxh[" << jj << "]: " 
                << util::full_precision(dot_product(dxh_[jj], dxh_[jj]))
                << std::endl;
  }
  // costJ0Jb
  conf.set("costJ0Jb", util::full_precision(costJ0Jb_));
  Log::info() << classname() << "::checkpoint costJ0Jb: " 
              << util::full_precision(costJ0Jb_) << std::endl;

  Log::info() << classname() << "::checkpoint done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void DRMinimizer<MODEL>::restart(const eckit::Configuration & conf) {
  Log::info() << classname() << "::restart starting" << std::endl;

  // GradJb
  ASSERT(conf.has("gradJb"));
  eckit::LocalConfiguration gradjbconf(conf, "gradJb");
  if (gradJb_ == 0) {
    gradJb_.reset(new CtrlInc_(J_.jb()));
  } 
  gradJb_->read(gradjbconf);
  Log::info() << classname() << "::restart gradJb: " 
              << util::full_precision(dot_product(*gradJb_, *gradJb_)) 
              << std::endl;
  // dxh
  ASSERT(conf.has("dxh"));
  std::vector<eckit::LocalConfiguration> dxhconfs;
  conf.get("dxh", dxhconfs);
  for (unsigned int jj = 0; jj < dxhconfs.size(); ++jj) {
    CtrlInc_ dxh(J_.jb());
    dxh.read(dxhconfs[jj]);
    dxh_.push_back(dxh);
    Log::info() << classname() << "::restart dxh[" << jj << "]: " 
                << util::full_precision(dot_product(dxh_[jj], dxh_[jj]))
                << std::endl;
  }
  // costJ0Jb
  ASSERT(conf.has("costJ0Jb"));
  costJ0Jb_ = conf.getDouble("costJ0Jb");
  Log::info() << classname() << "::restart costJ0Jb: " 
              << util::full_precision(costJ0Jb_) << std::endl;

  Log::info() << classname() << "::restart done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_DRMINIMIZER_H_
