/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_INCRCTLVEC_H_
#define OOPS_ASSIMILATION_INCRCTLVEC_H_

#include <string>

#include <boost/ptr_container/ptr_vector.hpp>

#include "oops/assimilation/CostJbState.h"
#include "oops/base/EnsemblesCollection.h"
#include "oops/interface/IncrModCtlVec.h"
#include "oops/interface/IncrEnsCtlVec.h"

#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

  /// IncrementCtlVec Class
  /*!
   *  Part of the control vector corresponding to the increment part of the  
   *  control increment defined for a given time-step/sub-window. It has two 
   *  components, IncrModCtlVec and IncrEnsCtlVec acted upon by the static and 
   *  localized ensemble background error covariance, respectively.
   */

// -----------------------------------------------------------------------------

template <typename MODEL>
class IncrCtlVec : public util::Printable,
                   private util::ObjectCounter<IncrCtlVec<MODEL> > {
  typedef EnsemblesCollection<MODEL> EnsemblesCollection_;
  typedef IncrEnsCtlVec<MODEL>       IncrEnsCtlVec_;
  typedef IncrModCtlVec<MODEL>       IncrModCtlVec_;
  typedef CostJbState<MODEL>         JbState_;

 public:
  static const std::string classname() {return "oops::IncrCtlVec";}

/// Constructor, destructor
  explicit IncrCtlVec(const JbState_ &, const util::DateTime &);
  IncrCtlVec(const JbState_ &, const IncrCtlVec &);
  IncrCtlVec(const IncrCtlVec &, const bool copy = true);
  virtual ~IncrCtlVec();

/// Interfacing
  IncrModCtlVec_ & modCtlVec() {return *incrModCtlVec_;}
  const IncrModCtlVec_ & modCtlVec() const {return *incrModCtlVec_;}
  IncrEnsCtlVec_ & ensCtlVec(const int ii) {return incrEnsCtlVecCollection_[ii];}
  const IncrEnsCtlVec_ & ensCtlVec(const int ii) const {return incrEnsCtlVecCollection_[ii];}

/// Time
  const util::DateTime validTime() const {return time_;}

/// Linear algebra operators
  void zero();
  IncrCtlVec & operator =(const IncrCtlVec &);
  IncrCtlVec & operator+=(const IncrCtlVec &);
  IncrCtlVec & operator-=(const IncrCtlVec &);
  IncrCtlVec & operator*=(const double &);
  void axpy(const double &, const IncrCtlVec &, const bool check = true);
  double dot_product_with(const IncrCtlVec &) const;
  void schur_product_with(const IncrCtlVec &);
  void random();

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

 private:
  IncrModCtlVec_ * newIncrModCtlVec(const JbState_ &) const;
  boost::ptr_vector<IncrEnsCtlVec_> newIncrEnsCtlVecCollection(const JbState_ &) const;

  void print(std::ostream &) const;

  util::DateTime time_;
  unsigned int ensSize_;
  boost::scoped_ptr<IncrModCtlVec_> incrModCtlVec_;
  boost::ptr_vector<IncrEnsCtlVec_> incrEnsCtlVecCollection_;
};

// =============================================================================
/// Constructor, destructor
// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL>::IncrCtlVec(const JbState_ & jb, const util::DateTime & time) 
  : time_(time), ensSize_(EnsemblesCollection_::getInstance().size()), 
    incrModCtlVec_(0), incrEnsCtlVecCollection_(0)
{
  Log::trace() << "IncrCtlVec<MODEL>::IncrCtlVec starting" << std::endl;
  util::Timer timer(classname(), "Incr");

  incrModCtlVec_.reset(this->newIncrModCtlVec(jb));
  
  incrEnsCtlVecCollection_ = this->newIncrEnsCtlVecCollection(jb);

  Log::trace() << "IncrCtlVec<MODEL>::IncrCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL>::IncrCtlVec(const JbState_ & jb, 
                              const IncrCtlVec & other)
  : time_(other.time_), ensSize_(other.ensSize_), 
    incrModCtlVec_(0), incrEnsCtlVecCollection_(0)
{
  Log::trace() << "IncrCtlVec<MODEL>::IncrCtlVec starting" << std::endl;
  util::Timer timer(classname(), "Incr");

  // Create new control vectors at the resolution defined by jb
  incrModCtlVec_.reset(this->newIncrModCtlVec(jb));
  *incrModCtlVec_ = *other.incrModCtlVec_;

  incrEnsCtlVecCollection_ = this->newIncrEnsCtlVecCollection(jb);
  for (unsigned int jj = 0; jj < ensSize_; ++jj) {
    incrEnsCtlVecCollection_[jj] = other.incrEnsCtlVecCollection_[jj];
  }
  
  Log::trace() << "IncrCtlVec<MODEL>::IncrCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL>::IncrCtlVec(const IncrCtlVec & other, 
                              const bool copy) 
  : time_(other.time_), ensSize_(other.ensSize_), 
    incrModCtlVec_(0), incrEnsCtlVecCollection_(0)
{
  Log::trace() << "IncrCtlVec<MODEL>::IncrCtlVec copy starting" << std::endl;
  util::Timer timer(classname(), "Incr");

  incrModCtlVec_.reset(new IncrModCtlVec_(*other.incrModCtlVec_, copy));

  for (unsigned int jj = 0; jj < ensSize_; ++jj) {
    incrEnsCtlVecCollection_.push_back(new IncrEnsCtlVec_(other.incrEnsCtlVecCollection_[jj], copy));
  }

  Log::trace() << "IncrCtlVec<MODEL>::IncrCtlVec copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL>::~IncrCtlVec() {
  Log::trace() << "IncrCtlVec<MODEL>::~IncrCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~Incr");

  incrModCtlVec_.reset();

  Log::trace() << "IncrCtlVec<MODEL>::~IncrCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrCtlVec<MODEL>::zero() {
  Log::trace() << "IncrCtlVec<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");

  incrModCtlVec_->zero();

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      incrEnsCtlVecCollection_[jj].zero();
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL> & IncrCtlVec<MODEL>::operator=(const IncrCtlVec & rhs) {
  Log::trace() << "IncrCtlVec<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");

  *incrModCtlVec_ = *rhs.incrModCtlVec_;

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      incrEnsCtlVecCollection_[jj] = rhs.incrEnsCtlVecCollection_[jj];
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL> & IncrCtlVec<MODEL>::operator+=(const IncrCtlVec & rhs) {
  Log::trace() << "IncrCtlVec<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");

  *incrModCtlVec_ += *rhs.incrModCtlVec_;

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      incrEnsCtlVecCollection_[jj] += rhs.incrEnsCtlVecCollection_[jj];
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::operator+= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL> & IncrCtlVec<MODEL>::operator-=(const IncrCtlVec & rhs) {
  Log::trace() << "IncrCtlVec<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");

  *incrModCtlVec_ -= *rhs.incrModCtlVec_;

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      incrEnsCtlVecCollection_[jj] -= rhs.incrEnsCtlVecCollection_[jj];
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::operator-= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrCtlVec<MODEL> & IncrCtlVec<MODEL>::operator*=(const double & zz) {
  Log::trace() << "IncrCtlVec<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");

  *incrModCtlVec_ *= zz;

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      incrEnsCtlVecCollection_[jj] *= zz; 
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::operator*= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrCtlVec<MODEL>::axpy(const double & zz, const IncrCtlVec & dv, 
                             const bool check) {
  Log::trace() << "IncrCtlVec<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");

  incrModCtlVec_->axpy(zz, *dv.incrModCtlVec_, check);

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      incrEnsCtlVecCollection_[jj].axpy(zz, dv.incrEnsCtlVecCollection_[jj], check);
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::axpy done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double IncrCtlVec<MODEL>::dot_product_with(const IncrCtlVec & dv) const {
  Log::trace() << "IncrCtlVec<MODEL>::dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "dot_product_with");

  double zz = incrModCtlVec_->dot_product_with(*dv.incrModCtlVec_);

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      zz += incrEnsCtlVecCollection_[jj].dot_product_with(dv.incrEnsCtlVecCollection_[jj]);
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::dot_product_with done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrCtlVec<MODEL>::schur_product_with(const IncrCtlVec & dv) {
  Log::trace() << "IncrCtlVec<MODEL>::schur_product_with starting" << std::endl;
  util::Timer timer(classname(), "schur_product_with");

  incrModCtlVec_->schur_product_with(*dv.incrModCtlVec_);

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      incrEnsCtlVecCollection_[jj].schur_product_with(dv.incrEnsCtlVecCollection_[jj]);
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::schur_product_with done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrCtlVec<MODEL>::random() {
  Log::trace() << "IncrCtlVec<MODEL>::random starting" << std::endl;
  util::Timer timer(classname(), "random");

  incrModCtlVec_->random();

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      incrEnsCtlVecCollection_[jj].random();
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::random done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrCtlVec<MODEL>::read(const eckit::Configuration & conf) {
  Log::trace() << "IncrCtlVec<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");

  incrModCtlVec_->read(conf);
  //incrEnsCtlVec_->read(conf);

  Log::trace() << "IncrCtlVec<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrCtlVec<MODEL>::write(const eckit::Configuration & conf) const {
  Log::trace() << "IncrCtlVec<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");

  incrModCtlVec_->write(conf);
  //incrEnsCtlVec_->write(conf);

  Log::trace() << "IncrCtlVec<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double IncrCtlVec<MODEL>::norm() const {
  Log::trace() << "IncrCtlVec<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");

  double zz = incrModCtlVec_->norm();

  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      zz += incrEnsCtlVecCollection_[jj].norm();
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrCtlVec<MODEL>::print(std::ostream & os) const {
  Log::trace() << "IncrCtlVec<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");

  os << *incrModCtlVec_;
  if (ensSize_ != 0) {
    for (unsigned int jj = 0; jj < ensSize_; ++jj) {
      os << incrEnsCtlVecCollection_[jj] << std::endl;
    }
  }

  Log::trace() << "IncrCtlVec<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL> * IncrCtlVec<MODEL>::newIncrModCtlVec(const JbState_ & jb) const {
  return jb.covar().newIncrModCtlVec();
} 

// -----------------------------------------------------------------------------

template<typename MODEL>
boost::ptr_vector<IncrEnsCtlVec<MODEL> > IncrCtlVec<MODEL>::newIncrEnsCtlVecCollection(const JbState_ & jb) const {
  unsigned int ensSize = EnsemblesCollection_::getInstance().size();
  boost::ptr_vector<IncrEnsCtlVec<MODEL> > incrEnsCtlVecCollection;
  for (unsigned int jj = 0; jj < ensSize; ++jj) {
    incrEnsCtlVecCollection.push_back(jb.covar().newIncrEnsCtlVec());
  }
  return incrEnsCtlVecCollection;
}

// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_INTERFACE_INCRCtlVEC_H_
