/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_
#define OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_

#include <vector>
#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "eckit/config/Configuration.h"

#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/Minimizer.h"
#include "oops/base/PostProcessor.h"
#include "oops/base/StateInfo.h"
#include "oops/base/StateSaver.h"
#include "oops/interface/State.h"

#include "util/Logger.h"
#include "util/LogbookWriter.h"
#include "util/Restart.h"

namespace oops {

template<typename MODEL> 
class IncrementalAssimilation : private boost::noncopyable {
  typedef ControlIncrement<MODEL>    CtrlInc_;
  typedef ControlVariable<MODEL>     CtrlVariable_;
  typedef Minimizer<MODEL>           Minimizer_;
  typedef State<MODEL>               State_;
 public:
  static const std::string classname() {return "oops::IncrementalAssimilation";}

/// Constructor, destructor
  IncrementalAssimilation(CostFunction<MODEL> &);
  ~IncrementalAssimilation() {}

/// Perform assimilation
  boost::shared_ptr<CtrlVariable_> assimilate(const eckit::Configuration &);

 private:
  void checkpoint(eckit::LocalConfiguration & conf) const;
  void restart(const eckit::Configuration & conf) const;
  void connectToRestart() const;

  CostFunction<MODEL> & J_;
  boost::shared_ptr<ControlVariable<MODEL> > xx_;
  boost::scoped_ptr<Minimizer<MODEL> > minim_;

  size_t jouter_;
  mutable size_t jstart_;
};

// =============================================================================

template<typename MODEL>
IncrementalAssimilation<MODEL>::IncrementalAssimilation(CostFunction<MODEL> & J)
  : J_(J), xx_(), minim_(), jouter_(0), jstart_(0)
{
  Log::trace() << "IncrementalAssimilation:constructed" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
boost::shared_ptr<ControlVariable<MODEL> > 
  IncrementalAssimilation<MODEL>::assimilate(const eckit::Configuration & config) {

//  Initialize first guess from background
  xx_.reset(new CtrlVariable_(J_.jb().getBackground()));
  Log::trace() << "IncrementalAssimilation: first guess has been set up" << std::endl;

// Setup outer loop
  std::vector<eckit::LocalConfiguration> iterconfs;
  config.get("variational.iteration", iterconfs);
  size_t nouter = iterconfs.size();
  Log::info() << "Running incremental assimilation with " << nouter 
              << " outer iterations." << std::endl;

// Setup minimizer
  eckit::LocalConfiguration minConf(config, "minimizer");
  const size_t nnout = nouter;
  minConf.set("nouter", nnout);
  minim_.reset(MinFactory<MODEL>::create(minConf, J_));

// Restart mechanism 
  this->connectToRestart();
  util::Restart::setup({"FirstGuess","CostFunction","Minimizer","IncrementalAssimilation"});
  util::Restart::restart({"FirstGuess","CostFunction","IncrementalAssimilation"});

  for (jouter_ = jstart_; jouter_ < nouter; ++jouter_) {
    util::LogbookWriter<size_t> log("OuterLoop", jouter_); 

//  Get configuration for current outer iteration
    Log::info() << "IncrementalAssimilation: Configuration for outer iteration "
              << jouter_ << ":\n" << iterconfs[jouter_];

//  Setup for the trajectory run
    PostProcessor<State_> post;
    // Setup diagnostic prints
    if (iterconfs[jouter_].has("prints")) {
      const eckit::LocalConfiguration prtConfig(iterconfs[jouter_], "prints");
      post.enrollProcessor(new StateInfo<State_>("traj", prtConfig));
    }
    // Setup state saver
    boost::shared_ptr< StateSaver<State_> > stateSaver;
    if (iterconfs[jouter_].has("4DAnalysisFields")) {
      const eckit::LocalConfiguration config(iterconfs[jouter_], "4DAnalysisFields");
      const eckit::LocalConfiguration configLate4d(config, "late4d");
      stateSaver.reset(new StateSaver<State_>(configLate4d));
      post.enrollProcessor(stateSaver);
    }

//  Setup quadratic problem
    J_.linearize(*xx_, iterconfs[jouter_], post);

//  Restart minimizer
    if (jouter_ == jstart_)
      util::Restart::restart({"Minimizer"});

//  Minimization
    boost::scoped_ptr<CtrlInc_> dx(minim_->minimize(iterconfs[jouter_]));

//  Compute analysis in physical space
    if (!J_.classicIncrementalFormulation() && jouter_ != 0)
      // Reset fg if computing increment with respect to background
      xx_.reset(new CtrlVariable_(J_.jb().getBackground()));

    J_.addIncrement(*xx_, *dx);

//  Compute and save 4D analysis fields    
    if (iterconfs[jouter_].has("4DAnalysisFields")) {
      // Write 4v fields
      const eckit::LocalConfiguration config(iterconfs[jouter_], "4DAnalysisFields");
      const eckit::LocalConfiguration configAnalysis4v(config, "analysis4v");
      xx_->write(configAnalysis4v);

      // Calculate and write an fields
      const eckit::LocalConfiguration configLate4d(config, "late4d");
      J_.compute4DAnalysisFields(configLate4d, stateSaver, *dx);
      // Clean-up
      stateSaver.reset();
    }

//  Clean-up trajectory, etc...
    J_.resetLinearization();

//  Checkpointing
    util::Restart::checkpoint(iterconfs[jouter_], {"FirstGuess","CostFunction",
                                                   "Minimizer", "IncrementalAssimilation"});
  }
  return xx_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrementalAssimilation<MODEL>::checkpoint(eckit::LocalConfiguration & conf) const {
  conf.set("Iteration", jouter_);
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrementalAssimilation<MODEL>::restart(const eckit::Configuration & conf) const {
  jstart_ = conf.getUnsigned("Iteration");
  ++jstart_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrementalAssimilation<MODEL>::connectToRestart() const {
// Checkpoint connect elements
  util::Restart::connectChkptCallback("CostFunction", 
                                      [this](eckit::LocalConfiguration & conf) 
                                      { return J_.checkpoint(conf); });
  util::Restart::connectChkptCallback("Minimizer", 
                                      [this](eckit::LocalConfiguration & conf) 
                                      { return minim_->checkpoint(conf); });
  util::Restart::connectChkptCallback("FirstGuess", 
                                      [this](eckit::LocalConfiguration & conf) 
                                      { return xx_->write(conf); });
  util::Restart::connectChkptCallback("IncrementalAssimilation", 
                                      [this](eckit::LocalConfiguration & conf) 
                                      { return this->checkpoint(conf); });

// Restart connect elements
  util::Restart::connectRstrtCallback("CostFunction", 
                                      [this](const eckit::Configuration & conf) 
                                      { return J_.restart(conf); });
  util::Restart::connectRstrtCallback("Minimizer", 
                                      [this](const eckit::Configuration & conf) 
                                      { return minim_->restart(conf); });
  util::Restart::connectRstrtCallback("FirstGuess", 
                                      [this](const eckit::Configuration & conf) 
                                      { return xx_->read(conf); });
  util::Restart::connectRstrtCallback("IncrementalAssimilation",
                                      [this](const eckit::Configuration & conf) 
                                      { return this->restart(conf); });
}

// =============================================================================

}  // namespace oops

#endif  // OOPS_ASSIMILATION_INCREMENTALASSIMILATION_H_
