/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_QNEWTONSQRTLMP_H_
#define OOPS_ASSIMILATION_QNEWTONSQRTLMP_H_

#include <algorithm>
#include <string>
#include <vector>

#include <boost/shared_ptr.hpp>

#include "eckit/config/LocalConfiguration.h"
#include "util/Logger.h"
#include "util/dot_product.h"

namespace oops {

  /*  The QUASI-NEWTON  Limited Memory preconditioner matrix \f$ C \f$
   *  for the system matrix \f$ AB = (I + Ht Rinv H B) \f$
   *
   *  It is defined as
   *  \f$ C_(k+1)  = (I - rho_k rr_k q_k^T) C_k  (I - rho_k q_k p_k^T) + rho_k rr_k p_k^T\f$
   *
   *  For details, we refer to S. Gratton, A. Sartenaer and J. Tshimanga,
   *  SIAM J. Optim.,21(3),912–935,2011 and S. Gurol, PhD Manuscript, 2013.
   *
   *  The solvers represent matrices as objects that implement a "multiply"
   *  method. This class defines objects that apply the preconditioner matrix \f$ P \f$.
   */

// -----------------------------------------------------------------------------

template<typename MODEL> class QNewtonSqrtLMP {
  typedef CostFunction<MODEL>     CostFct_;
  typedef ControlVector<MODEL>    CtrlVec_;
 public:
  explicit QNewtonSqrtLMP(const eckit::Configuration &, const CostFct_ &);
  ~QNewtonSqrtLMP() {}

  void push(const CtrlVec_ &, const CtrlVec_ &, const CtrlVec_ &, const double &);
  void update();
  void changeResolution();

  void inverseMultiplySqrt(const CtrlVec_ &, CtrlVec_ &) const;
  void inverseMultiply(const CtrlVec_ &, CtrlVec_ &) const;


 private:
  const CostFct_ & J_;
  const bool active_;
  unsigned maxpairs_;
  unsigned maxnewpairs_;
  bool useoldpairs_;
  int maxouter_;
  int update_;

  std::vector<boost::shared_ptr<CtrlVec_> > P_;
  std::vector<boost::shared_ptr<CtrlVec_> > R_;
  std::vector<boost::shared_ptr<CtrlVec_> > AP_;
  std::vector<double> rhos_;
  std::vector<unsigned> usedpairIndx_;

  std::vector<CtrlVec_> savedP_;
  std::vector<CtrlVec_> savedR_;
  std::vector<CtrlVec_> savedAP_;
  std::vector<double> savedrhos_;
};

// =============================================================================

template<typename MODEL>
QNewtonSqrtLMP<MODEL>::QNewtonSqrtLMP(const eckit::Configuration & conf,
                                      const CostFct_ & J)
  : J_(J), active_(conf.has("preconditioner")), maxpairs_(0), maxnewpairs_(0),
    useoldpairs_(false), maxouter_(0), update_(1)
{
  if (active_) {
    maxouter_ = conf.getInt("nouter");
    Log::info() << "QNewtonSqrtLMP: maxouter : " << maxouter_ << std::endl;

    const eckit::LocalConfiguration precond(conf, "preconditioner");
    if (precond.has("maxpairs")) maxpairs_ = precond.getInt("maxpairs");
    if (maxpairs_ > 0) {
      std::string old;
      if (precond.has("useoldpairs")) old = precond.getString("useoldpairs");
      useoldpairs_ = (old == "on" || old == "true");
      if (useoldpairs_ && precond.has("maxnewpairs")) {
          maxnewpairs_ = precond.getInt("maxnewpairs");
          maxnewpairs_ = std::min(maxpairs_, maxnewpairs_);
      } else {
       maxnewpairs_ = maxpairs_;
      }
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void QNewtonSqrtLMP<MODEL>::push(const CtrlVec_ & p, const CtrlVec_ & ph,
                                 const CtrlVec_ & ap, const double & rho) {
  if (active_) {
    ASSERT(savedP_.size() <= maxnewpairs_);
    if (maxnewpairs_ > 0 && update_ < maxouter_) {
      if (savedP_.size() == maxnewpairs_) {
        savedP_.erase(savedP_.begin());
        savedR_.erase(savedR_.begin());
        savedAP_.erase(savedAP_.begin());
        savedrhos_.erase(savedrhos_.begin());
      }
      savedP_.push_back(p);
      savedR_.push_back(ph);
      savedAP_.push_back(ap);
      savedrhos_.push_back(1.0/rho);
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void QNewtonSqrtLMP<MODEL>::update() {
  if (active_) {
    Log::debug() << "QNewtonLMP size saved Ph    = " << savedR_.size() << std::endl;
    Log::debug() << "QNewtonLMP size saved P     = " << savedP_.size() << std::endl;
    Log::debug() << "QNewtonLMP size saved AP    = " << savedAP_.size() << std::endl;
    Log::debug() << "QNewtonLMP size saved rhos  = " << savedrhos_.size() << std::endl;

    const unsigned nvec = savedR_.size();
    ASSERT(savedP_.size() == nvec);
    ASSERT(savedAP_.size() == nvec);
    ASSERT(savedrhos_.size() == nvec);

    if (!useoldpairs_ || nvec >= maxpairs_) {
      R_.clear();
      P_.clear();
      AP_.clear();
      rhos_.clear();
      for (unsigned kiter = 0; kiter < usedpairIndx_.size(); ++kiter) {
        usedpairIndx_[kiter] = 0;
      }
    }

    if (nvec > 0 && update_ < maxouter_) {
      Log::info() << "QNewtonLMP: update " << update_ << ", max = " << maxouter_-1 << std::endl;
      unsigned newpairs = std::min(nvec, maxpairs_);
      unsigned oldpairs = P_.size();
      unsigned rmpairs = 0;
//    First remove pairs we no longer need.
      if (oldpairs + newpairs > maxpairs_) {
        rmpairs = oldpairs + newpairs - maxpairs_;
        for (unsigned jv = 0; jv < rmpairs; ++jv) {
          R_.erase(R_.begin());
          P_.erase(P_.begin());
          AP_.erase(AP_.begin());
          rhos_.erase(rhos_.begin());
        }
        oldpairs -= rmpairs;
//      Keep information on how many pairs are used at each minimization loop
        unsigned removed = rmpairs;
        for (unsigned kiter = 0; kiter < usedpairIndx_.size(); ++kiter) {
          unsigned rmiter = std::min(usedpairIndx_[kiter], removed);
          usedpairIndx_[kiter] -= rmiter;
          removed -= rmiter;
        }
        ASSERT(removed == 0);
      }
      ASSERT(P_.size() == oldpairs);
      ASSERT(oldpairs + newpairs <= maxpairs_);

//    Add the new pairs.
      for (unsigned jv = nvec - newpairs; jv < nvec; ++jv) {
        P_.push_back(boost::shared_ptr<CtrlVec_>(new CtrlVec_(savedP_[jv])));
        R_.push_back(boost::shared_ptr<CtrlVec_>(new CtrlVec_(savedR_[jv])));
        AP_.push_back(boost::shared_ptr<CtrlVec_>(new CtrlVec_(savedAP_[jv])));
        rhos_.push_back(savedrhos_[jv]);
      }
      ASSERT(P_.size() == oldpairs + newpairs);

      Log::info() << "Number of inner iterations       : " << nvec << std::endl;
      Log::info() << "Number of maximum pairs          : " << maxpairs_ << std::endl;
      Log::info() << "Number of used total pairs       : " << oldpairs + newpairs << std::endl;
      Log::info() << "Number of new pairs              : " << newpairs << std::endl;
      Log::info() << "Number of removed old pairs      : " << rmpairs << std::endl;
      for (unsigned kiter = 0; kiter < usedpairIndx_.size(); ++kiter) {
        Log::info() << "Number of used pairs from outer loop " << kiter + 1
                  << " : "                                   << usedpairIndx_[kiter];
      }
    }

    ++update_;
    savedP_.clear();
    savedR_.clear();
    savedAP_.clear();
    savedrhos_.clear();
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void QNewtonSqrtLMP<MODEL>::changeResolution() {
  if (active_) {
    Log::info() << "QNewtonSqrtLMP:::changeResolution starting" << std::endl;
    for (auto & pp: P_)  pp.reset(new CtrlVec_(J_.jb(), *pp));
    for (auto & rr: R_)  rr.reset(new CtrlVec_(J_.jb(), *rr));
    for (auto & ap: AP_) ap.reset(new CtrlVec_(J_.jb(), *ap));
    Log::info() << "QNewtonSqrtLMP:::changeResolution done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void QNewtonSqrtLMP<MODEL>::inverseMultiplySqrt(const CtrlVec_ & a, CtrlVec_ & b) const {
  b = a;
  const unsigned nvec = P_.size();
  if (nvec != 0) {
    for (unsigned iiter = 0; iiter < nvec; ++iiter) {
      double factor = sqrt(rhos_[iiter])*dot_product(*R_[iiter], b)/sqrt(dot_product(*R_[iiter], *R_[iiter]));
      factor += rhos_[iiter]*dot_product(*AP_[iiter],b);
      b.axpy(-factor, *P_[iiter]);
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void QNewtonSqrtLMP<MODEL>::inverseMultiply(const CtrlVec_ & a, CtrlVec_ & b) const {
  b = a;
  const unsigned nvec = P_.size();
  if (nvec != 0) {
    std::vector<double> etas(nvec, 0);
    //etas.clear();
    for (int iiter = nvec-1; iiter >= 0; iiter--) {
      double eta = dot_product(b, *P_[iiter]);
      eta *= rhos_[iiter];
      etas[iiter]=eta;
      b.axpy(-eta, *AP_[iiter]);
    }
    b *= dot_product(*AP_[nvec-1], *P_[nvec-1])/dot_product(*AP_[nvec-1], *AP_[nvec-1]);
    for (unsigned iiter = 0; iiter < nvec; ++iiter) {
      double sigma = dot_product(b, *AP_[iiter]);
      sigma *= rhos_[iiter];
      sigma -= etas[iiter];
      b.axpy(-sigma, *P_[iiter]);
    }
  }
}


// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_QNEWTONSQRTLMP_H_
