/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_SQRTBPCGMINIMIZER_H_
#define OOPS_ASSIMILATION_SQRTBPCGMINIMIZER_H_

#include <cmath>
#include <string>
#include <vector>

#include "eckit/config/Configuration.h"
#include "util/Logger.h"
#include "util/LogbookWriter.h"
#include "util/dot_product.h"
#include "util/formats.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/JbMatrix.h"
#include "oops/assimilation/SQRTMinimizer.h"
#include "oops/assimilation/QNewtonSqrtLMP.h"
#include "oops/assimilation/UtHtRinvHUMatrix.h"

namespace oops {

/// SQRTBPCG Minimizer
/*!
 * \brief Preconditioned Conjugate Gradients solver for the square root preconditioned
 * formulation of the Hessian
 *
 * This solver is based on the standard Preconditioned Conjugate
 * Gradients solver for Ax=b (G. H. Golub and C. F. Van Loan, Matrix
 * Computations).
 *
 * A must be square, symmetric, positive definite.

 * On entry:
 * -    dv      =  starting point, \f$ dv_{0} \f$.
 * -    rr      = \f$ (sum dv^{b}_{i} + ) U^T H^T R^{-1} d \f$

 * On exit, dv will contain the solution \f$ U^{-1} dx \f$ 
 *  The solution \f$ dx = U dv \f$ is recovered in the SQRTMinimizer class
 *  The return value is the achieved reduction in preconditioned residual norm.
 *
 *  Iteration will stop if the maximum iteration limit "maxiter" is reached
 *  or if the residual norm reduces by a factor of "tolerance".
 *
 */

// -----------------------------------------------------------------------------

template<typename MODEL> class SQRTBPCGMinimizer : public SQRTMinimizer<MODEL> {
  typedef BMatrix<MODEL>          Bmat_;
  typedef CostFunction<MODEL>     CostFct_;
  typedef ControlVector<MODEL>    CtrlVec_;
  typedef JbMatrix<MODEL>         Jbmat_;
  typedef UtHtRinvHUMatrix<MODEL> UtHtRinvHU_;

 public:
  const std::string classname() const final {return "SQRTBPCGMinimizer";}

/// Constructor, destructor
  SQRTBPCGMinimizer(const eckit::Configuration &, const CostFct_ &);
  ~SQRTBPCGMinimizer() {}

 private:
  double solve(CtrlVec_ &, CtrlVec_ &, const Jbmat_ &, const UtHtRinvHU_ &,
               const bool &, const size_t &, const size_t &, const double &,
               const double &, const std::vector<double> &, const double &) final;

/// Checkpointing, restart of the limited memory preconditioner
  void checkpointLMP(eckit::LocalConfiguration &) const final;
  void restartLMP(const eckit::Configuration &) final;

/// Quadratic cost function calculations
  void setupQuadCost(const double & costJ0Jb, 
                     const std::vector<double> & costJ0JbObsAux,
                     const double & costJ0JoJc);
  void calcQuadCost(const CtrlVec_ &, const CtrlVec_ &, const CtrlVec_ &);
  void printQuadCost(const CtrlVec_ &, const size_t &) const;

/// Data members
  double costJ0_;
  double costJ0Jb_;
  std::vector<double> costJ0JbObsAux_;
  double costJ0JoJc_;

  double costJ_;
  double costJb_;
  std::vector<double> costJbObsAux_;
  double costJoJc_;

  QNewtonSqrtLMP<MODEL> lmp_;
};

// =============================================================================

template<typename MODEL>
SQRTBPCGMinimizer<MODEL>::SQRTBPCGMinimizer(const eckit::Configuration & conf, const CostFct_ & J)
  : SQRTMinimizer<MODEL>(conf, J), costJ0_(0), costJ0Jb_(0), costJ0JbObsAux_(), costJ0JoJc_(0),
    costJ_(0), costJb_(0), costJbObsAux_(), costJoJc_(0), lmp_(conf, J)
{}

// -----------------------------------------------------------------------------

template<typename MODEL>
double SQRTBPCGMinimizer<MODEL>::solve(CtrlVec_ & dv, CtrlVec_ & rr,
                                   const Jbmat_ & Jb, const UtHtRinvHU_ & UtHtRinvHU,
                                   const bool & linfoconv,
                                   const size_t & maxiter,
                                   const size_t & miniter,
                                   const double & tolerance,
                                   const double & costJ0Jb, 
                                   const std::vector<double> & costJ0JbObsAux,
                                   const double & costJ0JoJc) {
  // dv   control vector
  // rr   (sum dv_i^{b} +) U^T G^T R^{-1} d

  CtrlVec_ qq(dv);
  CtrlVec_ pp(dv);
  CtrlVec_ pr(dv);
  CtrlVec_ jbpp(dv);
  CtrlVec_ jbdv(dv);
  CtrlVec_ zz(dv);
  CtrlVec_ r0(dv);
  CtrlVec_ ww(dv);

  // vectors for re-orthogonalization
  std::vector<CtrlVec_> vvecs;
  std::vector<CtrlVec_> zvecs;
  std::vector<double> scals;

  // setup calculation of the quadratic cost function
  this->setupQuadCost(costJ0Jb, costJ0JbObsAux, costJ0JoJc);

  // change resolution of LMP vectors
  lmp_.changeResolution();

  // r_{0}
  r0 = rr;

  // r_{0}^T r_{0}
  Log::info() << "normr0 " << dot_product(rr, rr) << std::endl;

  // z_{0} = LMP r_{0}
  lmp_.inverseMultiply(rr, zz);
  // p_{0} = z_{0}
  pp = zz;

  // r_{0}^T z_{0}
  double dotZZ0  = dot_product(rr, zz);
  Log::info() << "normr0z0 " << dotZZ0 << std::endl;
  double normReduction = 1.0;
  double zdots = dotZZ0;
  double zdots_old = 0.0;

  vvecs.push_back(rr);
  zvecs.push_back(zz);
  scals.push_back(1.0/dotZZ0);

  Log::info() << std::endl;
  for (size_t jiter = 0; jiter < maxiter; ++jiter) {
    Log::info() << " SQRTPCG Starting Iteration " << jiter+1 << std::endl;

    util::LogbookWriter<int> log("InnerLoop", jiter);

    if (jiter > 0) {
      // beta_{i} = z_{i+1}^T z_{i+1} / z_{i}^T z_{i}
      double beta = zdots/zdots_old;

      // p_{i+1} = z_{i+1} + beta*p_{i}
      pp *= beta;
      pp += zz;
    }

    // q_{i} = ( P^-1/2  Jb P^-T/2 + U^T H^T R^{-1} H U ) p_{i}
    UtHtRinvHU.multiply(pp, qq);
    Jb.multiply(pp, jbpp); // U^T B^-1 U != I for VarBC
    qq += jbpp;

    // alpha_{i} = z_{i}^T z_{i} / q_{i}^T p_{i}
    double rho = dot_product(pp, qq);
    double alpha = zdots/rho;

    // dv_{i+1} = dv_{i} + alpha * p_{i}
    dv.axpy(alpha, pp);

    // jbdv_{i+1} = jbdv_{i} + alpha * jbh_{i}
    jbdv.axpy(alpha, jbpp);
    
    // r_{i+1} = r_{i} - alpha * q_{i}
    rr.axpy(-alpha, qq);

    // Compute the quadratic cost function
    this->calcQuadCost(dv, jbdv, r0);

    // Re-orthogonalization
    for (size_t jj = 0; jj < jiter; ++jj) {
      double proj = scals[jj] * dot_product(rr, zvecs[jj]);
      rr.axpy(-proj, vvecs[jj]);
    }

    lmp_.inverseMultiply(rr, zz);

    // z_{i}^T z_{i}
    zdots_old = zdots;
    // z_{i+1}^T z_{i+1}
    zdots = dot_product(rr, zz);

    Log::info() << "zdots " << zdots << "      iteration    " << jiter << std::endl;

    // z_{i+1}^T z_{i+1} / z_{0}^T z_{0}
    normReduction = sqrt(zdots/dotZZ0);

    Log::info() << "SQRTPCG end of iteration " << jiter+1 << std::endl
                << "  Norm reduction ("        << std::setw(2) << jiter+1 << ") = "
                  << util::full_precision(normReduction) << std::endl; 

    this->printQuadCost(dv, jiter);

    if (normReduction < tolerance && jiter >= miniter) {
      Log::info() << "SQRTPCG: Achieved required reduction in residual norm." << std::endl;
      break;
    }

    vvecs.push_back(rr);
    zvecs.push_back(zz);
    scals.push_back(1.0/zdots);

    lmp_.push(pp, zz, qq, rho);
  }
  lmp_.update();

  return normReduction;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTBPCGMinimizer<MODEL>::setupQuadCost(const double & costJ0Jb,
                                            const std::vector<double> & costJ0JbObsAux,
                                            const double & costJ0JoJc) {
  // J0
  costJ0Jb_ = costJ0Jb;
  costJ0JbObsAux_.clear(); costJ0JbObsAux_ = costJ0JbObsAux;
  costJ0JoJc_ = costJ0JoJc;

  // J0
  costJ0_ = costJ0Jb_ + costJ0JoJc_;

  // reset
  costJ_ = 0;
  costJb_ = 0;
  costJbObsAux_.clear(); costJbObsAux_ = costJ0JbObsAux;
  costJoJc_ = 0;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTBPCGMinimizer<MODEL>::calcQuadCost(const CtrlVec_ & dv,
                                           const CtrlVec_ & jbdv, 
                                           const CtrlVec_ & r0) {
    // Compute the quadratic cost function
    // J[dv_{i}] = J[0] - 0.5 dv_{i}^T r_{0}
    costJ_ = costJ0_ - 0.5 * dot_product(dv, r0);
    Log::info() << "dvtro " << util::full_precision( dot_product(dv, r0) ) << std::endl;
    // Jb[dv_{i}] = 0.5 dv_{i}^T Jb.dv_{i}
    costJb_ = costJ0Jb_ + 0.5 * dot_product(dv, jbdv);
    // JbObsAux[dv_{i}.obsVar()]
    for (size_t jj = 0; jj < costJ0JbObsAux_.size(); ++jj) {
      const std::string & key = dv.obsVar().getKeys()[jj];
      costJbObsAux_[jj] = costJ0JbObsAux_[jj]
                      + 0.5 * dot_product(dv.obsVar().obsAuxCtlVec(key),
                                          jbdv.obsVar().obsAuxCtlVec(key));
    }
    // Jo[dv_{i}] + Jc[dv_{i}] = J[dv_{i}] - Jb[dv_{i}]
    costJoJc_ = costJ_ - costJb_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTBPCGMinimizer<MODEL>::printQuadCost(const CtrlVec_ & dv,
                                            const size_t & jiter) const {
  
  // Print the quadratic cost function
  Log::info() << "  Quadratic cost function: J   (" << std::setw(2) << jiter+1 << ") = "
                << util::full_precision(costJ_)         << std::endl
              << "  Quadratic cost function: Jb  (" << std::setw(2) << jiter+1 << ") = "
                << util::full_precision(costJb_)        << std::endl;
  for (size_t jj = 0; jj < costJ0JbObsAux_.size(); ++jj) {
    Log::info() << "  Quadratic cost function: Jb" << dv.obsVar().getKeys()[jj]
                  << "  (" << std::setw(2) << jiter+1 << ") = "
                  << util::full_precision(costJbObsAux_[jj])   << std::endl;
  }
  Log::info() << "  Quadratic cost function: JoJc(" << std::setw(2) << jiter+1 << ") = "
                << util::full_precision(costJoJc_)      << std::endl << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTBPCGMinimizer<MODEL>::checkpointLMP(eckit::LocalConfiguration &) const {
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTBPCGMinimizer<MODEL>::restartLMP(const eckit::Configuration &) {
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_SQRTBPCGMINIMIZER_H_
