/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_SQRTMINIMIZER_H_
#define OOPS_ASSIMILATION_SQRTMINIMIZER_H_

#include <iterator>
#include <map>
#include <string>
#include <vector>

#include <boost/scoped_ptr.hpp>

#include "eckit/config/Configuration.h"

#include "oops/assimilation/BMatrix.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/JbMatrix.h"
#include "oops/assimilation/UtHtRinvHdxFG.h"
#include "oops/assimilation/UtHtRinvHUMatrix.h"
#include "oops/assimilation/Minimizer.h"

#include "util/dot_product.h"
#include "util/formats.h"
#include "util/Logger.h"

namespace oops {

/// SQRT Minimizers
/*!
 * SQRTMinimizer is the base class for all minimizers that use \f$ U\f$ to
 * precondition the variational minimisation problem and use the auxiliary
 * variable \f$ dx = U \hat{v} \f$ or \f$ dx = U \hat{v} + x_b - x_fg \f$
 */

// -----------------------------------------------------------------------------

template<typename MODEL> class SQRTMinimizer : public Minimizer<MODEL> {
  typedef BMatrix<MODEL>           Bmat_;
  typedef CostFunction<MODEL>      CostFct_;
  typedef ControlIncrement<MODEL>  CtrlInc_;
  typedef ControlVector<MODEL>     CtrlVec_;
  typedef JbMatrix<MODEL>          Jbmat_;
  typedef UtHtRinvHUMatrix<MODEL>  UtHtRinvHU_;
  typedef UtHtRinvHdxFG<MODEL>     UtHtRinvHdxFG_;
  typedef Minimizer<MODEL>         Minimizer_;

 public:
  virtual const std::string classname() const override =0;

/// Constructor, destructor
  SQRTMinimizer(const eckit::Configuration & conf, const CostFct_ & J)
   : Minimizer_(J), J_(J), conf_(conf), linfoConv_(false), ninnerMin_(0), 
     gradJb_(0), dv0_(0), dvh_(), costJ0Jb_(0),
     costJ0JbObsAux_(J.jb().jbObsBias().size(), 0.0), setObsAuxFormulation_(false), 
     obsAuxFormulation_(), iter_(0), maxiter_(999) 
  {
     if (conf.has("linfoconv")) linfoConv_ = conf.getBool("linfoconv");
     if (conf.has("ninnermin")) ninnerMin_ = conf.getUnsigned("ninnermin");
     this->setupFormulation();
  }
  ~SQRTMinimizer() {}

/// Checkpointing, restart
  void checkpoint(eckit::LocalConfiguration &) const;
  void restart(const eckit::Configuration &);

 private:
  CtrlInc_ * doMinimize(const eckit::Configuration &) override;
  virtual double solve(CtrlVec_ &, CtrlVec_ &, 
                       const Jbmat_ &,
                       const UtHtRinvHU_ &, 
                       const bool &,
                       const size_t &,
                       const size_t &,
                       const double &, 
                       const double &,
                       const std::vector<double> &, 
                       const double &) =0;

/// Checkpointing, restart of the limited memory preconditioner
  virtual void checkpointLMP(eckit::LocalConfiguration &) const =0;
  virtual void restartLMP(const eckit::Configuration &) =0;

  void checkpointGradJb(eckit::LocalConfiguration &) const;
  void checkpointDv0(eckit::LocalConfiguration &) const;
  void checkpointDvh(eckit::LocalConfiguration &) const;
  void checkpointAux(eckit::LocalConfiguration &) const;

  void restartGradJb(const eckit::Configuration &);
  void restartDv0(const eckit::Configuration &);
  void restartDvh(const eckit::Configuration &);
  void restartAux(const  eckit::Configuration &);

/// Quadratic cost function calculation (Jb component)
  void updateJbCostFunction(const CtrlVec_ & dv, const CtrlVec_ & jbdv);

/// Formulation
  void setupFormulation();

/// ObsAux formulation
  void setObsAuxCtlVec(CtrlVec_ & dv) const;

/// Data members
  const CostFct_ & J_;
  const eckit::LocalConfiguration & conf_;

  bool linfoConv_;
  size_t ninnerMin_;

  boost::scoped_ptr<CtrlVec_> gradJb_;
  boost::scoped_ptr<CtrlVec_> dv0_;
  std::vector<CtrlVec_> dvh_;

  double costJ0Jb_;
  std::vector<double> costJ0JbObsAux_;

  mutable bool setObsAuxFormulation_;
  mutable std::map<std::string, bool> obsAuxFormulation_;

 protected:
  size_t iter_;
  mutable size_t maxiter_;
};

// =============================================================================

template<typename MODEL>
ControlIncrement<MODEL> * SQRTMinimizer<MODEL>::doMinimize(const eckit::Configuration & config) {
  iter_++;

  ASSERT(iter_ <= maxiter_);
  size_t ninner = config.getUnsigned("ninner");
  ninner < ninnerMin_ ? ninner = ninnerMin_ : ninner;
  double gnreduc = config.getDouble("gradient_norm_reduction");

  bool runOnlineAdjTest = false;
  if (config.has("onlineDiagnostics")) {
    const eckit::LocalConfiguration onlineDiag(config, "onlineDiagnostics");
    runOnlineAdjTest = onlineDiag.getBool("onlineAdjTest");
  }

// Interpolate gradJB/Set initial guess to the new resolution
  if (J_.classicIncrementalFormulation()) {
    if (gradJb_ == 0) {
      gradJb_.reset(new CtrlVec_(J_.jb()));
    } else {
      gradJb_.reset(new CtrlVec_(J_.jb(), *gradJb_));
    }
  } else {
    if (dv0_ == 0) {
      dv0_.reset(new CtrlVec_(J_.jb()));
    } else {
      dv0_.reset(new CtrlVec_(J_.jb(), *dv0_));
    }
  }

  Log::info() << std::endl;
  Log::info() << classname() << ": starting" << std::endl;

// Define the matrices
  const Jbmat_      Jb(J_);
  const Bmat_       B(J_);
  const UtHtRinvHU_ UtHtRinvHU(J_, runOnlineAdjTest);

// Compute U^T H^T R^{-1} d
  CtrlInc_ fgrhs(J_.jb().getFirstGuessIncrement(), false);
  J_.computeGradientFG(fgrhs);
  CtrlVec_ rhs(J_.jb());
  B.multiplySqrtTrans(fgrhs, rhs);

// Calculate first guess gradient
  if (J_.classicIncrementalFormulation()) {
    // -SUM dv_i + U^T H^T R^{-1} d 
    J_.jb().addGradientFG(rhs, *gradJb_);
  } else {
    // U^T H^T R^{-1} [d + H (x_b - x_fg)]
    const UtHtRinvHdxFG_ UtHtRinvHdxFG(J_);
    CtrlVec_ ut_ht_rinv_h_dxfg(J_.jb());
    UtHtRinvHdxFG.compute(ut_ht_rinv_h_dxfg);
    rhs += ut_ht_rinv_h_dxfg;
  }


// Set starting point of the minimization
  // dx_k = U dv_k
  CtrlVec_ dv(J_.jb());
  // or dx_k = U dv_k + x_b - x_fg; initialise dv_k to dv_{k-1}
  if (!J_.classicIncrementalFormulation())
    dv = *dv0_;

// Set J[0] = 0.5 (x_i - x_b)^T B^{-1} (x_i - x_b) + 0.5 d^T R^{-1} d
  const double costJ0JoJc = J_.getCostJoJc();

// Solve the linear system
  double reduc = this->solve(dv, rhs, Jb, UtHtRinvHU, 
                             linfoConv_, ninner, ninnerMin_, gnreduc,
                             costJ0Jb_, costJ0JbObsAux_, costJ0JoJc);

// Set ObsAux componenet of the control vector
  this->setObsAuxCtlVec(dv);

// Calculate the increment dx_k
  CtrlInc_ * dx = new CtrlInc_(J_.jb().getFirstGuessIncrement(), false);
  B.multiplySqrt(dv, *dx);

// Print
  if (linfoConv_) {
    Log::test() << classname() << ": final relative gain in information content = " << reduc << std::endl;
  } else {
    Log::test() << classname() << ": reduction in residual norm = " << reduc << std::endl;
  }
  Log::info() << classname() << " output increment:" << *dx << std::endl;

// Update gradient Jb/Save inital guess for next minimization
  if (J_.classicIncrementalFormulation()) {
    *gradJb_ += dv;
  } else {
    *dv0_ = dv;
  }

// Calculate quadratic cost function
  CtrlVec_ jbdv(dv, false);
  Jb.multiply(dv, jbdv);

// Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
  this->updateJbCostFunction(dv, jbdv);

  Log::info() << classname() << ": done" << std::endl;

  return dx;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::setupFormulation() {
  if (conf_.has("obsAuxFormulations") ||
      conf_.has("preconditioner")) {
    Log::info() << "SQRTMinimizer<MODEL>::setupFormulation()" << std::endl; 
    ASSERT(conf_.has("minimizations"));
    maxiter_ = conf_.getUnsigned("minimizations");
    if (conf_.has("obsAuxFormulations")) {
      Log::info() << "SQRTMinimizer<MODEL>::setupFormulation() set ObsAux* formulations" 
                  << std::endl; 
      setObsAuxFormulation_ = true;
      std::vector<eckit::LocalConfiguration> configs;
      conf_.get("obsAuxFormulations", configs);
      for (size_t jj = 0; jj < configs.size(); ++jj) {
         std::string key(configs[jj].getString("component"));
         bool value(configs[jj].getBool("incremental"));
         obsAuxFormulation_.insert(std::make_pair(key, value));
         Log::info() << "   ObsAuxComponent: " << key
                     << ", incremental formulation: " << value
                     << std::endl;
      }
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::setObsAuxCtlVec(CtrlVec_ & dv) const {
   if (setObsAuxFormulation_ && iter_ < maxiter_) {
      Log::info() << "SQRTMinimizer<MODEL>::setObsAuxCtlVec(dv)" << std::endl;
      std::map<std::string, bool>::const_iterator icst = obsAuxFormulation_.begin();
      while (icst != obsAuxFormulation_.end())
      {
         if (!icst->second) {
            dv.obsVar().obsAuxCtlVec(icst->first).zero();
            Log::info() << "  set component of the control vector: "
                        << icst->first << " to zero" << std::endl;
         }
         ++icst;
      }
   }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::updateJbCostFunction(const CtrlVec_ & dv, 
                                                const CtrlVec_ & jbdv) { 
  // Store jbdvs
  if (J_.classicIncrementalFormulation()) {
    dvh_.push_back(jbdv);

    // Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
    costJ0Jb_ += 0.5 * dot_product(dv, jbdv);

    // Update Jb component of J[0] for ObsAux* classes
    for (size_t jj = 0; jj < costJ0JbObsAux_.size(); ++jj) {
      const std::string & key = dv.obsVar().getKeys()[jj];
      costJ0JbObsAux_[jj] += 0.5 * dot_product(dv.obsVar().obsAuxCtlVec(key),
                                               jbdv.obsVar().obsAuxCtlVec(key));
    }
    for (size_t jouter = 1; jouter < dvh_.size(); ++jouter) {
      CtrlVec_ dvhtmp(J_.jb(), dvh_[jouter-1]);
      costJ0Jb_ += dot_product(dv, dvhtmp);
      for (size_t jj = 0; jj < costJ0JbObsAux_.size(); ++jj) {
        const std::string & key = dv.obsVar().getKeys()[jj];
        costJ0JbObsAux_[jj] += dot_product(dv.obsVar().obsAuxCtlVec(key),
                                           dvhtmp.obsVar().obsAuxCtlVec(key));
      }
    }
  } else {
    // Update Jb component of J[0]: 0.5 (x_i - x_b)^T B^-1 (x_i - x_b)
    costJ0Jb_ = 0.5 * dot_product(dv, jbdv);

    // Update Jb component of J[0] for ObsAux* classes
    for (size_t jj = 0; jj < costJ0JbObsAux_.size(); ++jj) {
      const std::string & key = dv.obsVar().getKeys()[jj];
      costJ0JbObsAux_[jj] = 0.5 * dot_product(dv.obsVar().obsAuxCtlVec(key),
                                              jbdv.obsVar().obsAuxCtlVec(key));
    }
  }

  // Print out costJ0Jb and its ObsAux components
  Log::info() << "SQRTMinimizer: costJ0Jb      : " 
              << util::full_precision(costJ0Jb_) 
              << std::endl;

  for (size_t jj = 0; jj < costJ0JbObsAux_.size(); ++jj) {
    Log::info() << "SQRTMinimizer: costJ0JbVarBC_: " 
                << util::full_precision(costJ0JbObsAux_[jj]) 
                << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpoint(eckit::LocalConfiguration & conf) const {
  Log::info() << " " << classname() << "::checkpoint starting" << std::endl;

  // GradJb
  checkpointGradJb(conf);

  // dv0
  checkpointDv0(conf);

  // dvh
  checkpointDvh(conf);

  // costJ0Jb, costJ0JbObsAux
  checkpointAux(conf);

  // LMP
  if (conf.has("preconditioner")) 
    this->checkpointLMP(conf);

  // iteration
  conf.set("outerIteration", iter_);

  Log::info() << " " << classname() << "::checkpoint done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restart(const eckit::Configuration & conf) {
  Log::info() << " " << classname() << "::restart starting" << std::endl;

  // GradJb
  restartGradJb(conf);

  // dv0
  restartDv0(conf);

  // dvh
  restartDvh(conf);

  // costJ0Jb, costJ0JbObsAux
  restartAux(conf);

  // LMP
  if (conf.has("preconditioner"))
    this->restartLMP(conf);

  // iteration
  iter_ = conf.getUnsigned("outerIteration");

  Log::info() << " " << classname() << "::restart done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpointGradJb(eckit::LocalConfiguration & conf) const {
  // GradJb
  if (J_.classicIncrementalFormulation()) {
    ASSERT(conf.has("gradJb"));
    {
      eckit::LocalConfiguration gradjbconf(conf, "gradJb");
      gradJb_->write(gradjbconf);
      Log::info() << " " << classname() << "::checkpoint gradJb: "
                  << util::full_precision(dot_product(*gradJb_, *gradJb_))
                  << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpointDv0(eckit::LocalConfiguration & conf) const {
  // dv0
  if (!J_.classicIncrementalFormulation()) {
    ASSERT(conf.has("dv0"));
    {
      eckit::LocalConfiguration dv0conf(conf, "dv0");
      ASSERT(dv0_);
      dv0_->write(dv0conf);
      Log::info() << " " << classname() << "::checkpoint dv0: "
                  << util::full_precision(dot_product(*dv0_, *dv0_))
                  << std::endl;
    }
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpointDvh(eckit::LocalConfiguration & conf) const {
  // dvh
  ASSERT(conf.has("dvh"));
  {
    std::vector<eckit::LocalConfiguration> controlVectorConf;
    conf.get("dvh", controlVectorConf);

    // ctlvec component conf
    std::vector<eckit::LocalConfiguration> ctlvecConf;
    controlVectorConf[0].get("ctlvec", ctlvecConf);

    // ctlvecObsAux component conf
    std::vector<eckit::LocalConfiguration> ctlvecObsAuxConf;
    controlVectorConf[0].get("ctlvecObsAux", ctlvecObsAuxConf);

    // ctlvecModAux compnent conf
    std::vector<eckit::LocalConfiguration> ctlvecModAuxConf;
    controlVectorConf[0].get("ctlvecModAux", ctlvecModAuxConf);

    // Construct restart configuration and write control vectors
    std::vector<eckit::LocalConfiguration> controlVectorRestartConf;
    for (size_t ii = 0; ii < dvh_.size(); ++ii) {
      eckit::LocalConfiguration controlVectorRestartSingleConf;

      // ctlvec component
      std::vector<eckit::LocalConfiguration> ctlvecRestartConf;
      for (size_t jj = 0; jj < ctlvecConf.size(); ++jj) {
        std::string filename = ctlvecConf[jj].getString("filename");
        eckit::LocalConfiguration ctlvec(ctlvecConf[jj]);
        ctlvec.set("filename", filename + "." + std::to_string(ii));
        ctlvecRestartConf.push_back(ctlvec);
      }
      controlVectorRestartSingleConf.set("ctlvec", ctlvecRestartConf);

      // ctlvecObsAux component
      std::vector<eckit::LocalConfiguration> ctlvecObsAuxRestartConf;
      for (size_t jj = 0; jj < ctlvecObsAuxConf.size(); ++jj) {
        std::string filename = ctlvecObsAuxConf[jj].getString("filename");
        eckit::LocalConfiguration ctlvecObsAux(ctlvecObsAuxConf[jj]);
        ctlvecObsAux.set("filename", filename + "." + std::to_string(ii));
        ctlvecObsAuxRestartConf.push_back(ctlvecObsAux);        
      }
      controlVectorRestartSingleConf.set("ctlvecObsAux", ctlvecObsAuxRestartConf);

      // ctlvecModAux component
      std::vector<eckit::LocalConfiguration> ctlvecModAuxRestartConf;
      for (size_t jj = 0; jj < ctlvecModAuxConf.size(); ++jj) {
        std::string filename = ctlvecModAuxConf[jj].getString("filename");
        eckit::LocalConfiguration ctlvecModAux(ctlvecModAuxConf[jj]);
        ctlvecModAux.set("filename", filename + "." + std::to_string(ii));
        ctlvecModAuxRestartConf.push_back(ctlvecModAux);
      }
      controlVectorRestartSingleConf.set("ctlvecModAux", ctlvecModAuxRestartConf);
 
      // Write control vector
      dvh_[ii].write(controlVectorRestartSingleConf);
      Log::info() << " " << classname() << "::checkpoint dvh[" << ii << "]: "
                  << util::full_precision(dot_product(dvh_[ii], dvh_[ii]))
                  << std::endl;
      controlVectorRestartConf.push_back(controlVectorRestartSingleConf);
    }
    conf.set("dvh", controlVectorRestartConf);
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::checkpointAux(eckit::LocalConfiguration & conf) const {
  // costJ0Jb
  conf.set("costJ0Jb", util::full_precision(costJ0Jb_));
  Log::info() << " " << classname() << "::checkpoint costJ0Jb: "
              << util::full_precision(costJ0Jb_) << std::endl;

  // costJ0JbObsAux
  
  // Convert to string to retain full precision
  std::vector<std::string> costJ0JbObAuxStr;
  for(auto const& value: costJ0JbObsAux_) {
    costJ0JbObAuxStr.push_back(util::full_precision(value));
  }
  conf.set("costJ0JbObsAux", costJ0JbObAuxStr);
  Log::info() << " " << classname() << "::checkpoint costJoJbObAux: " 
              << util::full_precision(costJ0JbObsAux_) << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restartGradJb(const eckit::Configuration & conf) {
  // GradJb
  if (J_.classicIncrementalFormulation()) {
    ASSERT(conf.has("gradJb"));
    eckit::LocalConfiguration gradjbconf(conf, "gradJb");
    if (gradJb_ == 0) {
      gradJb_.reset(new CtrlVec_(J_.jb()));
    }
    gradJb_->read(gradjbconf);
    Log::info() << " " << classname() << "::restart gradJb: "
                << util::full_precision(dot_product(*gradJb_, *gradJb_))
                << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restartDv0(const eckit::Configuration & conf) {
  // GradJb
  if (!J_.classicIncrementalFormulation()) {
    ASSERT(conf.has("dv0"));
    eckit::LocalConfiguration dv0conf(conf, "dv0");
    if (dv0_ == 0) {
      dv0_.reset(new CtrlVec_(J_.jb()));
    }
    dv0_->read(dv0conf);
    Log::info() << " " << classname() << "::restart dv0: "
                << util::full_precision(dot_product(*dv0_, *dv0_))
                << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restartDvh(const eckit::Configuration & conf) {
  // dvh
  ASSERT(conf.has("dvh"));
  std::vector<eckit::LocalConfiguration> controlVectorConfs;
  conf.get("dvh", controlVectorConfs);
  for (size_t jj = 0; jj < controlVectorConfs.size(); ++jj) {
    CtrlVec_ dvh(J_.jb());
    dvh.read(controlVectorConfs[jj]);
    dvh_.push_back(dvh);
    Log::info() << " " << classname() << "::restart dvh[" << jj << "]: "
                << util::full_precision(dot_product(dvh_[jj], dvh_[jj]))
                << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTMinimizer<MODEL>::restartAux(const eckit::Configuration & conf) {
  // costJ0Jb
  ASSERT(conf.has("costJ0Jb"));
  costJ0Jb_ = conf.getDouble("costJ0Jb");
  Log::info() << " " << classname() << "::restart costJ0Jb: "
              << util::full_precision(costJ0Jb_) << std::endl;

  // costJ0JbObsAux
  if (conf.has("costJ0JbObsAux")) {
    costJ0JbObsAux_ = conf.getDoubleVector("costJ0JbObsAux");
    Log::info() << " " << classname() << "::restart costJ0JbObsAux: "
                << util::full_precision(costJ0JbObsAux_) << std::endl;
  }
}

// -----------------------------------------------------------------------------



}  // namespace oops

#endif  // OOPS_ASSIMILATION_SQRTMINIMIZ_H_
