/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_SQRTPLANCZOSMINIMIZER_H_
#define OOPS_ASSIMILATION_SQRTPLANCZOSMINIMIZER_H_

#include <cmath>
#include <string>
#include <iostream>
#include <vector>

#include <boost/ptr_container/ptr_vector.hpp>

#include "util/Logger.h"
#include "util/LogbookWriter.h"
#include "util/dot_product.h"
#include "util/formats.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/JbMatrix.h"
#include "oops/assimilation/SpectralSqrtLMP.h"
#include "oops/assimilation/SQRTMinimizer.h"
#include "oops/assimilation/TriDiagSolve.h"
#include "oops/assimilation/UtHtRinvHUMatrix.h"

namespace oops {

/// SQRTPLanczos Minimizer
/*!
 * \brief Preconditioned Lanczos solver for the square root preconditioned
 * formulation of the Hessian
 *
 * A must be square, symmetric, positive definite.

 * On entry:
 * -    A       = \f$ (U^T B^-1 U + U^T H^T R^{-1} U) \f$ 
 * -    dv      =  starting point, \f$ dv_{0} \f$.
 * -    rr      = \f$ (-sum dv^{b}_{i} + ) U^T H^T R^{-1} d  - A dv0 \f$
 * - or
 * -    rr      = \f$ U^T H^T R^{-1} [d + H (x_b - x_fg)] - A dv0 \f$

 * On exit, dv will contain the solution \f$ U^{-1} dx \f$
 *  The solution \f$ dx = U dv \f$ is recovered in the SQRTMinimizer class
 *  The return value is the achieved reduction in preconditioned residual norm.
 *
 *  Iteration will stop if the maximum iteration limit "maxiter" is reached
 *  or if the residual norm reduces by a factor of "tolerance".
 *
 */

// -----------------------------------------------------------------------------

template<typename MODEL> class SQRTPLanczosMinimizer : public SQRTMinimizer<MODEL> {
  typedef CostFunction<MODEL>     CostFct_;
  typedef ControlVector<MODEL>    CtrlVec_;
  typedef JbMatrix<MODEL>         Jbmat_;
  typedef UtHtRinvHUMatrix<MODEL> UtHtRinvHU_;

 public:
  const std::string classname() const final {return "SQRTPLanczosMinimizer";}

/// Constructor, destructor
  SQRTPLanczosMinimizer(const eckit::Configuration &, const CostFct_ &);
  ~SQRTPLanczosMinimizer() {}

 private:
  double solve(CtrlVec_ &, CtrlVec_ &, const Jbmat_ &, const UtHtRinvHU_ &,
               const bool &, const size_t &, const size_t &, const double &, 
               const double &, const std::vector<double> &, const double &) final;

  void releaseResources();

/// Checkpointing, restart of the limited memory preconditioner
  void checkpointLMP(eckit::LocalConfiguration &) const final;
  void restartLMP(const eckit::Configuration &) final;

/// Quadratic cost function calculations
  void setupQuadCost(const double & costJ0Jb,
                     const double & costJ0JoJc);
  void calcQuadCost(const CtrlVec_ &,
                    const std::vector<double> &);
  void printQuadCost(const size_t &) const;

/// Other diagnostics
  bool calcRitzInformation();
  void printRitzInformation(const size_t &) const;

/// Data memebers
  const CostFct_ & J_;

/// LMP
  SpectralSqrtLMP<MODEL> lmp_;

/// Local
  std::vector<boost::shared_ptr<CtrlVec_> > vvecs_;
  std::vector<double> alphas_;
  std::vector<double> betas_;

  std::vector<double> evals_;
  std::vector<double> eigvals_;
  std::vector<double> erritz_;
  std::vector<double> erritzlm_;
  std::vector<std::vector<double> > evecs_;

  std::vector<std::string> error_messages_;

/// Quadratic cost function calculations data members
  double costJ0_;
  double costJ0Jb_;
  double costJ0JoJc_;

  double costJ_;
  double costJb_;
  double costJbLocal_; // component of costJb from current minimization
  double costJoJc_;
};

// =============================================================================

template<typename MODEL>
SQRTPLanczosMinimizer<MODEL>::SQRTPLanczosMinimizer(const eckit::Configuration & conf,
                                                    const CostFct_ & J)
  : SQRTMinimizer<MODEL>(conf, J), J_(J), lmp_(conf, J), 
    vvecs_(), alphas_(), betas_(),
    costJ0_(0), costJ0Jb_(0), costJ0JoJc_(0),
    costJ_(0), costJb_(0), costJbLocal_(0), costJoJc_(0) 
{}

// -----------------------------------------------------------------------------

template<typename MODEL>
double SQRTPLanczosMinimizer<MODEL>::solve(CtrlVec_ & dv, CtrlVec_ & rr,
                                           const Jbmat_ & Jb, 
                                           const UtHtRinvHU_ & UtHtRinvHU,
                                           const bool & linfoconv, 
                                           const size_t & maxiter,
                                           const size_t & miniter,
                                           const double & tolerance,
                                           const double & costJ0Jb,
                                           const std::vector<double> & costJ0JbObsAux,
                                           const double & costJ0JoJc) {
  // dv   control vector 

  CtrlVec_ dv0(dv);
  CtrlVec_ mdv(dv,false);
  CtrlVec_ mvv(dv,false);
  CtrlVec_ jbzz(dv,false);

  CtrlVec_ mrr(rr,false);
  CtrlVec_ vv(rr,false);
  CtrlVec_ zz(rr,false);

  std::vector<double> ss;
  std::vector<double> dd;

  // J0
  this->setupQuadCost(costJ0Jb, costJ0JoJc);

  // Change resolution of LMP vectors
  lmp_.changeResolution();

  // Postprocess preconditioning vectors
  lmp_.postprocess();

  // Caclulate residual

  UtHtRinvHU.multiply(dv0, zz);
  Jb.multiply(dv0, jbzz);
  zz += jbzz;
  rr -= zz;

  // v_{0} = P^-1/2 r_{0}
  lmp_.inverseMultiplySqrt(rr, mrr);
  vv = mrr;

  // beta_{0} = sqrt( v_{0}^T v_{0} )
  double beta = sqrt(dot_product(vv, vv));
  const double beta0 = beta;

  // v_{1} = r_{0} / beta_{0}
  vv *= 1.0/beta;

  // vvecs[0] = v_{1} ---> for re-orthogonalization
  vvecs_.push_back(boost::shared_ptr<CtrlVec_>(new CtrlVec_(vv)));

  double normReduction = 1.0;
  double zdfi = 0.0;
  bool convergenceReached = false;
  size_t jiter = 0; 

  while (jiter < maxiter) {
    Log::info() << "SQRTPLanczos Starting Iteration " << jiter+1 << std::endl;

    util::LogbookWriter<size_t> log("InnerLoop", jiter+1);

    // v_{i+1} = P^-1/2 ( I + U^T H^T R^{-1} H U ) P^-1/2 v_{i}
    lmp_.inverseMultiplySqrt(vv, mvv);

    UtHtRinvHU.multiply(mvv, zz);
    Jb.multiply(mvv, jbzz);
    zz += jbzz;

    lmp_.inverseMultiplySqrt(zz, vv);

    // v_{i+1} = v_{i+1} - beta * v_{i-1}
    if (jiter > 0) vv.axpy(-beta, *vvecs_[jiter-1]);

    // alpha_{i} = v_{i+1}^T v_{i}
    double alpha = dot_product(*vvecs_[jiter], vv);
    if (alpha <= 0.0) {
      error_messages_.push_back("SQRTPLanczos: stopping J'' not positive definite");
      break;
    }

    // v_{i+1} = v_{i+1} - alpha_{i} v_{i}
    vv.axpy(-alpha, *vvecs_[jiter]);  

    // Re-orthogonalization
    for (size_t jj = 0; jj < jiter; ++jj) {
      double proj = dot_product(vv, *vvecs_[jj]);
      vv.axpy(-proj, *vvecs_[jj]);
    }

    // beta_{i+1} = sqrt( vv_{i+1}^t, vv_{i+1} )
    beta = sqrt(dot_product(vv, vv));

    // v_{i+1} = v_{i+1} / beta_{i+1}
    vv *= 1.0/beta;

    // vvecs[i+1] = v_{i+1}
    vvecs_.push_back(boost::shared_ptr<CtrlVec_>(new CtrlVec_(vv)));

    alphas_.push_back(alpha);

    if (jiter == 0) {
      ss.push_back(beta0/alpha);
      dd.push_back(beta0);
    } else {
      // Solve the tridiagonal system T_{i} s_{i} = beta0 * e_1
      dd.push_back(beta0*dot_product(*vvecs_[0], vv));
      TriDiagSolve(alphas_, betas_, dd, ss);
    }

    betas_.push_back(beta);

    // Compute the quadratic cost function
    this->calcQuadCost(mrr, ss);

    // Compute the Ritz information
    bool ritzFailureCheckFlag = this->calcRitzInformation();

    // Convergence criterion
    if (linfoconv) {
      // Information content based convergence criterion
      //  - Information content defined as 0.5 dv^T dv
      //  - We should measure the change in the solution vector length
      //    in the un-preconditioned metric
      //  - This would require additional call to the preconditioner
      double zdfi_old = zdfi;
      zdfi = costJbLocal_; 
      ASSERT(zdfi != 0.0);
      normReduction = (zdfi - zdfi_old)/zdfi;
      Log::info() << "SQRTPLanczos end of iteration " << jiter+1 << std::endl
                  << "  Relative information content gain (" << std::setw(2) << jiter+1 << ") = "
                  << util::full_precision(normReduction) << std::endl;
    } else {
      // Gradient reduction based convergence criterion
      // Gradient norm in precond metric --> sqrt(r'z) --> beta * s_{i}
      double rznorm = beta*std::abs(ss[jiter]);
      normReduction = rznorm/beta0;
      Log::info() << "SQRTPLanczos end of iteration " << jiter+1 << std::endl
                  << "  Norm reduction (" << std::setw(2) << jiter+1 << ") = "
                  << util::full_precision(normReduction) << std::endl;
    }

    // Print iteration summary
    this->printQuadCost(jiter);
    this->printRitzInformation(jiter);

    // Check if convergence criteria are met
    if (ritzFailureCheckFlag) break;
    if (normReduction < tolerance && jiter >= miniter) {
      convergenceReached = true;
      break;
    }
 
    // Increment iteration counter
    ++jiter;    
  }

  // Print summary
  Log::info() << "Summary of SQRTPLanczos solver: "                        << std::endl
              << " Information based convergence criterion: " << linfoconv << std::endl
              << " Number of iterations performed: "          << jiter     << std::endl
              << " Maximum allowed number of iterations: "    << maxiter   << std::endl
              << " Minimum allowed number of iterations: "    << miniter   << std::endl;
  if (linfoconv) {
    Log::info() << " Required relative gain in information content: " << tolerance     << std::endl
                << " Last relative gain in information content: "     << normReduction << std::endl;
  } else {
    Log::info() << " Required reduction in norm of gradient: " << tolerance     << std::endl
                << " Achieved reduction in norm of gradient: " << normReduction << std::endl;
  }
  if (convergenceReached) {
    Log::info() << " Requested convergence criteria met. Well done. " << std::endl;
  } else {
    Log::info() << " Failed to meet convergence criteria. " << std::endl;
    for (auto const & err: error_messages_) {
      Log::info() << "  " << err << std::endl;
    } 
  }
  Log::info() << std::endl;

  // Calculate the solution
  mdv.zero();
  for (size_t jj = 0; jj < ss.size(); ++jj) {
    mdv.axpy(ss[jj], *vvecs_[jj]);
  }

  // Transform control variable back to space with scalar product
  lmp_.inverseMultiplySqrt(mdv,dv);

  dv += dv0;

  // Update LMP
  if (SQRTMinimizer<MODEL>::iter_ < SQRTMinimizer<MODEL>::maxiter_) 
    lmp_.update(vvecs_, alphas_, betas_);

  // Clean up
  releaseResources();

  return normReduction;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTPLanczosMinimizer<MODEL>::releaseResources() {
  vvecs_.clear();
  alphas_.clear();
  betas_.clear();
  evals_.clear();
  eigvals_.clear();
  evecs_.clear();
  erritz_.clear();
  erritzlm_.clear();
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTPLanczosMinimizer<MODEL>::setupQuadCost(const double & costJ0Jb,
                                                 const double & costJ0JoJc) {
  // J0
  costJ0Jb_ = costJ0Jb;
  costJ0JoJc_ = costJ0JoJc;

  // J0
  costJ0_ = costJ0Jb_ + costJ0JoJc_;

  // reset
  costJ_ = 0;
  costJb_ = 0;
  costJbLocal_ = 0;
  costJoJc_ = 0;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTPLanczosMinimizer<MODEL>::calcQuadCost(const CtrlVec_ & rr,
                                                const std::vector<double> & ss) {
  // Compute the quadratic cost function
  // J[du_{i}] = J[0] - 0.5 s_{i}^T Z_{i}^T r_{0}
  // Jb[du_{i}] = 0.5 s_{i}^T V_{i}^T Z_{i} s_{i}
  // Note: here we calculate the quadratic cost function in the preconditioned
  //       space; 
  costJ_ = costJ0_;
  costJb_ = costJ0Jb_;
  costJbLocal_ = 0.0;
  for (size_t jj = 0; jj < ss.size(); ++jj) {
    costJ_ -= 0.5 * ss[jj] * dot_product(*vvecs_[jj], rr);
    costJbLocal_ += 0.5 * ss[jj] * dot_product(*vvecs_[jj], *vvecs_[jj]) * ss[jj];
  }
  costJb_ += costJbLocal_;
  costJoJc_ = costJ_ - costJb_;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTPLanczosMinimizer<MODEL>::printQuadCost(const size_t & jiter) const {
  // Print the quadratic cost function
  Log::info() << "  Quadratic cost function: J   (" << std::setw(2) << jiter+1 << ") = "
                << util::full_precision(costJ_)         << std::endl
              << "  Quadratic cost function: Jb  (" << std::setw(2) << jiter+1 << ") = "
                << util::full_precision(costJb_)        << std::endl
              << "  Quadratic cost function: JoJc(" << std::setw(2) << jiter+1 << ") = "
                << util::full_precision(costJoJc_)      << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
bool SQRTPLanczosMinimizer<MODEL>::calcRitzInformation() {
  ASSERT(alphas_.size() == betas_.size());
  const unsigned nvec = alphas_.size();
  bool ritzFailureCheckFlag = false;
  // Save the leading converged eigenvalue
  std::vector<double> eigvals_old; 
  if (eigvals_.size() > 0) eigvals_old.push_back(eigvals_[0]);

  if (nvec > 0) {
    evals_.clear();
    eigvals_.clear();
    evecs_.clear();
    erritz_.clear();
    erritzlm_.clear();
//  Compute spectrum of tri-diagonal matrix
    TriDiagSpectrum(alphas_, betas_, evals_, evecs_);

//  Determine the converged eigenvalues
    for (unsigned jiter = 0; jiter < nvec; ++jiter) {
      double erritz = std::abs(evecs_[jiter][nvec-1]*betas_[nvec-1]);
      double lambda = evals_[jiter];
      if (lambda < 0) {
        // NEGATIVE RITZ VALUE DETECTED" 
        ritzFailureCheckFlag = true;
        error_messages_.push_back("Negative Ritz value detected");
        break;
      }
//    Add new converged eigenvalues
      double erritzlm = 0.0001*lambda;
      if (erritz <= erritzlm) {
        eigvals_.push_back(lambda);
        erritz_.push_back(erritz);
        erritzlm_.push_back(erritzlm);
      }
    }

// Check if Ritz values are not exploding
    if (eigvals_.size() != 0 && eigvals_old.size() != 0) {
      if (eigvals_[0] > 1.01 * eigvals_old[0]) {
        // RITZ VALUES EXPLODE!
        ritzFailureCheckFlag = true;
        Log::info() << "SQRTPLanczos: Ritz values explode" << std::endl
                    << "Leading Ritz value: " << eigvals_[0] << std::endl
                    << "Leading converged eigenvalue: " << eigvals_old[0] << std::endl;
        error_messages_.push_back("Ritz values explode");
      }
    }
  }
  return ritzFailureCheckFlag;
}
                                  
// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTPLanczosMinimizer<MODEL>::printRitzInformation(const size_t & jiter) const {
  ASSERT(alphas_.size() == betas_.size());
  const unsigned nvec = alphas_.size();

  if (nvec > 0) {
    Log::info() << "  Converged Ritz values (" << jiter+1 << "):" << std::endl;
    for (auto const & eigval: eigvals_) {
      Log::info() << "    Ritz value: " << util::full_precision(eigval) << std::endl;
    }
    for (auto const & err: erritz_) {
      Log::info() << "    Error bounds: " << util::full_precision(err) << std::endl;
    }
    for (auto const & errlm: erritzlm_) {
      Log::info() << "    Error bound limits: " << util::full_precision(errlm) << std::endl;
    }
    Log::info() << std::endl;
  }
  return;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTPLanczosMinimizer<MODEL>::checkpointLMP(eckit::LocalConfiguration & conf) const {
  ASSERT(conf.has("preconditioner"));

  // Checkpoint lmp
  lmp_.checkpoint(conf);
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SQRTPLanczosMinimizer<MODEL>::restartLMP(const eckit::Configuration & conf) {
  ASSERT(conf.has("preconditioner"));

  // Restart lmp
  lmp_.restart(conf);
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_SQRTPLANCZOSMINIMIZER_H_
