/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_SPECTRALSQRTLMP_H_
#define OOPS_ASSIMILATION_SPECTRALSQRTLMP_H_

#include <algorithm>
#include <cmath>
#include <numeric>
#include <string>
#include <vector>

#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/shared_ptr.hpp>

#include "eckit/config/LocalConfiguration.h"
#include "util/dot_product.h"
#include "util/formats.h"
#include "util/Logger.h"
#include "oops/assimilation/CholDecomp.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/ControlVector.h"
#include "oops/assimilation/SymMatSpectrum.h"
#include "oops/assimilation/TriDiagSpectrum.h"

namespace oops {

  /*  The SPECTRAL preconditioner matrix \f$ P \f$ for the system matrix A;
   *  It is defined as \f$ P_(k+1)  = P_k  +  U (D^{-1} - I_l) U' \f$
   *  where
   *  \f$ D \f$ is a l-by-l diagonal matrix whose diagonal entries are \f$ lambda_i \f$,
   *  \f$ U = [u_1, u_2, ..., u_l] \f$ is a column matrix.
   *
   *  The pairs \f$ (lambda_i,u_i) \f$ define the Ritz pairs of the matrix
   *  \f$ PA \f$ with respect to the subspace \f$ K(PA,Pr0) \f$
   *   Note that \f$ r0 \f$ is the initial residual.
   *
   */
  /*!
   *  The solvers represent matrices as objects that implement a "multiply"
   *  method. This class defines objects that apply the preconditioner matrix \f$ P \f$.
   */

// -----------------------------------------------------------------------------

template<typename MODEL> class SpectralSqrtLMP {
  typedef CostFunction<MODEL>     CostFct_;
  typedef ControlVector<MODEL>    CtrlVec_;
 public:
  const std::string classname() const {return "SpectralSqrtLMP";}

/// Constructor, destructor
  explicit SpectralSqrtLMP(const eckit::Configuration &, const CostFct_ &);
  ~SpectralSqrtLMP() {}

/// Methods
  void update(const std::vector<boost::shared_ptr<CtrlVec_> > &, 
              const std::vector<double> &, 
              const std::vector<double> &);
  void postprocess();
  void changeResolution();
  void inverseMultiply(const CtrlVec_ &, CtrlVec_ &) const;
  void inverseMultiplySqrt(const CtrlVec_ &, CtrlVec_ &) const;
  void multiplySqrt(const CtrlVec_ &, CtrlVec_ &) const;

/// Checkpointing, restart of the limited memory preconditioner
  void checkpoint(eckit::LocalConfiguration &) const;
  void restart(const eckit::Configuration &);

 private:
  const CostFct_ & J_;

  const bool active_;
  bool orthogonalize_;
  bool useoldpairs_;
  size_t maxpairs_;
  size_t update_;
  double mineigval_;
  double errthreshold_;

  std::vector<size_t> usedpairIndx_;
  std::vector<double> eigvals_;                   
  std::vector<boost::shared_ptr<CtrlVec_> > U_;
  std::vector<boost::shared_ptr<CtrlVec_> > Q_;
};

// =============================================================================

template<typename MODEL>
SpectralSqrtLMP<MODEL>::SpectralSqrtLMP(const eckit::Configuration & conf, 
                                        const CostFct_ & J)
  : J_(J), active_(conf.has("preconditioner")), 
    orthogonalize_(true), useoldpairs_(false), maxpairs_(0),
    update_(0), mineigval_(1.0), errthreshold_(0.001), usedpairIndx_(),
    eigvals_(), U_(), Q_()
{
  if (active_) {
    Log::info() << "SpectralSqrtLMP::SpectralSqrtLMP constructor starting" << std::endl;

    const eckit::LocalConfiguration precond(conf, "preconditioner");
    if (precond.has("maxpairs")) maxpairs_ = precond.getInt("maxpairs");
    Log::info() << "  maxpairs_ = " << maxpairs_ << std::endl;
    if (precond.has("error_threshold")) errthreshold_ = precond.getDouble("error_threshold");
    Log::info() << "  errthreshold_ = " << errthreshold_ << std::endl;

    if (maxpairs_ > 0) {
      std::string old;
      if (precond.has("orthogonalize")) old = precond.getString("orthogonalize");
      orthogonalize_ = (old == "on" || old == "true");
      Log::info() << "  orthogonalize_ = " << orthogonalize_ << std::endl;
      if (precond.has("useoldpairs")) old = precond.getString("useoldpairs");
      useoldpairs_ = (old == "on" || old == "true");
      Log::info() << "  useoldpairs_ = " << useoldpairs_ << std::endl;
      if (precond.has("mineigval")) mineigval_ = precond.getDouble("mineigval");
      Log::info() << "  mineigval_ = " << mineigval_ << std::endl;
    }
    
    Log::info() << "SpectralSqrtLMP::SpectralSqrtLMP constructor done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SpectralSqrtLMP<MODEL>::update(const std::vector<boost::shared_ptr<CtrlVec_> > & Zhl,
                                    const std::vector<double> & alphas,
                                    const std::vector<double> & betas) {
  if (active_) {
    Log::info() << "SpectralSqrtLMP::update starting" << std::endl;
//  If useoldpairs = false, use only current information
    if (!useoldpairs_) {
      eigvals_.clear();
      usedpairIndx_.clear();
      U_.clear();
      Q_.clear();
    }

    const size_t nvec = Zhl.size();
    if (nvec > 0) {
      ASSERT(alphas.size() ==  nvec-1);
      ASSERT(betas.size() == nvec-1);

//    Compute spectrum of tri-diagonal matrix
      std::vector<double> evals;
      std::vector<std::vector<double> > evecs;
      TriDiagSpectrum(alphas, betas, evals, evecs);

//    Determine the converged eigenvalues
      std::vector<size_t> convIndx;
      std::vector<double> neweigvals;
      for (int ji = nvec-2; ji >= 0; ji--) {
        double erritz = evecs[ji][nvec-2]*betas[nvec-2];
        double lambda = evals[ji];
//      Add new converged eigenvalues
        if (std::abs(erritz) <= errthreshold_*lambda &&
            lambda > mineigval_ && 
            eigvals_.size() < maxpairs_) {
          Log::info() << "Eigenvalue accepted: " << lambda
                      << ", error= " << std::abs(erritz)
                      << ", threshold= " << errthreshold_*lambda
                      << std::endl;
          convIndx.push_back(ji);
          eigvals_.push_back(lambda);
          neweigvals.push_back(lambda);
        } else {
          if (eigvals_.size() >= maxpairs_)
            Log::info() << "Warning: Maximum allowed number of pairs exceeded" << std::endl;

          Log::info() << "Eigenvalue not accepted: " << lambda
                      << ", error= " << std::abs(erritz)
                      << ", threshold= " << errthreshold_*lambda
                      << std::endl;
        }
      }
//    Keep information on how many pairs are used at each minimization loop
      usedpairIndx_.push_back(convIndx.size());
      Log::info() << " Outer loop                    : " << update_                           << std::endl
                  << " Number of inner iterations    : " << Zhl.size()-1                      << std::endl
                  << " Maximum number of pairs       : " << maxpairs_                         << std::endl
                  << " Number of new converged pairs : " << convIndx.size()                   << std::endl
                  << " New converged eigenvalues     : " << util::full_precision(neweigvals)  << std::endl;
      size_t count = 0;                  
      for (unsigned kiter = 0; kiter < usedpairIndx_.size(); ++kiter) {
        count += usedpairIndx_[kiter];
        Log::info() << " Number of used pairs from outer loop " << kiter
                    << " : " << usedpairIndx_[kiter] << std::endl;
      }
      Log::info() << " Total number of used pairs    : " << count << std::endl;
//    Calculate/Update the matrix U
      std::vector<boost::shared_ptr<CtrlVec_> > Unew;
      for (size_t ji = 0; ji < convIndx.size(); ++ji) {
        boost::shared_ptr<CtrlVec_> uu(new CtrlVec_(J_.jb()));
        uu->zero();
        for (size_t ii = 0; ii < nvec-1; ++ii) {
          uu->axpy(evecs[convIndx[ji]][ii], *Zhl[ii]);
        }
        Unew.push_back(uu);
      }
//    Insert new eigenvectors      
      for (size_t ji = 0; ji < Unew.size(); ++ji) 
        U_.push_back(Unew[ji]);
    }
    ++update_;
    Log::info() << "SpectralSqrtLMP::update done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SpectralSqrtLMP<MODEL>::postprocess() {
  if (active_) {
/*  The Ritz vectors combined from different minimizations are not necessarily orthogonal,
 *  while the preconditioner requires a set of orthogonal vectors. We first compute 
 *  eigenvectors v_i of (QU) (QU)^T, where U is a rectangular matrix with columns being 
 *  Ritz vectors and Q is a Hausholder matrix - QU is an upper triangular matrix and
 *  (QU)(QU)^T has only a non-zero block in the top left corner of size LxL, where L is
 *  the number of Ritz vectors forming U. The required preconditioning vectors are given 
 *  by w_i = Q^T v_i and the required eigenvalues are given by eigenvalues of the LxL matrix 
 */ 
    Log::info() << "SpectralSqrtLMP::postprocess starting" << std::endl;

    if (orthogonalize_) {
      Log::info() << "SpectralSqrtLMP::postprocess orthogonalization starting" << std::endl;
      size_t nvecs = U_.size();
      ASSERT(nvecs <= maxpairs_);
    
      // Copy and transform preconditioning vectors/eigenvectors
      Q_.clear();
      for (size_t jj = 0; jj < nvecs; ++jj) {
        boost::shared_ptr<CtrlVec_> qqvec(new CtrlVec_(*U_[jj]));
        double scale = std::sqrt(eigvals_[jj]-1);
        (*qqvec) *= scale;
        Q_.push_back(qqvec);
      }

      // qumat matrix - stored in a column major order
      std::vector<double> qumat(nvecs*nvecs);
      for (size_t jj = 0; jj < nvecs; ++jj) {
        std::vector<double> qqvec(jj+1);
        Q_[jj]->state()[0].modCtlVec().getSection(qqvec, 1, jj+1);
        double zz = dot_product(*Q_[jj], *Q_[jj]) - 
                    std::inner_product(qqvec.begin(), qqvec.end(), qqvec.begin(), 0.0);
        if (qqvec[jj] < 0) {
          qumat[jj + jj*nvecs] = -std::sqrt(zz + qqvec[jj] * qqvec[jj]);
        } else {
          qumat[jj + jj*nvecs] =  std::sqrt(zz + qqvec[jj] * qqvec[jj]);
        }
        qqvec[jj] = qqvec[jj] - qumat[jj + jj*nvecs];
        for (size_t jk = 0; jk < jj; ++jk) {
          qumat[jk + jj*nvecs] = qqvec[jk];
        }
        zz = zz + qqvec[jj] * qqvec[jj];
        std::vector<double> zzvec(jj+1);
        zzvec[jj] = qqvec[jj];
        Q_[jj]->state()[0].modCtlVec().setSection(zzvec, 1, jj+1);
        *Q_[jj] *= std::sqrt(2.0/zz);

        // We now have Householder vectors in Q[jj-1] and the non-zero elements of the 
        // transformed vector in qumat. Now apply the Householder transfromations to the
        // remaining vectors.

        for (size_t ji = jj+1; ji < nvecs; ++ji) {
          double zz = dot_product(*Q_[jj], *Q_[ji]);
          Q_[ji]->axpy(-zz, *Q_[jj]);
        }
      }

      // Multiply the upper triangle by its transpose and find eigenvectors and eigenvalues
      for (size_t jj = 0; jj < nvecs; ++jj) {
        for (size_t ji = jj+1; ji < nvecs; ++ji) {
          qumat[ji + jj*nvecs] = 0.0;
        }
      }

      // ququtmat matrix - stored in a column major order
      std::vector<double> ququtmat(nvecs*nvecs);
      // ququtmat = qumat * qumat^T
      for (size_t jj=0; jj < nvecs; ++jj) {
        for (size_t ji=jj; ji < nvecs; ++ji) {
          ququtmat[ji + jj*nvecs] = 0.0;
          for (size_t jk=ji; jk < nvecs; ++jk) {
            ququtmat[ji + jj*nvecs] = ququtmat[ji + jj*nvecs] + 
                                      qumat[ji + jk*nvecs] * qumat[jj + jk*nvecs];
          }
        }
      }

      // Find eigenpairs of ququtmat (q_i, v_i)
      std::vector<double> evals;
      std::vector<std::vector<double> > evecs;
      SymMatSpectrum(ququtmat, evals, evecs);

      // Transform eigenvalues
      for (auto & val: evals) {
          val = val + 1.0;
      }

      Log::info() << "Eigenvalues: " << util::full_precision(evals) << std::endl;

      // Order eigenvalues 
      std::vector<int> indices(evals.size());
      std::size_t n(0);
      std::generate(std::begin(indices), 
                    std::end(indices), 
                    [&]{ return n++; });

      std::sort(std::begin(indices), 
                std::end(indices), 
                [&](int ii1, int ii2){ return evals[ii1] > evals[ii2]; });

      Log::info() << "Sorted eigenvalues (" << indices.size() << "):" << std::endl;
      for (const auto & ind: indices) 
        Log::info() << "Index: " << ind << ", value: " << evals[ind] << std::endl;

      // Accept or reject eigenvectors
      indices.erase(std::remove_if(indices.begin(), 
                                   indices.end(), 
                                   [&](int n) { return evals[n] <= 1.0; }), 
                    indices.end());

      Log::info() << "Sorted accepted eigenvalues (" << indices.size() << "):" << std::endl;
      for (const auto & ind: indices) 
        Log::info() << "Index: " << ind << ", value: " << evals[ind] << std::endl;

      if (indices.size() == 0)
        Log::info() << "Warning: cannot form preconditioner - no positive eigenvalues; " 
                    << "minimisation will not be preconditioned" << std::endl;

      // Transform back w_i = Q^T v_i
      U_.clear();
      eigvals_.clear();
      for (auto const & ind: indices) {
        // Limit eigenvalues
        evals[ind] = std::min(10.0, evals[ind]);
        eigvals_.push_back(evals[ind]);
        boost::shared_ptr<CtrlVec_> ww(new CtrlVec_(J_.jb()));
        ww->zero();
        ww->state()[0].modCtlVec().setSection(evecs[ind], 1, nvecs);
        for (int ji = nvecs-1; ji >= 0; --ji) {
          double zz = dot_product(*Q_[ji], *ww);
          ww->axpy(-zz, *Q_[ji]);
        }
        U_.push_back(ww);
      }
      // Release Q
      Q_.clear();

      Log::info() << "SpectralSqrtLMP::postprocess orthogonalization done" << std::endl;
    } else {
      // Limit eigenvalues
      for (auto & val: eigvals_)
        val = std::min(10.0, val);
    }
    Log::info() << "SpectralSqrtLMP::postprocess done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SpectralSqrtLMP<MODEL>::changeResolution() {
  if (active_) {
//  Interpolate to the new resolution
    Log::info() << "SpectralSqrtLMP::changeResolution starting" << std::endl;
    for (size_t jj = 0; jj < U_.size(); ++jj) {
       U_[jj].reset(new CtrlVec_(J_.jb(), *U_[jj]));
    }
    Log::info() << "SpectralSqrtLMP::changeResolution done" << std::endl;
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SpectralSqrtLMP<MODEL>::inverseMultiply(const CtrlVec_ & a, CtrlVec_ & b) const {
  b = a;  // P_0 = I

  for (size_t ii = 0; ii < U_.size(); ++ii) {
    double zeval = eigvals_[ii];
    ASSERT(zeval);
    double zz = 1.0/zeval - 1.0;
    b.axpy(zz*dot_product(a, *U_[ii]), *U_[ii]);
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SpectralSqrtLMP<MODEL>::inverseMultiplySqrt(const CtrlVec_ & a, CtrlVec_ & b) const {
  b = a;  // P_0 = I

  for (size_t ii = 0; ii < U_.size(); ++ii) {
    double zeval = eigvals_[ii];
    ASSERT(zeval);
    double zz = 1.0/std::sqrt(zeval) - 1.0;
    b.axpy(zz*dot_product(a, *U_[ii]), *U_[ii]);
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SpectralSqrtLMP<MODEL>::multiplySqrt(const CtrlVec_ & a, CtrlVec_ & b) const {
  b = a;  // P_0 = I

  for (size_t ii = 0; ii < U_.size(); ++ii) {
    double zeval = eigvals_[ii];
    ASSERT(zeval);
    double zz = sqrt(zeval) - 1.0;
    b.axpy(zz*dot_product(a, *U_[ii]), *U_[ii]);
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SpectralSqrtLMP<MODEL>::checkpoint(eckit::LocalConfiguration & conf) const {
  if (active_) {
    std::vector<eckit::LocalConfiguration> controlVectorConf;
    conf.get("preconditioner", controlVectorConf);

    // ctlvec component conf
    std::vector<eckit::LocalConfiguration> ctlvecConf;
    controlVectorConf[0].get("ctlvec", ctlvecConf);

    // ctlvecObsAux component conf
    std::vector<eckit::LocalConfiguration> ctlvecObsAuxConf;
    controlVectorConf[0].get("ctlvecObsAux", ctlvecObsAuxConf);

    // ctlvecModAux compnent conf
    std::vector<eckit::LocalConfiguration> ctlvecModAuxConf;
    controlVectorConf[0].get("ctlvecModAux", ctlvecModAuxConf);

    // Construct restart configuration and write control vectors
    std::vector<eckit::LocalConfiguration> controlVectorRestartConf;
    for (size_t ii = 0; ii < U_.size(); ++ii) {
      eckit::LocalConfiguration controlVectorRestartSingleConf;

      // ctlvec component
      std::vector<eckit::LocalConfiguration> ctlvecRestartConf;
      for (size_t jj = 0; jj < ctlvecConf.size(); ++jj) {
        std::string filename = ctlvecConf[jj].getString("filename");
        eckit::LocalConfiguration ctlvec(ctlvecConf[jj]);
        ctlvec.set("filename", filename + "." + std::to_string(ii));
        ctlvecRestartConf.push_back(ctlvec);
      }
      controlVectorRestartSingleConf.set("ctlvec", ctlvecRestartConf);

      // ctlvecObsAux component
      std::vector<eckit::LocalConfiguration> ctlvecObsAuxRestartConf;
      for (size_t jj = 0; jj < ctlvecObsAuxConf.size(); ++jj) {
        std::string filename = ctlvecObsAuxConf[jj].getString("filename");
        eckit::LocalConfiguration ctlvecObsAux(ctlvecObsAuxConf[jj]);
        ctlvecObsAux.set("filename", filename + "." + std::to_string(ii));
        ctlvecObsAuxRestartConf.push_back(ctlvecObsAux);
      }
      controlVectorRestartSingleConf.set("ctlvecObsAux", ctlvecObsAuxRestartConf);

      // ctlvecModAux component
      std::vector<eckit::LocalConfiguration> ctlvecModAuxRestartConf;
      for (size_t jj = 0; jj < ctlvecModAuxConf.size(); ++jj) {
        std::string filename = ctlvecModAuxConf[jj].getString("filename");
        eckit::LocalConfiguration ctlvecModAux(ctlvecModAuxConf[jj]);
        ctlvecModAux.set("filename", filename + "." + std::to_string(ii));
        ctlvecModAuxRestartConf.push_back(ctlvecModAux);
      }
      controlVectorRestartSingleConf.set("ctlvecModAux", ctlvecModAuxRestartConf);

      // Write control vector
      U_[ii]->write(controlVectorRestartSingleConf);
      Log::info() << " " <<  classname() << "::checkpoint lmp[" << ii << "]: "
                  << util::full_precision(dot_product(*U_[ii], *U_[ii]))
                  << std::endl;
      controlVectorRestartConf.push_back(controlVectorRestartSingleConf);
    }
    conf.set("preconditioner", controlVectorRestartConf);

    // Additional LMP checkpointing
    eckit::LocalConfiguration auxconf;

    // Convert to string to retain full precision
    std::vector<std::string> eigvalstr;
    for(auto const& value: eigvals_) {
      eigvalstr.push_back(util::full_precision(value));
    }
    auxconf.set("eigvals", eigvalstr);
    Log::info() << " " << classname() << "::checkpoint eigvals " 
                << util::full_precision(eigvals_) << std::endl;

    auxconf.set("usedpairIndx", usedpairIndx_);
    Log::info() << " " << classname() << "::checkpoint usedpairIndx " 
                << util::full_precision(usedpairIndx_) << std::endl;

    auxconf.set("update", update_);
    Log::info() << " " << classname() << "::checkpoint update " 
                << util::full_precision(update_) << std::endl;

    // Set auxiliary components of the preconditioner restart config
    conf.set("preconditioner_aux", auxconf);
  }
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void SpectralSqrtLMP<MODEL>::restart(const eckit::Configuration & conf) {
  if (active_) {
    ASSERT(conf.has("preconditioner") && conf.has("preconditioner_aux"));

    // Restart preconditioning vectors
    std::vector<eckit::LocalConfiguration> precvecs;
    conf.get("preconditioner", precvecs);  
    for (size_t jj = 0; jj < precvecs.size(); ++jj) {
      boost::shared_ptr<CtrlVec_> u(new CtrlVec_(J_.jb()));
      u->read(precvecs[jj]);
      U_.push_back(u);
      Log::info() << " " << classname() << "::restart lmp[" << jj << "]: "
                  << util::full_precision(dot_product(*U_[jj], *U_[jj]))
                  << std::endl;
    }

    // Restart auxiliary data 
    eckit::LocalConfiguration auxconf = conf.getSubConfiguration("preconditioner_aux");
    eigvals_ = auxconf.getDoubleVector("eigvals"); 
    usedpairIndx_ = auxconf.getUnsignedVector("usedpairIndx");
    update_ = auxconf.getUnsigned("update");

    Log::info() << " " << classname() << "::restart eigvals: "
                << util::full_precision(eigvals_)      << std::endl;
    Log::info() << " " << classname() << "::restart usedpairIndx: " 
                << util::full_precision(usedpairIndx_) << std::endl;
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_SPECTRALSQRTLMP_H_
