/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_ASSIMILATION_UTHTRINVHUDXFG_H_
#define OOPS_ASSIMILATION_UTHTRINVHUDXFG_H_

#include <boost/noncopyable.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>

#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/ControlIncrement.h"
#include "oops/assimilation/DualVector.h"
#include "oops/base/PostProcessorTL.h"
#include "oops/base/PostProcessorAD.h"
#include "oops/interface/Increment.h"
#include "util/dot_product.h"
#include "util/formats.h"
#include "util/Logger.h"
#include "util/PrintAdjTest.h"

namespace oops {

/// Calculate \f$ U^T H^T R^{-1} H dx_FG \f$.

template<typename MODEL> class UtHtRinvHdxFG : private boost::noncopyable {
  typedef Increment<MODEL>        Increment_;
  typedef ControlIncrement<MODEL> CtrlInc_;
  typedef ControlVector<MODEL>    CtrlVec_;
  typedef CostFunction<MODEL>     CostFct_;

 public:
  explicit UtHtRinvHdxFG(const CostFct_ & j);

  void compute(CtrlVec_ & dz) const;

 private:
  CostFct_ const & j_;
  mutable CtrlInc_ dx_;
  mutable CtrlInc_ mdx_;
  mutable CtrlInc_ dz_;
};

// -----------------------------------------------------------------------------

template<typename MODEL> 
UtHtRinvHdxFG<MODEL>::UtHtRinvHdxFG(const CostFct_ & j)
  : j_(j),
    dx_(j.jb().getFirstGuessIncrement(), true),
    mdx_(j.jb().getFirstGuessIncrement(), false),
    dz_(j.jb().getFirstGuessIncrement(), false)
{}

// -----------------------------------------------------------------------------

template<typename MODEL> 
void UtHtRinvHdxFG<MODEL>::compute(CtrlVec_ & dvout) const {
// Setup TL terms of cost function
  PostProcessorTL<Increment_> costtl;
  for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
    costtl.enrollProcessor(j_.jterm(jj).setupTL(dx_));
  }

// Run TLM
  mdx_ = dx_;
  j_.runTLM(mdx_, costtl);

// Get TLM outputs, multiply by covariance inverses, and setup ADJ forcing terms
  j_.zeroAD(dz_);
  PostProcessorAD<Increment_> costad;

  DualVector<MODEL> ww;
  DualVector<MODEL> zz;
  
  for (unsigned jj = 0; jj < j_.nterms(); ++jj) {
    ww.append(costtl.releaseOutputFromTL(jj));
    zz.append(j_.jterm(jj).multiplyCoInv(*ww.getv(jj)));
    costad.enrollProcessor(j_.jterm(jj).setupAD(zz.getv(jj), dz_));
  }

// Run ADJ
  j_.runADJ(dz_, costad);

// Multiply by the transpose of the square root of the covariance
  j_.jb().multiplyBsqrtTrans(dz_, dvout);
  
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_ASSIMILATION_UTHTRINVHUDXFG_H_
