/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_DEPARTURES_H_
#define OOPS_BASE_DEPARTURES_H_

#include <cstddef>
#include <iostream>
#include <string>
#include <vector>

#include <boost/shared_ptr.hpp>

#include "oops/assimilation/ControlObsVector.h"
#include "oops/base/GeneralizedDepartures.h"
#include "oops/base/ObsSpaces.h"
#include "oops/interface/ObsAuxIncrement.h"
#include "oops/interface/ObsErrorBase.h"
#include "oops/interface/ObsVector.h"
#include "util/Logger.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/Printable.h"
#include "util/dot_product.h"

namespace oops {

template<typename MODEL> class Departures;

/// Difference between two observation vectors.
/*!
 * A departure is the difference between two observations.
 * The archetypal example is \f$ \mathbf{y} - {\cal H}(\mathbf{x}) \f$.
 *
 * Keeping an observation space vector here is necessary for the implementation
 * of generic observation error covariance matrices.
 */

// -----------------------------------------------------------------------------
template <typename MODEL>
class Departures : public util::Printable,
                   public GeneralizedDepartures {
  typedef ControlObsVector<MODEL>    ControlObsVector_;
  typedef ObsAuxIncrement<MODEL>     ObsAuxIncr_;
  typedef ObsErrorBase<MODEL>        ObsErrorBase_;
  typedef ObsSpaces<MODEL>           ObsSpaces_;

 public:
// Constructors and destructor
  explicit Departures(const ObsSpaces_ &);
  explicit Departures(std::vector<boost::shared_ptr<ControlObsVector_> >);
  explicit Departures(const Departures &);
  ~Departures();

/// Access
  std::size_t size() const {return dep_.size();}
  ControlObsVector_ & operator[](const std::size_t ii) {return *dep_.at(ii);} 
  const ControlObsVector_ & operator[](const std::size_t ii) const {return *dep_.at(ii);} 

// Linear algebra operators
  Departures & operator=(const Departures &);
  Departures & operator+=(const Departures &);
  Departures & operator-=(const Departures &);
  Departures & operator*=(const double &);
  Departures & operator*=(const Departures &);
  Departures & operator/=(const Departures &);
  void zero();
  void invert();
  void axpy(const double &, const Departures &);
  double dot_product_with(const Departures &) const;

/// Get departue values
  const ControlObsVector_ & depvalues() const {return *dep_;}

/// I/O
  void save(const eckit::Configuration &) const;
  void read(const eckit::Configuration &);

 private:
  void print(std::ostream &) const;

  std::vector<boost::shared_ptr<ControlObsVector_> > dep_;
};

// =============================================================================

template<typename MODEL>
Departures<MODEL>::Departures(const ObsSpaces_ & obsgeom)
 : dep_(0)
{
  for (std::size_t jj = 0; jj < obsgeom.size(); ++jj) {
    boost::shared_ptr<ControlObsVector_> tmp(new ControlObsVector_(obsgeom[jj]));
    dep_.push_back(tmp);
  }
  Log::trace() << "Departures created" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL>::Departures(std::vector<boost::shared_ptr<ControlObsVector_> > val)
 : dep_(val)
{
  Log::trace() << "Departures created" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL>::Departures(const Departures & other)
 : dep_(0)
{
// We want deep copy here
  for (std::size_t jj = 0; jj < other.dep_.size(); ++jj) {
    boost::shared_ptr<ControlObsVector_> tmp(new ControlObsVector_(*other.dep_[jj]));
    dep_.push_back(tmp);
  }
  Log::trace() << "Departures copy-created" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL>::~Departures() {
  Log::trace() << "Departures destructed" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL> & Departures<MODEL>::operator=(const Departures & rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] = *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL> & Departures<MODEL>::operator+=(const Departures & rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] += *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL> & Departures<MODEL>::operator-=(const Departures & rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] -= *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL> & Departures<MODEL>::operator*=(const double & zz) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] *= zz;
  }
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL> & Departures<MODEL>::operator*=(const Departures & rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] *= *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
Departures<MODEL> & Departures<MODEL>::operator/=(const Departures & rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    *dep_[jj] /= *rhs.dep_[jj];
  }
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void Departures<MODEL>::zero() {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->zero();
  }
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void Departures<MODEL>::invert() {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->invert();
  }
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void Departures<MODEL>::axpy(const double & zz, const Departures & rhs) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->axpy(zz, *rhs.dep_[jj]);
  }
}
// -----------------------------------------------------------------------------
template<typename MODEL>
double Departures<MODEL>::dot_product_with(const Departures & other) const {
  double zz = 0.0;
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    zz += dot_product(*dep_[jj], *other.dep_[jj]);
  }
  return zz;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Departures<MODEL>::save(const eckit::Configuration & conf) const {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->save(conf);
  }
  Log::trace() << "Departures:Departures have been saved: " << conf << std::endl;
}
// -----------------------------------------------------------------------------
template <typename MODEL>
void Departures<MODEL>::read(const eckit::Configuration & conf) {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    dep_[jj]->read(conf);
  }
  Log::trace() << "Departures:Departures have been read: " << conf << std::endl;
}
// -----------------------------------------------------------------------------

template <typename MODEL>
void Departures<MODEL>::print(std::ostream & os) const {
  for (std::size_t jj = 0; jj < dep_.size(); ++jj) {
    os << *dep_[jj];
  }
}
// -----------------------------------------------------------------------------
}  // namespace oops

#endif  // OOPS_BASE_DEPARTURES_H_
