/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSSPACES_H_
#define OOPS_BASE_OBSSPACES_H_

#include <cstddef>
#include <ostream>
#include <string>
#include <vector>

#include <boost/noncopyable.hpp>
#include <boost/shared_ptr.hpp>

#include "eckit/config/LocalConfiguration.h"
#include "oops/interface/ObservationSpace.h"
#include "util/DateTime.h"
#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {
  template <typename T> class Departures;
  template <typename T> class GeoVaLs;

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsSpaces : public util::Printable,
                  private boost::noncopyable,
                  private util::ObjectCounter<ObsSpaces<MODEL> > {
  typedef Departures<MODEL>         Departures_;
  typedef GeoVaLs<MODEL>            GeoVaLs_;
  typedef ObservationSpace<MODEL>   ObsSpace_;

  typedef std::vector<boost::shared_ptr<GeoVaLs_> > GeoVaLsVec_;
 public:
  static const std::string classname() {return "oops::ObsSpaces";}

  ObsSpaces(const eckit::Configuration &, 
            const util::DateTime &, 
            const util::DateTime &,
            const bool lscreened =false);
  ~ObsSpaces();

/// Access
  std::size_t size() const {return spaces_.size();}
  const ObsSpace_ & operator[](const std::size_t ii) const {return *spaces_.at(ii);} 
  ObsSpace_ & operator[](const std::size_t ii) {return *spaces_.at(ii);} 

/// Assimilation window
  const util::DateTime & windowStart() const {return wbgn_;}
  const util::DateTime & windowEnd() const {return wend_;}

/// Screen observations  
  void screenObservations(const Departures_ &,
                          const GeoVaLsVec_ &) const;

/// Other
  const bool & isScreened() const {return lscreened_;}
  void printJo(const Departures_ &, const Departures_ &) const;  // To be changed
  void write(const eckit::Configuration &) const {}
  void read(const eckit::Configuration &) {}

 private:
  void print(std::ostream &) const;

  const util::DateTime wbgn_;
  const util::DateTime wend_;
  const bool lscreened_;
  std::vector<boost::shared_ptr<ObsSpace_> > spaces_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsSpaces<MODEL>::ObsSpaces(const eckit::Configuration & conf,
                            const util::DateTime & bgn, 
                            const util::DateTime & end,
                            const bool lscreened)
 : wbgn_(bgn), wend_(end), lscreened_(lscreened), spaces_(0)
{
  std::vector<eckit::LocalConfiguration> obsconf;
  conf.get("ObsTypes", obsconf);
  for (std::size_t jj = 0; jj < obsconf.size(); ++jj) {
    Log::debug() << "ObsSpaces::ObsSpaces : conf " << obsconf[jj] << std::endl;
    boost::shared_ptr<ObsSpace_> tmp(new ObsSpace_(obsconf[jj], bgn, end, lscreened));
    spaces_.push_back(tmp);
//  Generate locations etc... if required
    if (obsconf[jj].has("Generate")) {
      const eckit::LocalConfiguration gconf(obsconf[jj], "Generate");
      spaces_[jj]->generateDistribution(gconf);
    }
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsSpaces<MODEL>::~ObsSpaces() {}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsSpaces<MODEL>::screenObservations(const Departures_ & dep,
                                          const GeoVaLsVec_ & geovals) const {
  for (std::size_t jj = 0; jj < spaces_.size(); ++jj) {
    spaces_[jj]->screenObservations(dep[jj], *geovals[jj]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsSpaces<MODEL>::printJo(const Departures_ & dy, const Departures_ & grad) const {
  for (std::size_t jj = 0; jj < spaces_.size(); ++jj) {
    spaces_[jj]->printJo(dy[jj], grad[jj]);
  }
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsSpaces<MODEL>::print(std::ostream & os) const {
  for (std::size_t jj = 0; jj < spaces_.size(); ++jj) {
    os << *spaces_[jj];
  }
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSSPACES_H_
