/*
 * (C) Copyright 2009-2016 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_BASE_OBSERVERTL_H_
#define OOPS_BASE_OBSERVERTL_H_

#include <memory>

#include <boost/ptr_container/ptr_map.hpp>
#include <boost/ptr_container/ptr_vector.hpp>

#include "oops/base/Departures.h"
#include "oops/base/LinearObsOperators.h"
#include "oops/base/ObsSpaces.h"
#include "oops/base/PostBaseTL.h"
#include "oops/interface/GeoVaLs.h"
#include "oops/interface/Locations.h"
#include "oops/interface/ObsAuxIncrement.h"
#include "util/DateTime.h"
#include "util/Duration.h"

namespace oops {

/// Computes observation equivalent TL and AD to/from increments.

template <typename MODEL, typename INCR> class ObserverTL : public PostBaseTL<INCR> {
  typedef Departures<MODEL>          Departures_;
  typedef GeoVaLs<MODEL>             GeoVaLs_;
  typedef LinearObsOperators<MODEL>  LinearObsOperators_;
  typedef Locations<MODEL>           Locations_;
  typedef ObsAuxIncrement<MODEL>     ObsAuxIncr_;
  typedef ObsSpaces<MODEL>           ObsSpaces_;

 public:
  ObserverTL(const ObsSpaces_ &, const LinearObsOperators_ &, const ObsAuxIncr_ &,
             const util::Duration &, const bool subwin = false);
  ~ObserverTL() {}

  Departures_ * releaseOutputFromTL() override {return new Departures_(ydep_);}
  void setupTL(const ObsAuxIncr_ & dybiasparam) {dybiasparam_ = &dybiasparam;}

 private:
// Methods
  void doInitializeTL(const INCR &, const util::DateTime &, const util::Duration &) override;
  void doProcessingTL(const INCR &) override;
  void doFinalizeTL(const INCR &) override;

// Obs operator
  const ObsSpaces_ & obspaces_;
  const LinearObsOperators_ & hoptlad_;

// Data
  Departures_ ydep_;
  mutable const ObsAuxIncr_ * dybiasparam_;

  util::DateTime winbgn_;   //!< Begining of assimilation window
  util::DateTime winend_;   //!< End of assimilation window
  util::DateTime bgn_;      //!< Begining of currently active observations
  util::DateTime end_;      //!< End of currently active observations
  util::Duration hslot_;    //!< Half time slot
  const bool subwindows_;

  typedef typename boost::ptr_map<util::DateTime, boost::ptr_vector<GeoVaLs_> > gvalsMap_;
  typedef typename gvalsMap_::iterator gvalsIter_;
  gvalsMap_ gvals_;
};

// ====================================================================================

template <typename MODEL, typename INCR>
ObserverTL<MODEL, INCR>::ObserverTL(const ObsSpaces_ & obsdb,
                                    const LinearObsOperators_ & hoptlad,
                                    const ObsAuxIncr_ & dybiasparam,
                                    const util::Duration & tslot, const bool subwin)
  : PostBaseTL<INCR>(obsdb.windowStart(), obsdb.windowEnd()),
    obspaces_(obsdb), hoptlad_(hoptlad),
    ydep_(obspaces_), dybiasparam_(&dybiasparam),
    winbgn_(obsdb.windowStart()), winend_(obsdb.windowEnd()),
    bgn_(winbgn_), end_(winend_), hslot_(tslot/2), subwindows_(subwin)
{}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverTL<MODEL, INCR>::doInitializeTL(const INCR & dx,
                 const util::DateTime & end, const util::Duration & tstep) {
  const util::DateTime bgn(dx.validTime());
  if (hslot_ == util::Duration(0)) hslot_ = tstep/2;
  if (subwindows_) {
    if (bgn == end) {
      bgn_ = bgn - hslot_;
      end_ = end + hslot_;
    } else {
      bgn_ = bgn;
      end_ = end;
    }
  }
  if (bgn_ < winbgn_) bgn_ = winbgn_;
  if (end_ > winend_) end_ = winend_;

  gvalsIter_ it = gvals_.find(bgn_);
  if (it == gvals_.end()) {
    boost::ptr_vector<GeoVaLs_> * gvalsvec = new boost::ptr_vector<GeoVaLs_>();
    for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
      gvalsvec->push_back(new GeoVaLs_(obspaces_[jj], hoptlad_.variables(jj), 
                                       dx, bgn_, end_));
    }
    gvals_.insert(bgn_, gvalsvec);
    it = gvals_.find(bgn_);
    ASSERT(it != gvals_.end());
  }
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    (*it->second)[jj].zero(); 
  } 
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverTL<MODEL, INCR>::doProcessingTL(const INCR & dx) {
  util::DateTime t1(dx.validTime()-hslot_);
  util::DateTime t2(dx.validTime()+hslot_);
  if (t1 < bgn_) t1 = bgn_;
  if (t2 > end_) t2 = end_;

  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
//  Get locations info for interpolator
    Locations_ locs(obspaces_[jj], t1, t2);

//  Interpolate state variables to obs locations
    gvalsIter_ it = gvals_.find(bgn_);
    ASSERT(it != gvals_.end());
    dx.interpolateTL(locs, (*it->second)[jj]);
  }
}
// -----------------------------------------------------------------------------
template <typename MODEL, typename INCR>
void ObserverTL<MODEL, INCR>::doFinalizeTL(const INCR &) {
  gvalsIter_ it = gvals_.find(bgn_);
  ASSERT(it != gvals_.end());
  for (std::size_t jj = 0; jj < obspaces_.size(); ++jj) {
    // H(dx) + P(dB)
    hoptlad_[jj].obsEquivTL((*it->second)[jj],
                            ydep_[jj].obsvector(),
                            *dybiasparam_);
    // P(dB)
    if (ydep_[jj].hasBias())
      hoptlad_[jj].obsBiasEquivTL((*it->second)[jj],
                                  ydep_[jj].biasvector(),
                                  *dybiasparam_);
  }
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_BASE_OBSERVERTL_H_
