/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_INCRENSCTLVEC_H_
#define OOPS_INTERFACE_INCRENSCTLVEC_H_

#include <string>

#include <boost/scoped_ptr.hpp>

#include "oops/interface/Geometry.h"
#include "oops/interface/Localization.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

template <typename MODEL> class Localization;

  /// IncrEnsCtlVec Class
  /*!
   *  Augmented part of the control vector
   */

// -----------------------------------------------------------------------------

template <typename MODEL>
class IncrEnsCtlVec : public util::Printable,
                      private util::ObjectCounter<IncrEnsCtlVec<MODEL> > {
  typedef typename MODEL::IncrEnsCtlVec IncrEnsCtlVec_;
  typedef Localization<MODEL>           Localization_;
  typedef Geometry<MODEL>               Geometry_;

 public:
  static const std::string classname() {return "oops::IncrEnsCtlVec";}

/// Constructor, destructor
  IncrEnsCtlVec() : incrEnsCtlVec_(0) {}
  IncrEnsCtlVec(const Geometry_ &, const Localization_ &);
  IncrEnsCtlVec(const Geometry_ &, const Localization_ &, const IncrEnsCtlVec &);
  IncrEnsCtlVec(const IncrEnsCtlVec &, const bool copy = true);
  virtual ~IncrEnsCtlVec();

/// Linear algebra operators
  void zero();
  IncrEnsCtlVec & operator =(const IncrEnsCtlVec &);
  IncrEnsCtlVec & operator+=(const IncrEnsCtlVec &);
  IncrEnsCtlVec & operator-=(const IncrEnsCtlVec &);
  IncrEnsCtlVec & operator*=(const double &);
  void axpy(const double &, const IncrEnsCtlVec &, const bool check = true);
  double dot_product_with(const IncrEnsCtlVec &) const;
  void schur_product_with(const IncrEnsCtlVec &);
  void random();

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

/// Interfacing
  IncrEnsCtlVec_ & ctlVec() {return *incrEnsCtlVec_;}
  const IncrEnsCtlVec_ & ctlVec() const {return *incrEnsCtlVec_;}

 private:
  void print(std::ostream &) const;
  boost::scoped_ptr<IncrEnsCtlVec_> incrEnsCtlVec_;
};

// =============================================================================
/// Constructor, destructor
// -----------------------------------------------------------------------------

template<typename MODEL>
IncrEnsCtlVec<MODEL>::IncrEnsCtlVec(const Geometry_ & resol, 
                                    const Localization_ & loc)
  : incrEnsCtlVec_(0)
{
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec starting" << std::endl;
  util::Timer timer(classname(), "Incr");
  incrEnsCtlVec_.reset(new IncrEnsCtlVec_(resol.geometry(), 
                                          loc.localization()));
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrEnsCtlVec<MODEL>::IncrEnsCtlVec(const Geometry_ & resol, 
                                    const Localization_ & loc,
                                    const IncrEnsCtlVec & other)
  : incrEnsCtlVec_(0)
{
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec starting" << std::endl;
  util::Timer timer(classname(), "Incr");
  if (other.incrEnsCtlVec_ != 0) {
    incrEnsCtlVec_.reset(new IncrEnsCtlVec_(resol.geometry(), 
                                            loc.localization(),
                                            *other.incrEnsCtlVec_));
  }
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrEnsCtlVec<MODEL>::IncrEnsCtlVec(const IncrEnsCtlVec & other, 
                                    const bool copy) 
  : incrEnsCtlVec_(0)
{
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec copy starting" << std::endl;
  util::Timer timer(classname(), "Incr");
  if (other.incrEnsCtlVec_ != 0) incrEnsCtlVec_.reset(new IncrEnsCtlVec_(*other.incrEnsCtlVec_, copy));
  Log::trace() << "IncrEnsCtlVec<MODEL>::IncrEnsCtlVec copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrEnsCtlVec<MODEL>::~IncrEnsCtlVec() {
  Log::trace() << "IncrEnsCtlVec<MODEL>::~IncrEnsCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~Incr");
  incrEnsCtlVec_.reset();
  Log::trace() << "IncrEnsCtlVec<MODEL>::~IncrEnsCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrEnsCtlVec<MODEL>::zero() {
  Log::trace() << "IncrEnsCtlVec<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  if (incrEnsCtlVec_ != 0) incrEnsCtlVec_->zero();
  Log::trace() << "IncrEnsCtlVec<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrEnsCtlVec<MODEL> & IncrEnsCtlVec<MODEL>::operator=(const IncrEnsCtlVec & rhs) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  if (incrEnsCtlVec_ != 0) *incrEnsCtlVec_ = *rhs.incrEnsCtlVec_;
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrEnsCtlVec<MODEL> & IncrEnsCtlVec<MODEL>::operator+=(const IncrEnsCtlVec & rhs) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  if (incrEnsCtlVec_ != 0) *incrEnsCtlVec_ += *rhs.incrEnsCtlVec_;
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator+= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrEnsCtlVec<MODEL> & IncrEnsCtlVec<MODEL>::operator-=(const IncrEnsCtlVec & rhs) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  if (incrEnsCtlVec_ != 0) *incrEnsCtlVec_ -= *rhs.incrEnsCtlVec_;
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator-= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrEnsCtlVec<MODEL> & IncrEnsCtlVec<MODEL>::operator*=(const double & zz) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  if (incrEnsCtlVec_ != 0) *incrEnsCtlVec_ *= zz;
  Log::trace() << "IncrEnsCtlVec<MODEL>::operator*= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrEnsCtlVec<MODEL>::axpy(const double & zz, const IncrEnsCtlVec & dx, 
                                  const bool check) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  if (incrEnsCtlVec_ != 0) incrEnsCtlVec_->axpy(zz, *dx.incrEnsCtlVec_, check);
  Log::trace() << "IncrEnsCtlVec<MODEL>::axpy done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double IncrEnsCtlVec<MODEL>::dot_product_with(const IncrEnsCtlVec & dx) const {
  Log::trace() << "IncrEnsCtlVec<MODEL>::dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "dot_product_with");
  double zz = 0;
  if (incrEnsCtlVec_ != 0) zz = incrEnsCtlVec_->dot_product_with(*dx.incrEnsCtlVec_);
  Log::trace() << "IncrEnsCtlVec<MODEL>::dot_product_with done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrEnsCtlVec<MODEL>::schur_product_with(const IncrEnsCtlVec & dx) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::schur_product_with starting" << std::endl;
  util::Timer timer(classname(), "schur_product_with");
  if (incrEnsCtlVec_ != 0) incrEnsCtlVec_->schur_product_with(*dx.incrEnsCtlVec_);
  Log::trace() << "IncrEnsCtlVec<MODEL>::schur_product_with done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrEnsCtlVec<MODEL>::random() {
  Log::trace() << "IncrEnsCtlVec<MODEL>::random starting" << std::endl;
  util::Timer timer(classname(), "random");
  if (incrEnsCtlVec_ != 0) incrEnsCtlVec_->random();
  Log::trace() << "IncrEnsCtlVec<MODEL>::random done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrEnsCtlVec<MODEL>::read(const eckit::Configuration & conf) {
  Log::trace() << "IncrEnsCtlVec<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  if (incrEnsCtlVec_ != 0) incrEnsCtlVec_->read(conf);
  Log::trace() << "IncrEnsCtlVec<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrEnsCtlVec<MODEL>::write(const eckit::Configuration & conf) const {
  Log::trace() << "IncrEnsCtlVec<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  if (incrEnsCtlVec_ != 0) incrEnsCtlVec_->write(conf);
  Log::trace() << "IncrEnsCtlVec<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double IncrEnsCtlVec<MODEL>::norm() const {
  Log::trace() << "IncrEnsCtlVec<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = 0;
  if (incrEnsCtlVec_ != 0) incrEnsCtlVec_->norm();
  Log::trace() << "IncrEnsCtlVec<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrEnsCtlVec<MODEL>::print(std::ostream & os) const {
  Log::trace() << "IncrEnsCtlVec<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  if (incrEnsCtlVec_ != 0) os << *incrEnsCtlVec_;
  Log::trace() << "IncrEnsCtlVec<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_INCRENSCTLVEC_H_
