/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_INCRMODCTLVEC_H_
#define OOPS_INTERFACE_INCRMODCTLVEC_H_

#include <string>

#include <boost/scoped_ptr.hpp>

#include "oops/interface/ErrorCovariance.h"
#include "util/DateTime.h"
#include "util/Duration.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace oops {

  /// IncrModCtlVec Class
  /*!
   *  Increment part of the control vector (acted upon by the static background
   *  error covariance matrix)
   */

// -----------------------------------------------------------------------------

template <typename MODEL>
class IncrModCtlVec : public util::Printable,
                      private util::ObjectCounter<IncrModCtlVec<MODEL> > {
  typedef typename MODEL::IncrModCtlVec IncrModCtlVec_;
  typedef ErrorCovariance<MODEL>        ErrorCovariance_;

 public:
  static const std::string classname() {return "oops::IncrModCtlVec";}

/// Constructor, destructor
  IncrModCtlVec();
  IncrModCtlVec(const ErrorCovariance_ &);
  IncrModCtlVec(const ErrorCovariance_ &, const IncrModCtlVec &);
  IncrModCtlVec(const IncrModCtlVec &, const bool copy = true);
  virtual ~IncrModCtlVec();

/// Linear algebra operators
  void zero();
  IncrModCtlVec & operator =(const IncrModCtlVec &);
  IncrModCtlVec & operator+=(const IncrModCtlVec &);
  IncrModCtlVec & operator-=(const IncrModCtlVec &);
  IncrModCtlVec & operator*=(const double &);
  void axpy(const double &, const IncrModCtlVec &, const bool check = true);
  double dot_product_with(const IncrModCtlVec &) const;
  void schur_product_with(const IncrModCtlVec &);
  void random();

/// Section
  void getSection(std::vector<double> &,
                  const int &,
                  const int &) const;
  void setSection(const std::vector<double> &,
                  const int &,
                  const int &);

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

/// Interfacing
  IncrModCtlVec_ & ctlVec() {return *incrModCtlVec_;}
  const IncrModCtlVec_ & ctlVec() const {return *incrModCtlVec_;}

 private:
  void print(std::ostream &) const;
  boost::scoped_ptr<IncrModCtlVec_> incrModCtlVec_;
};

// =============================================================================
/// Constructor, destructor
// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec()
  : incrModCtlVec_(0)
{
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec(const ErrorCovariance_ & cov)
  : incrModCtlVec_(0)
{
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec starting" << std::endl;
  util::Timer timer(classname(), "Incr");
  incrModCtlVec_.reset(new IncrModCtlVec_(cov.errorCovariance()));
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------


template<typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec(const ErrorCovariance_ & cov,
                                    const IncrModCtlVec & other)
  : incrModCtlVec_(0)
{
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec starting" << std::endl;
  util::Timer timer(classname(), "Incr");
  if (other.incrModCtlVec_ != 0) {
    incrModCtlVec_.reset(new IncrModCtlVec_(cov.errorCovariance(),
                                            *other.incrModCtlVec_));
  }
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL>::IncrModCtlVec(const IncrModCtlVec & other, 
                                    const bool copy) 
  : incrModCtlVec_(0)
{
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec copy starting" << std::endl;
  util::Timer timer(classname(), "Incr");
  if (other.incrModCtlVec_ != 0) {
    incrModCtlVec_.reset(new IncrModCtlVec_(*other.incrModCtlVec_, copy));
  }
  Log::trace() << "IncrModCtlVec<MODEL>::IncrModCtlVec copy done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL>::~IncrModCtlVec() {
  Log::trace() << "IncrModCtlVec<MODEL>::~IncrModCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~Incr");
  incrModCtlVec_.reset();
  Log::trace() << "IncrModCtlVec<MODEL>::~IncrModCtlVec done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::zero() {
  Log::trace() << "IncrModCtlVec<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  if (incrModCtlVec_ != 0) incrModCtlVec_->zero();
  Log::trace() << "IncrModCtlVec<MODEL>::zero done" << std::endl;
}

// -----------------------------------------------------------------------------

//template<typename MODEL>
//void IncrModCtlVec<MODEL>::zero(const util::DateTime & tt) {
//  Log::trace() << "IncrModCtlVec<MODEL>::zero starting" << std::endl;
//  util::Timer timer(classname(), "zero");
//  incrModCtlVec_->zero(tt);
//  Log::trace() << "IncrModCtlVec<MODEL>::zero done" << std::endl;
//}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL> & IncrModCtlVec<MODEL>::operator=(const IncrModCtlVec & rhs) {
  Log::trace() << "IncrModCtlVec<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  if (incrModCtlVec_ != 0) *incrModCtlVec_ = *rhs.incrModCtlVec_;
  Log::trace() << "IncrModCtlVec<MODEL>::operator= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL> & IncrModCtlVec<MODEL>::operator+=(const IncrModCtlVec & rhs) {
  Log::trace() << "IncrModCtlVec<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  if (incrModCtlVec_ != 0) *incrModCtlVec_ += *rhs.incrModCtlVec_;
  Log::trace() << "IncrModCtlVec<MODEL>::operator+= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL> & IncrModCtlVec<MODEL>::operator-=(const IncrModCtlVec & rhs) {
  Log::trace() << "IncrModCtlVec<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  if (incrModCtlVec_ != 0) *incrModCtlVec_ -= *rhs.incrModCtlVec_;
  Log::trace() << "IncrModCtlVec<MODEL>::operator-= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
IncrModCtlVec<MODEL> & IncrModCtlVec<MODEL>::operator*=(const double & zz) {
  Log::trace() << "IncrModCtlVec<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  if (incrModCtlVec_ != 0) *incrModCtlVec_ *= zz;
  Log::trace() << "IncrModCtlVec<MODEL>::operator*= done" << std::endl;
  return *this;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::axpy(const double & zz, const IncrModCtlVec & dx, 
                                  const bool check) {
  Log::trace() << "IncrModCtlVec<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  if (incrModCtlVec_ != 0) incrModCtlVec_->axpy(zz, *dx.incrModCtlVec_, check);
  Log::trace() << "IncrModCtlVec<MODEL>::axpy done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double IncrModCtlVec<MODEL>::dot_product_with(const IncrModCtlVec & dx) const {
  Log::trace() << "IncrModCtlVec<MODEL>::dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "dot_product_with");
  double zz = 0.0;
  if (incrModCtlVec_ != 0) zz += incrModCtlVec_->dot_product_with(*dx.incrModCtlVec_);
  Log::trace() << "IncrModCtlVec<MODEL>::dot_product_with done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::schur_product_with(const IncrModCtlVec & dx) {
  Log::trace() << "IncrModCtlVec<MODEL>::schur_product_with starting" << std::endl;
  util::Timer timer(classname(), "schur_product_with");
  if (incrModCtlVec_ != 0) incrModCtlVec_->schur_product_with(*dx.incrModCtlVec_);
  Log::trace() << "IncrModCtlVec<MODEL>::schur_product_with done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::random() {
  Log::trace() << "IncrModCtlVec<MODEL>::random starting" << std::endl;
  util::Timer timer(classname(), "random");
  if (incrModCtlVec_ != 0) incrModCtlVec_->random();
  Log::trace() << "IncrModCtlVec<MODEL>::random done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::getSection(std::vector<double> & section,
                                      const int & iistart,
                                      const int & iiend) const {
  Log::trace() << "IncrModCtlVec<MODEL>::getSection starting" << std::endl;
  util::Timer timer(classname(), "getSection");
  if (incrModCtlVec_ != 0) incrModCtlVec_->getSection(section, iistart, iiend);
  Log::trace() << "IncrModCtlVec<MODEL>::getSection done" << std::endl;

}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::setSection(const std::vector<double> & section,
                const int & iistart,
                const int & iiend) {
  Log::trace() << "IncrModCtlVec<MODEL>::setSection starting" << std::endl;
  util::Timer timer(classname(), "setSection");
  if (incrModCtlVec_ != 0) incrModCtlVec_->setSection(section, iistart, iiend);
  Log::trace() << "IncrModCtlVec<MODEL>::setSection done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::read(const eckit::Configuration & conf) {
  Log::trace() << "IncrModCtlVec<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  if (incrModCtlVec_ != 0) incrModCtlVec_->read(conf);
  Log::trace() << "IncrModCtlVec<MODEL>::read done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::write(const eckit::Configuration & conf) const {
  Log::trace() << "IncrModCtlVec<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  if (incrModCtlVec_ != 0) incrModCtlVec_->write(conf);
  Log::trace() << "IncrModCtlVec<MODEL>::write done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
double IncrModCtlVec<MODEL>::norm() const {
  Log::trace() << "IncrModCtlVec<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = 0.0;
  if (incrModCtlVec_ != 0) zz += incrModCtlVec_->norm();
  Log::trace() << "IncrModCtlVec<MODEL>::norm done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void IncrModCtlVec<MODEL>::print(std::ostream & os) const {
  Log::trace() << "IncrModCtlVec<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  if (incrModCtlVec_ != 0) os << *incrModCtlVec_;
  Log::trace() << "IncrModCtlVec<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_INCRMODCTLVEC_H_
