/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_MODELAUXCTLVEC_H_
#define OOPS_INTERFACE_MODELAUXCTLVEC_H_

#include <iostream>
#include <string>

#include <boost/scoped_ptr.hpp>

#include "oops/interface/ModelAuxControl.h"

#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
}

namespace oops {

  /// ModelAuxCtlVec Class
  /*!
   *  Part of the control vector corresponding to the ModelAuxIncrement part of the
   *  control increment
   */

// -----------------------------------------------------------------------------

template <typename MODEL>
class ModelAuxCtlVec : public util::Printable,
                       private util::ObjectCounter<ModelAuxCtlVec<MODEL> > {
  typedef typename MODEL::ModelAuxCtlVec     ModelAuxCtlVec_;
  typedef typename MODEL::ModelAuxCovariance ModelAuxCovariance_;
  typedef ModelAuxControl<MODEL>             ModelAuxControl_;

 public:
  static const std::string classname() {return "oops::ModelAuxCtlVec";}

/// Constructor, destructor
  ModelAuxCtlVec(const ModelAuxCovariance_ &);
  ModelAuxCtlVec(const ModelAuxCovariance_ &, const ModelAuxCtlVec &);
  ModelAuxCtlVec(const ModelAuxCtlVec &, const bool copy = true);
  ~ModelAuxCtlVec();

/// Interfacing
  const ModelAuxCtlVec_ & modelAuxCtlVec() const {return *aux_;}
  ModelAuxCtlVec_ & modelAuxCtlVec() {return *aux_;}

/// Linear algebra operators
  void zero();
  ModelAuxCtlVec & operator=(const ModelAuxCtlVec &);
  ModelAuxCtlVec & operator+=(const ModelAuxCtlVec &);
  ModelAuxCtlVec & operator-=(const ModelAuxCtlVec &);
  ModelAuxCtlVec & operator*=(const double &);
  void axpy(const double &, const ModelAuxCtlVec &);
  double dot_product_with(const ModelAuxCtlVec &) const;

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

 private:
  void print(std::ostream &) const;
  boost::scoped_ptr<ModelAuxCtlVec_> aux_;
};

// -----------------------------------------------------------------------------
template<typename MODEL>
ModelAuxCtlVec<MODEL>::ModelAuxCtlVec(const ModelAuxCovariance_ & cov) 
  : aux_()
{
  Log::trace() << "ModelAuxCtlVec<MODEL>::ModelAuxCtlVec starting" << std::endl;
  util::Timer timer(classname(), "ModelAuxCtlVec");
  aux_.reset(new ModelAuxCtlVec_(cov));
  Log::trace() << "ModelAuxCtlVec<MODEL>::ModelAuxCtlVec done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ModelAuxCtlVec<MODEL>::ModelAuxCtlVec(const ModelAuxCovariance_ & cov,
                                      const ModelAuxCtlVec & other) 
  : aux_()
{
  Log::trace() << "ModelAuxCtlVec<MODEL>::ModelAuxCtlVec interpolated starting" << std::endl;
  util::Timer timer(classname(), "ModelAuxCtlVec");
  aux_.reset(new ModelAuxCtlVec_(cov, *other.aux_));
  Log::trace() << "ModelAuxCtlVec<MODEL>::ModelAuxCtlVec interpolated done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ModelAuxCtlVec<MODEL>::ModelAuxCtlVec(const ModelAuxCtlVec & other,
                                      const bool copy) 
  : aux_()
{
  Log::trace() << "ModelAuxCtlVec<MODEL>::ModelAuxCtlVec copy starting" << std::endl;
  util::Timer timer(classname(), "ModelAuxCtlVec");
  aux_.reset(new ModelAuxCtlVec_(*other.aux_, copy));
  Log::trace() << "ModelAuxCtlVec<MODEL>::ModelAuxCtlVec copy done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ModelAuxCtlVec<MODEL>::~ModelAuxCtlVec() {
  Log::trace() << "ModelAuxCtlVec<MODEL>::~ModelAuxCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~ModelAuxCtlVec");
  aux_.reset();
  Log::trace() << "ModelAuxCtlVec<MODEL>::~ModelAuxCtlVec done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ModelAuxCtlVec<MODEL>::zero() {
  Log::trace() << "ModelAuxCtlVec<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  aux_->zero();
  Log::trace() << "ModelAuxCtlVec<MODEL>::zero done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ModelAuxCtlVec<MODEL> & ModelAuxCtlVec<MODEL>::operator=(const ModelAuxCtlVec & rhs) {
  Log::trace() << "ModelAuxCtlVec<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  *aux_ = *rhs.aux_;
  Log::trace() << "ModelAuxCtlVec<MODEL>::operator= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ModelAuxCtlVec<MODEL> & ModelAuxCtlVec<MODEL>::operator+=(const ModelAuxCtlVec & rhs) {
  Log::trace() << "ModelAuxCtlVec<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  *aux_ += *rhs.aux_;
  Log::trace() << "ModelAuxCtlVec<MODEL>::operator+= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ModelAuxCtlVec<MODEL> & ModelAuxCtlVec<MODEL>::operator-=(const ModelAuxCtlVec & rhs) {
  Log::trace() << "ModelAuxCtlVec<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  *aux_ -= *rhs.aux_;
  Log::trace() << "ModelAuxCtlVec<MODEL>::operator-= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ModelAuxCtlVec<MODEL> & ModelAuxCtlVec<MODEL>::operator*=(const double & zz) {
  Log::trace() << "ModelAuxCtlVec<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  *aux_ *= zz;
  Log::trace() << "ModelAuxCtlVec<MODEL>::operator*= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ModelAuxCtlVec<MODEL>::axpy(const double & zz, const ModelAuxCtlVec & dx) {
  Log::trace() << "ModelAuxCtlVec<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  aux_->axpy(zz, *dx.aux_);
  Log::trace() << "ModelAuxCtlVec<MODEL>::axpy done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
double ModelAuxCtlVec<MODEL>::dot_product_with(const ModelAuxCtlVec & dx) const {
  Log::trace() << "ModelAuxCtlVec<MODEL>::dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "dot_product_with");
  double zz = aux_->dot_product_with(*dx.aux_);
  Log::trace() << "ModelAuxCtlVec<MODEL>::dot_product_with done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ModelAuxCtlVec<MODEL>::read(const eckit::Configuration & conf) {
  Log::trace() << "ModelAuxCtlVec<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");
  aux_->read(conf);
  Log::trace() << "ModelAuxCtlVec<MODEL>::read done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ModelAuxCtlVec<MODEL>::write(const eckit::Configuration & conf) const {
  Log::trace() << "ModelAuxCtlVec<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");
  aux_->write(conf);
  Log::trace() << "ModelAuxCtlVec<MODEL>::write done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
double ModelAuxCtlVec<MODEL>::norm() const {
  Log::trace() << "ModelAuxCtlVec<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = aux_->norm();
  Log::trace() << "ModelAuxCtlVec<MODEL>::norm done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ModelAuxCtlVec<MODEL>::print(std::ostream & os) const {
  Log::trace() << "ModelAuxCtlVec<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *aux_;
  Log::trace() << "ModelAuxCtlVec<MODEL>::print done" << std::endl;
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_MODELAUXCTLVEC_H_
