/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXCOVARIANCE_H_
#define OOPS_INTERFACE_OBSAUXCOVARIANCE_H_

#include <iostream>
#include <string>

#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_map.hpp>
#include <boost/scoped_ptr.hpp>

#include "oops/interface/ObsAuxControl.h"
#include "oops/interface/ObsAuxCovarianceBase.h"
#include "oops/interface/ObsAuxCtlVec.h"
#include "oops/interface/ObsAuxIncrement.h"

#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCovariance : public util::Printable,
                         private boost::noncopyable,
                         private util::ObjectCounter<ObsAuxCovariance<MODEL> > {
  typedef ObsAuxCovarianceBase   ObsAuxCovarianceBase_;
  typedef ObsAuxControl<MODEL>   ObsAuxControl_;
  typedef ObsAuxIncrement<MODEL> ObsAuxIncrement_;
  typedef ObsAuxCtlVec<MODEL>    ObsAuxCtlVec_;

  typedef boost::ptr_map<std::string, ObsAuxCovarianceBase_> ObsAuxCovariancePtrMap_;
  typedef typename ObsAuxCovariancePtrMap_::iterator iter_;
  typedef typename ObsAuxCovariancePtrMap_::const_iterator icst_;

 public:
  static const std::string classname() {return "oops::ObsAuxCovariance";}

/// Constructor, destructor
  explicit ObsAuxCovariance(const eckit::Configuration &);
  ~ObsAuxCovariance();

/// Update
  void update(const ObsAuxControl_ &, const eckit::Configuration &);

/// Operators
  void linearize(const ObsAuxControl_ &);
  void multiply(const ObsAuxIncrement_ &, ObsAuxIncrement_ &) const;
  void inverseMultiply(const ObsAuxIncrement_ &, ObsAuxIncrement_ &) const;
  void multiplySqrt(const ObsAuxCtlVec_ &, ObsAuxIncrement_ &) const;
  void multiplySqrtTrans(const ObsAuxIncrement_ &, ObsAuxCtlVec_ &) const;
  void randomize(ObsAuxIncrement_ &) const;

// Interfacing
  const eckit::Configuration & config() const {return *conf_;}

  ObsAuxCovarianceBase_ & errorCovariance(const std::string & id) {
    return this->get(id);
  }
  const ObsAuxCovarianceBase_ & errorCovariance(const std::string & id) const {
    return this->get(id);
  }
  const size_t & size() const {return size_;}

 private:
  ObsAuxCovarianceBase_ & get(const std::string &);
  const ObsAuxCovarianceBase_ & get(const std::string &) const;  
  void print(std::ostream &) const;

  mutable size_t size_;
  ObsAuxCovariancePtrMap_ covs_;
  boost::scoped_ptr<eckit::LocalConfiguration> conf_;
};

// =============================================================================

template<typename MODEL>
ObsAuxCovariance<MODEL>::ObsAuxCovariance(const eckit::Configuration & conf) 
  : size_(0), covs_(), conf_(new eckit::LocalConfiguration(conf))
{
  Log::trace() << "ObsAuxCovariance<MODEL>::ObsAuxCovariance starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxCovariance");

  std::vector<eckit::LocalConfiguration> covconfs;
  conf.get("ObsAuxCovariances", covconfs);
  size_ = covconfs.size();

  for (auto const & config: covconfs) {
    std::string key(config.getString("type"));
    ObsAuxCovarianceBase_ * auxcov = ObsAuxCovarianceFactory::create(config);
    covs_.insert(key, auxcov);
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::ObsAuxCovariance done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
ObsAuxCovariance<MODEL>::~ObsAuxCovariance() {
  Log::trace() << "ObsAuxCovariance<MODEL>::~ObsAuxCovariance starting" << std::endl;
  util::Timer timer(classname(), "~ObsAuxCovariance");
  Log::trace() << "ObsAuxCovariance<MODEL>::~ObsAuxCovariance done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
ObsAuxCovarianceBase & ObsAuxCovariance<MODEL>::get(const std::string & id) {
  iter_ it = covs_.find(id);
  ASSERT(it != covs_.end());
  return *it->second;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
const ObsAuxCovarianceBase & ObsAuxCovariance<MODEL>::get(const std::string & id) const {
  icst_ it = covs_.find(id);
  ASSERT(it != covs_.end());
  return *it->second;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariance<MODEL>::linearize(const ObsAuxControl_ & xx) {
  Log::trace() << "ObsAuxCovariance<MODEL>::linearize starting" << std::endl;
  util::Timer timer(classname(), "linearize");
  for (iter_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->linearize(xx.obsauxcontrol(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariance<MODEL>::update(const ObsAuxControl_ & xx, 
                                     const eckit::Configuration & conf) {
  Log::trace() << "ObsAuxCovariance<MODEL>::update starting" << std::endl;
  util::Timer timer(classname(), "update");

  conf_.reset(new eckit::LocalConfiguration(conf));

  std::vector<eckit::LocalConfiguration> confs;
  conf.get("ObsAuxCovariances", confs);

  for (iter_ it = covs_.begin(); it != covs_.end(); ++it) {
    for (auto const & config: confs) {
      if (config.getString("type") == it->first &&
          config.getBool("update")) {
        it->second->update(xx.obsauxcontrol(it->first), config);
      }
    }
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::update done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariance<MODEL>::multiply(const ObsAuxIncrement_ & dx1, 
                                       ObsAuxIncrement_ & dx2) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::multiply starting" << std::endl;
  util::Timer timer(classname(), "multiply");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->multiply(dx1.obsauxincrement(it->first), 
                         dx2.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::multiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariance<MODEL>::inverseMultiply(const ObsAuxIncrement_ & dx1,
                                              ObsAuxIncrement_ & dx2) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::inverseMultiply starting" << std::endl;
  util::Timer timer(classname(), "inverseMultiply");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->inverseMultiply(dx1.obsauxincrement(it->first), 
                                dx2.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::inverseMultiply done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariance<MODEL>::multiplySqrt(const ObsAuxCtlVec_ & dvin, 
                                           ObsAuxIncrement_ & dxout) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::multiplySqrt starting" << std::endl;
  util::Timer timer(classname(), "multiplySqrt");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->multiplySqrt(dvin.obsAuxCtlVec(it->first), 
                             dxout.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::multiplySqrt done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariance<MODEL>::multiplySqrtTrans(const ObsAuxIncrement_ & dxin, 
                                                ObsAuxCtlVec_ & dvout) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::multiplySqrtTrans starting" << std::endl;
  util::Timer timer(classname(), "multiplySqrtTrans");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->multiplySqrtTrans(dxin.obsauxincrement(it->first), 
                                  dvout.obsAuxCtlVec(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::multiplySqrtTrans done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariance<MODEL>::randomize(ObsAuxIncrement_ & dx) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::randomize starting" << std::endl;
  util::Timer timer(classname(), "randomize");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    it->second->randomize(dx.obsauxincrement(it->first));
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::randomize done" << std::endl;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsAuxCovariance<MODEL>::print(std::ostream & os) const {
  Log::trace() << "ObsAuxCovariance<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  for (icst_ it = covs_.begin(); it != covs_.end(); ++it) {
    os << it->first << ": " << *it->second;
  }
  Log::trace() << "ObsAuxCovariance<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXCOVARIANCE_H_
