/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSAUXCTLVEC_H_
#define OOPS_INTERFACE_OBSAUXCTLVEC_H_

#include <iostream>
#include <string>

#include <boost/ptr_container/ptr_map.hpp>

#include "oops/interface/ObsAuxCtlVecBase.h"

#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
}

namespace oops {

  /// ObsAuxCtlVec Class
  /*!
   *  Part of the control vector corresponding to the ObsAuxIncrement part of the
   *  control increment
   */

template <typename MODEL>
class ObsAuxCovariance;

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObsAuxCtlVec : public util::Printable,
                     private util::ObjectCounter<ObsAuxCtlVec<MODEL> > {
  typedef ObsAuxCtlVecBase        ObsAuxCtlVecBase_;
  typedef ObsAuxCovariance<MODEL> ObsAuxCovariance_;

  typedef boost::ptr_map<std::string, ObsAuxCtlVecBase_> ObsAuxCtlVecPtrMap_;
  typedef typename ObsAuxCtlVecPtrMap_::iterator iter_;
  typedef typename ObsAuxCtlVecPtrMap_::const_iterator icst_;

 public:
  static const std::string classname() {return "oops::ObsAuxCtlVec";}

/// Constructor, destructor
  ObsAuxCtlVec(const ObsAuxCovariance_ &);
  ObsAuxCtlVec(const ObsAuxCovariance_ &, const ObsAuxCtlVec &);
  ObsAuxCtlVec(const ObsAuxCtlVec &, const bool copy = true);
  ~ObsAuxCtlVec();

/// Linear algebra operators
  void zero();
  ObsAuxCtlVec & operator=(const ObsAuxCtlVec &);
  ObsAuxCtlVec & operator+=(const ObsAuxCtlVec &);
  ObsAuxCtlVec & operator-=(const ObsAuxCtlVec &);
  ObsAuxCtlVec & operator*=(const double &);
  void axpy(const double &, const ObsAuxCtlVec &);
  double dot_product_with(const ObsAuxCtlVec &) const;

/// I/O and diagnostics
  void read(const eckit::Configuration &);
  void write(const eckit::Configuration &) const;
  double norm() const;

/// Interfacing
  const ObsAuxCtlVecBase_ & obsAuxCtlVec(const std::string & id) const {
    return this->get(id);
  }
  ObsAuxCtlVecBase_ & obsAuxCtlVec(const std::string & id) {
    return this->get(id);
  }

  const std::vector<std::string> & getKeys() const {return keys_;}

  const size_t & size() const {return size_;}

 private:
  ObsAuxCtlVecBase_ & get(const std::string &);
  const ObsAuxCtlVecBase_ & get(const std::string &) const;
  void print(std::ostream &) const;

  mutable size_t size_;
  std::vector<std::string> keys_;
  ObsAuxCtlVecPtrMap_ ctlvecs_;
};

// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVec<MODEL>::ObsAuxCtlVec(const ObsAuxCovariance_ & cov) 
  : size_(0), keys_(), ctlvecs_()
{
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxCtlVec");

  std::vector<eckit::LocalConfiguration> covconfs;
  cov.config().get("ObsAuxCovariances", covconfs);
  size_ = covconfs.size();

  for (auto const & config: covconfs) {
    std::string key(config.getString("type"));
    ObsAuxCtlVecBase * auxctl = ObsAuxCtlVecFactory::create(cov.errorCovariance(key));
    ctlvecs_.insert(key, auxctl);
    keys_.push_back(key);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVec<MODEL>::ObsAuxCtlVec(const ObsAuxCovariance_ & cov,
                                  const ObsAuxCtlVec & other) 
  : size_(other.size_), keys_(), ctlvecs_()
{
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec interpolated starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxCtlVec");
  for (icst_ it = other.ctlvecs_.begin(); it != other.ctlvecs_.end(); ++it) {
    std::string incname = it->first;
    ObsAuxCtlVecBase * auxctl = it->second->clone(cov);
    ctlvecs_.insert(incname, auxctl);
    keys_.push_back(incname);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec interpolated done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVec<MODEL>::ObsAuxCtlVec(const ObsAuxCtlVec & other,
                                  const bool copy) 
  : size_(other.size_), keys_(), ctlvecs_()
{
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec copy starting" << std::endl;
  util::Timer timer(classname(), "ObsAuxCtlVec");
  for (icst_ it = other.ctlvecs_.begin(); it != other.ctlvecs_.end(); ++it) {
    std::string incname = it->first;
    ObsAuxCtlVecBase * auxctl = it->second->clone(copy);
    ctlvecs_.insert(incname, auxctl);
    keys_.push_back(incname);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::ObsAuxCtlVec copy done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVec<MODEL>::~ObsAuxCtlVec() {
  Log::trace() << "ObsAuxCtlVec<MODEL>::~ObsAuxCtlVec starting" << std::endl;
  util::Timer timer(classname(), "~ObsAuxCtlVec");
  Log::trace() << "ObsAuxCtlVec<MODEL>::~ObsAuxCtlVec done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVecBase & ObsAuxCtlVec<MODEL>::get(const std::string & id) {
  iter_ it = ctlvecs_.find(id);
  ASSERT(it != ctlvecs_.end());
  return *it->second;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
const ObsAuxCtlVecBase & ObsAuxCtlVec<MODEL>::get(const std::string & id) const {
  icst_ it = ctlvecs_.find(id);
  ASSERT(it != ctlvecs_.end());
  return *it->second;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ObsAuxCtlVec<MODEL>::zero() {
  Log::trace() << "ObsAuxCtlVec<MODEL>::zero starting" << std::endl;
  util::Timer timer(classname(), "zero");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    it->second->zero();
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::zero done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVec<MODEL> & ObsAuxCtlVec<MODEL>::operator=(const ObsAuxCtlVec & rhs) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator= starting" << std::endl;
  util::Timer timer(classname(), "operator=");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    *(it->second) = rhs.obsAuxCtlVec(it->first);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVec<MODEL> & ObsAuxCtlVec<MODEL>::operator+=(const ObsAuxCtlVec & rhs) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator+= starting" << std::endl;
  util::Timer timer(classname(), "operator+=");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    *(it->second) += rhs.obsAuxCtlVec(it->first);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator+= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVec<MODEL> & ObsAuxCtlVec<MODEL>::operator-=(const ObsAuxCtlVec & rhs) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator-= starting" << std::endl;
  util::Timer timer(classname(), "operator-=");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    *(it->second) -= rhs.obsAuxCtlVec(it->first);
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator-= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
ObsAuxCtlVec<MODEL> & ObsAuxCtlVec<MODEL>::operator*=(const double & zz) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator*= starting" << std::endl;
  util::Timer timer(classname(), "operator*=");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    *(it->second) *= zz;
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::operator*= done" << std::endl;
  return *this;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ObsAuxCtlVec<MODEL>::axpy(const double & zz, const ObsAuxCtlVec & dx) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::axpy starting" << std::endl;
  util::Timer timer(classname(), "axpy");
  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    it->second->axpy(zz, dx.obsAuxCtlVec(it->first));
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::axpy done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
double ObsAuxCtlVec<MODEL>::dot_product_with(const ObsAuxCtlVec & dx) const {
  Log::trace() << "ObsAuxCtlVec<MODEL>::dot_product_with starting" << std::endl;
  util::Timer timer(classname(), "dot_product_with");
  double zz = 0.0;
  for (icst_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    zz += it->second->dot_product_with(dx.obsAuxCtlVec(it->first));
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::dot_product_with done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ObsAuxCtlVec<MODEL>::read(const eckit::Configuration & conf) {
  Log::trace() << "ObsAuxCtlVec<MODEL>::read starting" << std::endl;
  util::Timer timer(classname(), "read");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("ctlvecObsAux")) conf.get("ctlvecObsAux", confs);

  for (iter_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    for (auto const & config: confs) {
      if (config.getString("type") == it->first) 
        it->second->read(config);
    }
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::read done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ObsAuxCtlVec<MODEL>::write(const eckit::Configuration & conf) const {
  Log::trace() << "ObsAuxCtlVec<MODEL>::write starting" << std::endl;
  util::Timer timer(classname(), "write");

  std::vector<eckit::LocalConfiguration> confs;
  if (conf.has("ctlvecObsAux")) conf.get("ctlvecObsAux", confs);

  for (icst_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    for (auto const & config: confs) {
      if (config.getString("type") == it->first) 
        it->second->write(config);
    }
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::write done" << std::endl;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
double ObsAuxCtlVec<MODEL>::norm() const {
  Log::trace() << "ObsAuxCtlVec<MODEL>::norm starting" << std::endl;
  util::Timer timer(classname(), "norm");
  double zz = 0.0;
  for (icst_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    zz += it->second->norm();
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::norm done" << std::endl;
  return zz;
}
// -----------------------------------------------------------------------------
template<typename MODEL>
void ObsAuxCtlVec<MODEL>::print(std::ostream & os) const {
  Log::trace() << "ObsAuxCtlVec<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  for (icst_ it = ctlvecs_.begin(); it != ctlvecs_.end(); ++it) {
    os << it->first << ": " << *it->second;
  }
  Log::trace() << "ObsAuxCtlVec<MODEL>::print done" << std::endl;
}
// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSAUXCTLVEC_H_
