/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSERRORCOVARIANCE_H_
#define OOPS_INTERFACE_OBSERRORCOVARIANCE_H_

#include <string>

#include <boost/noncopyable.hpp>
#include <boost/scoped_ptr.hpp>

#include "oops/assimilation/ControlObsVector.h"
#include "oops/interface/ObsErrorBase.h"
#include "oops/interface/ObservationSpace.h"
#include "oops/interface/ObsVector.h"
#include "util/Printable.h"

namespace eckit {
  class Configuration;
}

namespace oops {

// -----------------------------------------------------------------------------
/// Observation error covariance matrix
/*!
 *  This class provides the operations associated with the observation
 *  error covariance matrix. It wraps the actual observation error covariance
 *  which can be a model specific one or a generic one.
 *  The interface for the observation error comprises two levels (ObsErrorCovariance
 *  and ObsErrorBase) because we want run time polymorphism.
 *  The ObsErrorCovariance does conversion of arguments to templated ObsVector and
 *  the tracing and timing. The ObsErrorBase does the conversion to model specific
 *  ObsVector.
 */

template <typename MODEL>
class ObsErrorCovariance : public util::Printable,
                           private util::ObjectCounter<ObsErrorCovariance<MODEL> >,
                           private boost::noncopyable {
  typedef ObsErrorBase<MODEL>        ObsErrorBase_;
  typedef ObservationSpace<MODEL>    ObsSpace_;
  typedef ObsVector<MODEL>           ObsVector_;
  typedef ControlObsVector<MODEL>    ControlObsVector_;

 public:
  static const std::string classname() {return "oops::ObsErrorCovariance";}

  ObsErrorCovariance(const ObsSpace_ &, const eckit::Configuration &);
  ~ObsErrorCovariance();

/// Linearize and reset for inner loop if needed
  void linearize(const ControlObsVector_ &);
  void setupWeights(const ControlObsVector_ &);

/// Multiply a Departure by \f$R\f$ and \f$R^{-1}\f$
  ControlObsVector_ * multiply(const ControlObsVector_ &) const;
  ControlObsVector_ * inverseMultiply(const ControlObsVector_ &) const;

/// Generate random perturbation
  void randomize(ControlObsVector_ &) const;

/// Get mean error for Jo table
  double getRMSE() const;

 private:
  ControlObsVector_ * multiplyWghtSqrt_(const ControlObsVector_ &) const;
  ControlObsVector_ * inverseMultiply_(const ControlObsVector_ &) const;
  void print(std::ostream &) const;
  boost::scoped_ptr<ObsErrorBase_> covar_;
  boost::scoped_ptr<ObsErrorBase_> biascovar_;
  bool lvarqc_;
};

// ====================================================================================

template <typename MODEL>
ObsErrorCovariance<MODEL>::ObsErrorCovariance(const ObsSpace_ & obsdb,
                                              const eckit::Configuration & conf)
  : covar_(0), biascovar_(0), lvarqc_(false)
{
  Log::trace() << "ObsErrorCovariance<MODEL>::ObsErrorCovariance starting" << std::endl;
  util::Timer timer(classname(), "ObsErrorCovariance");
  eckit::LocalConfiguration obscovarconf(conf, "ObsErrorCovariance");
  covar_.reset(ObsErrorFactory<MODEL>::create(obsdb, obscovarconf));
  if (conf.has("BiasErrorCovariance")) {
    eckit::LocalConfiguration biascovarconf(conf, "BiasErrorCovariance");
    biascovar_.reset(ObsErrorFactory<MODEL>::create(obsdb, biascovarconf));
  }
  Log::trace() << "ObsErrorCovariance<MODEL>::ObsErrorCovariance done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObsErrorCovariance<MODEL>::~ObsErrorCovariance() {
  Log::trace() << "ObsErrorCovariance<MODEL>::~ObsErrorCovariance starting" << std::endl;
  util::Timer timer(classname(), "~ObsErrorCovariance");
  covar_.reset();
  biascovar_.reset();
  Log::trace() << "ObsErrorCovariance<MODEL>::~ObsErrorCovariance done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrorCovariance<MODEL>::linearize(const ControlObsVector_ & yy) {
  Log::trace() << "ObsErrorCovariance<MODEL>::linearize starting" << std::endl;
  util::Timer timer(classname(), "linearize");
  lvarqc_ = false;
  covar_->linearize(yy.obsvector().obsvector());
  if (biascovar_) 
    biascovar_->linearize(yy.biasvector().obsvector());
  Log::trace() << "ObsErrorCovariance<MODEL>::linearize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObsErrorCovariance<MODEL>::setupWeights(const ControlObsVector_ & yy) {
  Log::trace() << "ObsErrorCovariance<MODEL>::setupWeights starting" << std::endl;
  util::Timer timer(classname(), "setupWeights");
  covar_->setupWeights(yy.obsvector().obsvector());
  lvarqc_ = true;
  Log::trace() << "ObsErrorCovariance<MODEL>::setupWeights done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> * ObsErrorCovariance<MODEL>::multiply(const ControlObsVector_ & dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::multiply starting" << std::endl;
  util::Timer timer(classname(), "multiply");
  ObsVector_ * dyobs = new ObsVector_(covar_->multiply(dy.obsvector().obsvector()));
  ObsVector_ * dybias = nullptr;
  if (biascovar_) 
    dybias = new ObsVector_(biascovar_->multiply(dy.biasvector().obsvector()));
  ControlObsVector_ * dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::multiply done" << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> * ObsErrorCovariance<MODEL>::multiplyWghtSqrt_(const ControlObsVector_ & dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::multiplyWghtSqrt_ starting" << std::endl;
  util::Timer timer(classname(), "multiplyWghtSqrt_");
  ObsVector_ * dyobs = new ObsVector_(covar_->multiplyWghtSqrt(dy.obsvector().obsvector()));
  ObsVector_ * dybias = nullptr;
  if (biascovar_)
    dybias = new ObsVector_(covar_->multiplyWghtSqrt(dy.biasvector().obsvector()));
  ControlObsVector_ * dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::multiplyWghtSqrt_ done" << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> * ObsErrorCovariance<MODEL>::inverseMultiply_(const ControlObsVector_ & dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiply_ starting" << std::endl;
  util::Timer timer(classname(), "inverseMultiply_");
  ObsVector_ * dyobs = new ObsVector_(covar_->inverseMultiply(dy.obsvector().obsvector()));
  ObsVector_ * dybias = nullptr;
  if (biascovar_) 
    dybias = new ObsVector_(biascovar_->inverseMultiply(dy.biasvector().obsvector()));
  ControlObsVector_ * dyctl = new ControlObsVector_(dyobs, dybias);
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiply_ done" << std::endl;
  return dyctl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ControlObsVector<MODEL> * ObsErrorCovariance<MODEL>::inverseMultiply(const ControlObsVector_ & dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiply starting" << std::endl;
  util::Timer timer(classname(), "inverseMultiply");
  ControlObsVector<MODEL> * wrinvwdy = nullptr;
  if (lvarqc_) {
    ControlObsVector<MODEL> * wdy = this->multiplyWghtSqrt_(dy);
    ControlObsVector<MODEL> * rinvwdy = this->inverseMultiply_(*wdy);
    wrinvwdy = this->multiplyWghtSqrt_(*rinvwdy);
  } else {
    wrinvwdy = this->inverseMultiply_(dy);
  }
  Log::trace() << "ObsErrorCovariance<MODEL>::inverseMultiply done" << std::endl;
  return wrinvwdy;
}

// -----------------------------------------------------------------------------


template <typename MODEL>
void ObsErrorCovariance<MODEL>::randomize(ControlObsVector_ & dy) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::randomize starting" << std::endl;
  util::Timer timer(classname(), "randomize");
  covar_->randomize(dy.obsvector().obsvector());
  Log::trace() << "ObsErrorCovariance<MODEL>::randomize done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
double ObsErrorCovariance<MODEL>::getRMSE() const {
  Log::trace() << "ObsErrorCovariance<MODEL>::getRMSE starting" << std::endl;
  util::Timer timer(classname(), "getRMSE");
  double zz = covar_->getRMSE();
  Log::trace() << "ObsErrorCovariance<MODEL>::getRMSE done" << std::endl;
  return zz;
}

// -----------------------------------------------------------------------------

template<typename MODEL>
void ObsErrorCovariance<MODEL>::print(std::ostream & os) const {
  Log::trace() << "ObsErrorCovariance<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *covar_;
  if (biascovar_) os << *biascovar_;
  Log::trace() << "ObsErrorCovariance<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSERRORCOVARIANCE_H_
