/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_INTERFACE_OBSERVATIONSPACE_H_
#define OOPS_INTERFACE_OBSERVATIONSPACE_H_

#include <ostream>
#include <string>

#include <boost/shared_ptr.hpp>

#include "util/Logger.h"
#include "util/ObjectCounter.h"
#include "util/Printable.h"
#include "util/Timer.h"

namespace eckit {
  class Configuration;
  class LocalConfiguration;
}

namespace util {
  class DateTime;
}

namespace oops {
  template <typename T> class ControlObsVector;
  template <typename T> class GeoVaLs;

// -----------------------------------------------------------------------------

template <typename MODEL>
class ObservationSpace : public util::Printable,
                         private util::ObjectCounter<ObservationSpace<MODEL> > {
  typedef typename MODEL::ObsSpace  ObsSpace_;
  typedef ControlObsVector<MODEL>   ControlObsVector_;
  typedef GeoVaLs<MODEL>            GeoVaLs_;

 public:
  static const std::string classname() {return "oops::ObservationSpace";}

  ObservationSpace(const eckit::Configuration &,
                   const util::DateTime &, 
                   const util::DateTime &,
                   const bool lscreened =false);
  ~ObservationSpace();

/// Interfacing
  ObsSpace_ & observationspace() const {return *obsdb_;}
  const eckit::Configuration & config() const {return conf_;}

/// Assimilation window
  const util::DateTime & windowStart() const {return obsdb_->windowStart();}
  const util::DateTime & windowEnd() const {return obsdb_->windowEnd();}

// Screening  
  void screenObservations(const ControlObsVector_ &, const GeoVaLs_ &) const;

// Other
  const bool & isScreened() const {return obsdb_->isScreened();}
  void generateDistribution(const eckit::Configuration &);
  void printJo(const ControlObsVector_ &, const ControlObsVector_ &) const;

 private:
  void print(std::ostream &) const;

  eckit::LocalConfiguration conf_;
  boost::shared_ptr<ObsSpace_> obsdb_;
};

// -----------------------------------------------------------------------------

template <typename MODEL>
ObservationSpace<MODEL>::ObservationSpace(const eckit::Configuration & conf,
                                          const util::DateTime & bgn,
                                          const util::DateTime & end,
                                          const bool lscreened)
  : conf_(conf), obsdb_() {
  Log::trace() << "ObservationSpace<MODEL>::ObservationSpace starting" << std::endl;
  util::Timer timer(classname(), "ObservationSpace");
  obsdb_.reset(new ObsSpace_(conf, bgn, end, lscreened));
  Log::trace() << "ObservationSpace<MODEL>::ObservationSpace done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
ObservationSpace<MODEL>::~ObservationSpace() {
  Log::trace() << "ObservationSpace<MODEL>::~ObservationSpace" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::screenObservations(const ControlObsVector_ & dy, 
                                                 const GeoVaLs_ & gvals) const {
  Log::trace() << "ObservationSpace<MODEL>::screenObservations starting" << std::endl;
  util::Timer timer(classname(), "printJo");
  obsdb_->screenObservations(dy.obsvector().obsvector(), gvals.geovals());
  Log::trace() << "ObservationSpace<MODEL>::screenObservations done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::generateDistribution(const eckit::Configuration & conf) {
  Log::trace() << "ObservationSpace<MODEL>::generateDistribution starting" << std::endl;
  util::Timer timer(classname(), "generateDistribution");
  obsdb_->generateDistribution(conf);
  Log::trace() << "ObservationSpace<MODEL>::generateDistribution done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::printJo(const ControlObsVector_ & dy, const ControlObsVector_ & grad) const {
  Log::trace() << "ObservationSpace<MODEL>::printJo starting" << std::endl;
  util::Timer timer(classname(), "printJo");
  obsdb_->printJo(dy.obsvector().obsvector(), grad.obsvector().obsvector());
  Log::trace() << "ObservationSpace<MODEL>::printJo done" << std::endl;
}

// -----------------------------------------------------------------------------

template <typename MODEL>
void ObservationSpace<MODEL>::print(std::ostream & os) const {
  Log::trace() << "ObservationSpace<MODEL>::print starting" << std::endl;
  util::Timer timer(classname(), "print");
  os << *obsdb_;
  Log::trace() << "ObservationSpace<MODEL>::print done" << std::endl;
}

// -----------------------------------------------------------------------------

}  // namespace oops

#endif  // OOPS_INTERFACE_OBSERVATIONSPACE_H_
