/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef UTIL_LOGBOOKHELPER_H_
#define UTIL_LOGBOOKHELPER_H_

#include <map>
#include <string>

#include <boost/noncopyable.hpp>
#include <boost/scoped_ptr.hpp>

#include "util/Logbook.h"
#include "util/Logger.h"
#include "util/Printable.h"

#include "eckit/config/LocalConfiguration.h"
#include "eckit/exception/Exceptions.h"
#include "eckit/filesystem/PathName.h"

namespace util {

// -----------------------------------------------------------------------------

class LogbookHelper : public util::Printable,
                      private boost::noncopyable {
  typedef std::map<std::string, int>::const_iterator cit_;
  typedef std::function<void(const eckit::Configuration & conf)> CallbackFunction_;

 public:
  enum Execution : bool {
    Active = true,
    NotActive = false
  };

  enum Callback : bool {
    Call = true,
    DontCall = false
  };

/// Initialization, finalization
  static void start();
  static void stop();

/// Utilities
  template<typename ValueT> 
    static void startRecording(const std::string &, const ValueT &);
  static void finishRecording(const std::string &);

/// Connect callback - allow models to connect to the LogbookHelper
  static void connectCallback(const CallbackFunction_ cb);

/// Accessors
  static bool has(const std::string & s);
  template<typename ValueT> 
    static void query(const std::string & s, ValueT & v);
  template<typename ValueT> 
    static void record(const std::string & s, const ValueT & v, const bool callbacks =false);

/// Interfacing
  static const eckit::Configuration & getConstLogbook();
  static eckit::LocalConfiguration & getLogbook();

  ~LogbookHelper();
 private:
  LogbookHelper();
  static LogbookHelper & getInstance();
  void print(std::ostream & os) const;

  bool on_;
  std::map<std::string, bool> status_;
  std::map<std::string, int> counts_;
  std::vector<CallbackFunction_> callbacks_;
  boost::scoped_ptr<Logbook> logbook_;
};

// -----------------------------------------------------------------------------

template<typename ValueT>
void LogbookHelper::query(const std::string & s, ValueT & v) {
  ASSERT(getInstance().logbook_);
  return getInstance().logbook_->get<ValueT>(s, v);
}

// -----------------------------------------------------------------------------

template<typename ValueT>
void LogbookHelper::record(const std::string & s, const ValueT & v, const bool callbacks) {
  ASSERT(getInstance().logbook_);
  getInstance().logbook_->set<ValueT>(s, v);
  // Call callback functions
  if (callbacks)
    for (auto const & cb: getInstance().callbacks_) {
      cb(getConstLogbook());
    }
}

// -----------------------------------------------------------------------------

template<typename ValueT>
void LogbookHelper::startRecording(const std::string & name, const ValueT & val) {
  if (getInstance().on_) {
    getInstance().status_[name] = Execution::Active;
    getInstance().record<bool>(name+"_active", Execution::Active, Callback::DontCall);
    getInstance().record<ValueT>(name, val, Callback::Call);
  }
}

// -----------------------------------------------------------------------------

}  // namespace util

#endif  // UTIL_LOGBOOKHELPER_H_
