/*
 * (C) Copyright 2009-2016 ECMWF.
 * 
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0. 
 * In applying this licence, ECMWF does not waive the privileges and immunities 
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

#ifndef OOPS_UTIL_RESTART_H_
#define OOPS_UTIL_RESTART_H_

#include <map>
#include <string>

#include <boost/noncopyable.hpp>

#include "eckit/config/Configuration.h"
#include "util/Printable.h"
#include "util/RestartHelper.h"
#include "oops/assimilation/ControlVariable.h"
#include "oops/assimilation/CostFunction.h"
#include "oops/assimilation/Minimizer.h"

namespace util {

// -----------------------------------------------------------------------------

class Restart : public util::Printable,
                private boost::noncopyable {
 typedef std::function<void(eckit::LocalConfiguration & conf)> chkptCallbackFunction_;
 typedef std::function<void(const eckit::Configuration & conf)> rstrtCallbackFunction_;

 public:
  static void setup(const std::vector<std::string> & restarts);

  static void connectChkptCallback(const std::string & str,
                                   const chkptCallbackFunction_ & cb);
  static void connectRstrtCallback(const std::string & str,
                                   const rstrtCallbackFunction_ & cb);
  static void checkpoint(const eckit::Configuration & conf,  
                         const std::vector<std::string> & strs);
  static void restart(const std::vector<std::string> & strs);

  ~Restart() {}
 private:
  Restart();
  static Restart & getInstance();
  static eckit::LocalConfiguration & getConf(const std::string & str);
  static eckit::LocalConfiguration & newConf(const std::string & str);

  void print(std::ostream & os) const;

  bool isRestarted_;
  std::map<std::string, eckit::LocalConfiguration> restartConfigs_;
  std::map<std::string, chkptCallbackFunction_> cbChkpt_;
  std::map<std::string, rstrtCallbackFunction_> cbRstrt_;
};

// =============================================================================

Restart & Restart::getInstance() {
  static Restart theRestart;
  return theRestart;
}

// -----------------------------------------------------------------------------

void Restart::setup(const std::vector<std::string> & restarts) {
  if (util::RestartHelper::isReadIn()) {
    getInstance().isRestarted_ = true;
    for (auto const& restart: restarts) {
      ASSERT(util::RestartHelper::has(restart));
      util::RestartHelper::query<eckit::LocalConfiguration>
                                (restart, getInstance().newConf(restart));
    } 
    oops::Log::info() << getInstance() << std::endl;
  }
}

// -----------------------------------------------------------------------------

void Restart::connectChkptCallback(const std::string & str,
                                   const chkptCallbackFunction_ & cb) {
  auto search = getInstance().cbChkpt_.find(str);
  ASSERT(search == getInstance().cbChkpt_.end());
  getInstance().cbChkpt_[str] = cb;
}

// -----------------------------------------------------------------------------

void Restart::connectRstrtCallback(const std::string & str,
                                   const rstrtCallbackFunction_ & cb) {
  auto search = getInstance().cbRstrt_.find(str);
  ASSERT(search == getInstance().cbRstrt_.end());
  getInstance().cbRstrt_[str] = cb;
}

// -----------------------------------------------------------------------------

void Restart::checkpoint(const eckit::Configuration & conf, 
                         const std::vector<std::string> & strs) {
  if (conf.has("checkpoint")) {
    for (auto const & str: strs) {
      ASSERT(conf.has("checkpoint."+str));
      oops::Log::info() << "Restart::checkpoint " << str << ": " << std::endl;
      eckit::LocalConfiguration lconf(conf,"checkpoint."+str);
      getInstance().cbChkpt_[str](lconf);
      util::RestartHelper::record<eckit::LocalConfiguration>(str, lconf);
    }
    oops::Log::info() << "Restart::checkpoint write restart config to file starting"
                      << std::endl;
    util::RestartHelper::write();
    oops::Log::info() << "Restart::checkpoint write restart config to file done"
                      << std::endl;
  }
}

// -----------------------------------------------------------------------------

void Restart::restart(const std::vector<std::string> & strs) {
  if (getInstance().isRestarted_) {
    for (auto const & str: strs) {
      oops::Log::info() << "Restart::restart " << str << ": " << std::endl;
      eckit::LocalConfiguration lconf(getInstance().getConf(str));
      getInstance().cbRstrt_[str](lconf);
    }
  }
}

// -----------------------------------------------------------------------------

eckit::LocalConfiguration & Restart::getConf(const std::string & str) {
  auto search = getInstance().restartConfigs_.find(str);
  ASSERT(search != getInstance().restartConfigs_.end());
  return search->second;
}

// -----------------------------------------------------------------------------

eckit::LocalConfiguration & Restart::newConf(const std::string & str) {
  auto search = getInstance().restartConfigs_.find(str);
  ASSERT(search == getInstance().restartConfigs_.end());
  eckit::LocalConfiguration lconf;
  getInstance().restartConfigs_[str] = lconf;
  return getInstance().restartConfigs_[str];
}

// -----------------------------------------------------------------------------

Restart::Restart() 
  : isRestarted_(false) {}

// -----------------------------------------------------------------------------

void Restart::print(std::ostream & os) const {
  if (isRestarted_) {
    os << std::endl;
    os << "---------------------------------------------------------------------" << std::endl;
    os << "------------------------- Restart -----------------------------------" << std::endl;
    os << "---------------------------------------------------------------------" << std::endl;
    os << "  Recorded configuration: " << std::endl; 
    for (auto const& restart: restartConfigs_) {
      os << "    " << restart.first << ": " << restart.second << std::endl;
      os << "---------------------------------------------------------------------" << std::endl;
    }
    os << "----------- Restart configuration setup successful ------------------" << std::endl;
    os << "---------------------------------------------------------------------" << std::endl;
  }
}

// =============================================================================

}  // namespace util

#endif  // OOPS_UTIL_RESTART_H_
